#!/usr/bin/env python3
"""
AltSci BikeIM Maildir Gtk+3 Demo
by Javantea
Mar 26, 2015
Based on pyterm1.py by Javantea Jan 31, 2015
Based on mail1.py by Javantea Sept 27, 2007 - Jan 13, 2013

"""
from __future__ import print_function
import sys
import email
import os
import stat
import time
import json
from collections import Counter
import hashlib
import random
import io
import gi
gi.require_version('Gtk', '3.0')
try:
	gi.require_version('GtkSpell', '3.0')
	from gi.repository import GtkSpell
except (ImportError, ValueError):
	print("GtkSpell 3 would give you spell checking.")
	GtkSpell = None
#end try
from gi.repository import Gtk, Gdk, GdkPixbuf, Pango, GObject

import bikeim.mail1 as mail1
try:
	# Optional GPG encryption, decryption, sign, verify, import, and gen key
	import bikeim.gpg3 as gpg3
except ImportError:
	gpg3 = None
#end try

PY2 = mail1.PY2
if not PY2:
	baseUnicode = str
else:
	class FileNotFoundError(OSError):
		pass
	class NotADirectoryError(OSError):
		pass
	class FileExistsError(OSError):
		pass
	baseUnicode = unicode
#end if

DRAG_ACTION = Gdk.DragAction.COPY

def Gdk_color_parse(hex_color):
	color = Gdk.RGBA()
	color.parse(hex_color)
	return  color
#end def Gdk_color_parse(hex_color)

def sanitizeGtk(s):
	"""
	We don't need to sanitize in Gtk. Only in HTML.
	But often we get header objects, so we turn them into strings...
	"""
	return baseUnicode(s)
#end def sanitizeGtk(s)

def sanitizePy2(s):
	"""
	When str() works in Python 3 but unicode() fails in Python 2.
	"""
	if PY2:
		if isinstance(s, baseUnicode):
			return s
		elif isinstance(s, str):
			# FIXME: This doesn't make sense..
			try:
				return s.decode('utf-8', 'replace')
			except UnicodeDecodeError as ude:
				# Backslash escape maybe?
				return repr(s)
			#end try
		#end if
		return str(s)
	#end if
	# str in both?
	return baseUnicode(s)
#end def sanitizePy2(s)

def decodeHeaderGtk(h, toSanitize=True, debug=False, removeNewline=False):
	"""
	We don't need to sanitize in Gtk. Only in HTML.
	"""
	if h == None:
		if toSanitize: return sanitizeGtk(h)
		return ''
	#end if
	origh = h
	if isinstance(h, email.header.Header):
		h = baseUnicode(h)
		h = b''
		for string, whatever in origh._chunks:
			h += string.encode('ascii', 'surrogateescape')
		#next string, whatever
	#end if
	if PY2 and not isinstance(h, baseUnicode):
		# Fix unsanitized in python2, whatever.
		h = h.decode('utf-8', 'replace')
	#end if
	if isinstance(h, baseUnicode) and ('=?' in h):
		output = mail1.decodeQE(h, debug)
	else:
		# Seems legit... How could this be a bad choice? Oh yeah.. 
		# TODO: Ensure that we don't get hacked putting a decoder and a sanitizer together. Shit.
		# Since they are surrogate escaping, we should be surrogate escaping.
		if isinstance(h, baseUnicode):
			output = h
		else:
			intermediate = h
			output = mail1.guessEncoding(intermediate)
		#end if
	#end if
	if removeNewline:
		output = output.replace('\n', '')
	#end if
	if toSanitize:
		return sanitizeGtk(output)
	#end if
	return output
#end def decodeHeaderGtk(s, toSanitize=False, debug=False, removeNewline=False)

def sanitizeGtkButton(s):
	"""
	Buttons use _ as an accelerator marker. We don't use that.
	"""
	return s #.replace('_','__')
#end def sanitizeGtkButton(s)

def alert(s, top_msg='', parent=None):
	dialog = Gtk.MessageDialog(parent, 0, Gtk.MessageType.WARNING,
		Gtk.ButtonsType.OK, top_msg)
	if parent:
		dialog.set_title('{0} - Message'.format(parent.get_title()))
	dialog.format_secondary_text(s)
	response = dialog.run()
	dialog.destroy()
	return response
#end def alert(s, top_msg='', parent=None)

def yes_or_no(s, top_msg='', parent=None):
	dialog = Gtk.MessageDialog(parent, 0, Gtk.MessageType.WARNING,
		Gtk.ButtonsType.YES_NO, top_msg)
	if parent:
		dialog.set_title('{0} - Message'.format(parent.get_title()))
	dialog.format_secondary_text(s)
	response = dialog.run()
	dialog.destroy()
	return response
#end def yes_or_no(s, top_msg='', parent=None)

def toThread(msg_row, threaded_rows, show, show_j, auto_added):
	"""
	threading one row at a time
	msg_row is a single row [msg_icon,ssubject,smfrom,date,subject]
	threaded_rows is a list of already established top level rows
	show is a list of [filename [filename, ...]] which is self.messages
	returns Gtk.TreeIter if it belongs to a thread or None if it should be added to threaded_rows.
	show_j is the current index in show.
	auto_added is an offset to the index in show so that you can match threaded_rows with show
	This function modifies threaded_rows, show.
	"""
	ss = msg_row[4].replace('Re:','').replace('RE:','').strip(' ')
	for i, msg in enumerate(threaded_rows):
		ss2 = msg[4].replace('Re:','').replace('RE:','').strip(' ')
		if ss == ss2:
			threaded_rows[-1].append(msg_row)
			filename = show.pop(show_j)
			x = 0
			if show_j < i:
				# TODO: assert only in dev version?
				print('yes it can', show_j, i)
				x = 1
			#end if
			so = show[auto_added + i - x]
			if not isinstance(so, list):
				so = [so, []]
				show[auto_added + i - x] = so
			#end if
			so[1].append(filename)
			return i
		#end if
	#next i, msg
	return None
#end def toThread(msg_row, threaded_rows, show, show_j)

def groupBySender(messages, rows):
	"""
	Thread by sender and mailing list.
	
	"""
	output = {}
	show_i = {}
	sender_order = []
	for i in range(len(rows)):
		sender = rows[i][2]
		if sender not in sender_order:
			sender_order.append(sender)
		#end if
		if sender in output:
			output[sender][1].append(rows[i])
			show_i[sender][1].append(messages[i])
		else:
			output[sender] = [rows[i], []]
			show_i[sender] = [messages[i], []]
		#end if
	#next i
	show = []
	output_list = []
	#output_list = list(output.values())
	i = 0
	for sender in sender_order:
		#if i < 10:
		#	i += 1
		#	print('d', sender, show_i[sender])
		#end if
		show.append(show_i[sender])
		output_list.append(output[sender])
	#next sender
	return output_list, show
#end def groupBySender(messages, rows)

def toThreads(msgs_rows, show):
	"""
	msgs_rows is an array of [msg_icon,ssubject,smfrom,date,subject]
	show is a an array of filenames.
	We sort msgs_rows and show and then spit it back out.
	In general the ordering must be in the same chrono order except that 
	older things go under newer things they relate to.
	"""
	skips = []
	output = []
	show_out = []
	for i, msg in enumerate(msgs_rows):
		if i in skips: continue
		output.append(msg + [[]])
		show_out.append(show[i])
		(msg_icon,ssubject,smfrom,date, subject) = msg
		for j, msg2 in enumerate(msgs_rows[i+1:]):
			if (i+j+1) in skips: continue
			(msg_icon2,ssubject2,smfrom2,date2, subject2) = msg2
			ss = subject.replace('Re:','').replace('RE:','').strip(' ')
			ss2 = subject2.replace('Re:','').replace('RE:','').strip(' ')
			if (ss == ss2):
				output[-1][5].append(msg2)
				so = show_out[-1]
				if not isinstance(so, list):
					show_out[-1] = [so, []]
					so = show_out[-1]
				#end if
				so[1].append(show[i+j+1])
				skips.append(i+j+1)
			#end if
		#next j, msg2
	#next i, msg
	return output, show_out
#end def toThreads(msgs_rows, show)

def strToAddrList(addrstr, names=None, histogram_to=None):
	"""
	Take a list of receipients and turn them into a list of addresses.
	Examples:
	input:                  output:
	javantea@sono.us        ['javantea@sono.us']
	javantea@sono.us, a     ['javantea@sono.us', 'a']
	javantea@sono.us, <a    ['javantea@sono.us', 'a']
	javantea@sono.us, <a>   ['javantea@sono.us', 'a']
	"blah" <>               ['']
	
	Valid strings look like 'a <a@example.com>, b <b@example.com>'
	
	Optional arguments:
		names is a dict between addr and name.
		histogram_to is a Counter of addr and count.
		
	Looking for a canonicalizer? mail1.canonTo
	"""
	pos = 0
	#print('addr', addr.__class__, addr)
	addr_match = mail1.email_re.match(addrstr)
	addr_list = []
	while addr_match:
		name, addr, name2, addr2, addr3 = addr_match.groups()
		if not name and not addr and not name2 and not addr2 and not addr3: break
		pos += addr_match.span()[1] + 1
		if not addr:
			addr = addr2
			name = name2
		#end if
		if not addr:
			addr = addr3
			name = ''
		#end if
		if not addr:
			# <>
			addr_list.append('')
			if pos >= len(addrstr): break
			continue
		#end if
		addr = addr.replace('\n', '')
		addr = addr.strip()
		if not name: name = addr
		name = name.strip().strip('"').strip()
		addr_list.append(addr)
		if histogram_to != None:
			histogram_to[addr] += 1
		#end if
		if names != None:
			# TODO: make this more sensible.
			if addr not in names or addr != name:
				names[addr] = name.strip()
			#end if
		#end if
		#print('=>', name, addr)
		if pos >= len(addrstr): break
		addr_match = mail1.email_re.match(addrstr[pos:])
	#loop
	return addr_list
#end def strToAddrList(addrstr, names=None, histogram_to=None)

class ConfigWindow:
	"""
	Allows you to configure the mail client.
	"""
	def __init__(self, parent):
		"""
		If you give no parent, showAll will call gtk.main(), so don't screw this up.
		"""
		self.parent = parent
		
		self.builder = Gtk.Builder()
		self.builder.add_from_file(os.path.join(mail1.data_dir, 'glade/config.glade'))
		self.window = self.builder.get_object("config_dialog")
		self.window.set_transient_for(self.parent)
		self.inputs = {}
		for input_box in ["server", 'from', 'inbox', 'max_messages', 'keyid', 'spam', 'lang', 'server_name', 'server_fingerprint']:
			self.inputs[input_box] = self.builder.get_object(input_box + '_input')
		#next input_box
		for checkbox in ["hacker", 'update_loop', "autocomplete"]:
			self.inputs[checkbox] = self.builder.get_object(checkbox + '_checkbox')
		#next checkbox
	
	def load_config(self, config):
		"""
		Take a ConfigParser object and load values from it.
		"""
		self.config = config
		if 'Mail1' not in config:
			self.config['Mail1'] = {'inbox': 'inbox'}
		#end if
		if 'Mail1GUI' not in config:
			self.config['Mail1GUI'] = {'width': 800, 'height': 600, 'subject':384, 'from':144, 'date':200}
		#end if
		self.inputs['server'].set_text(config['Mail1'].get('server', ''))
		self.inputs['from'].set_text(config['Mail1'].get('from', ''))
		self.inputs['inbox'].set_text(config['Mail1'].get('inbox', 'inbox'))
		self.inputs['max_messages'].set_text(config['Mail1GUI'].get('max_messages', '10000'))
		self.inputs['keyid'].set_text(config['Mail1'].get('keyid', ''))
		self.inputs['spam'].set_text(config['Mail1'].get('spam', '~/Mail/.inbox.directory/spam1'))
		self.inputs['lang'].set_text(config['Mail1'].get('lang', ''))
		self.inputs['server_name'].set_text(config['Mail1'].get('server_name', ''))
		self.inputs['server_fingerprint'].set_text(config['Mail1'].get('server_fingerprint', ''))
		if config['Mail1'].getboolean('hacker', False):
			self.inputs['hacker'].set_active(True)
		else:
			self.inputs['hacker'].set_active(False)
		#end if
		if config['Mail1GUI'].getboolean('update_loop', False):
			self.inputs['update_loop'].set_active(True)
		else:
			self.inputs['update_loop'].set_active(False)
		#end if
		if 'autocomplete' in config['Mail1'] and config['Mail1'].getboolean('autocomplete'):
			self.inputs['autocomplete'].set_active(True)
		else:
			self.inputs['autocomplete'].set_active(False)
		#end if
		return True
	#end def load_config(config)

	def get_config(self):
		"""
		Take values from textboxes and return a config similar to the one provided in load_config.
		"""
		for key, input_box in self.inputs.items():
			if key in ['hacker', 'update_loop', 'autocomplete']: continue
			if key == 'max_messages':
				self.config['Mail1GUI'][key] = self.inputs[key].get_text()
			else:
				self.config['Mail1'][key] = input_box.get_text()
			#end if
		#next 
		self.config['Mail1']['hacker'] = str(self.inputs['hacker'].get_active())
		self.config['Mail1GUI']['update_loop'] = str(self.inputs['update_loop'].get_active())
		self.config['Mail1']['autocomplete'] = str(self.inputs['autocomplete'].get_active())
		return self.config
	#end def get_config(self)
	
	def showAll(self):
		"""
		Calls self.window.show_all()
		"""
		self.window.show_all()
		if self.parent == None:
			# This might screw stuff up...
			Gtk.main()
		#end if
		return True
	#end def showAll()
	
	def run(self):
		return self.window.run()
	#end def run()
	
	def destroy(self):
		return self.window.destroy()
	#end def destroy()
	
#end class ConfigWindow

def getMessage(message, index, directory):
	"""
	A simple function to abstract the indexing process.
	Modifies index.
	Returns msg which is a dict of:
	{'subject':subject, 'from':mfrom, 'to':to, 'date':date, 
	 'X-Spam-Level':spam_level, 'Received-SPF':spf, 'DKIM-Signature':dkim, 
	 'attach':attach}
	None of the fields are guaranteed except for 
	'subject', 'from', 'to', 'date'.
	"""
	if message not in index:
		fname = os.path.join(directory, message)
		fp = open(fname, 'rb')
		msg = mail1.message_from_binary_file(fp, fname)
		# Write to the index
		# These are Header objects, so we convert to str.
		msg_dict = {}
		msg_dict['subject'] = baseUnicode(msg['subject'])
		msg_dict['from']    = baseUnicode(msg['from'])
		msg_dict['to']      = baseUnicode(msg['to'])
		msg_dict['date']    = baseUnicode(msg['date'])
		if 'X-Spam-Level' in msg:
			msg_dict['X-Spam-Level'] = msg.get('X-Spam-Level', '')
		#end if
		if 'Received-SPF' in msg:
			msg_dict['Received-SPF'] = msg.get('Received-SPF', '')
		#end if
		if 'DKIM-Signature' in msg:
			msg_dict['DKIM-Signature']    = '1'
		#end if
		
		attach = False
		for b in msg.walk():
			ctype = b.get_content_type()
			if ctype not in MaildirWindow.legit_types:
				attach = True
				break
			#end if
		#next b
		if attach:
			msg['attach'] = msg_dict['attach'] = '1'
		#end if
		
		# A fairly cheap index.
		index[message] = msg_dict
	else:
		fname = os.path.join(directory, message)
		#indexed += 1
		# Pull from the index
		msg = index[message]
	#end if
	return msg
#end def getMessage(message, index, directory)

def getMessageRow(msg, config, blacklist=[], blacklist_ip=[], whitelist=[], iconList=[]):
	"""
	A large section of the drawDir code.
	"""
	date_local = config['Mail1GUI'].getboolean('date_local', False)
	subject = msg['subject']
	lsubject = decodeHeaderGtk(subject, True, removeNewline=True)
	mfrom = msg['from']
	mfrom = decodeHeaderGtk(mfrom, True, removeNewline=True)
	mto = msg['to']
	mto = decodeHeaderGtk(mto, True, removeNewline=True)
	date = decodeHeaderGtk(msg['date'], True, removeNewline=True)
	if date_local:
		# FIXME: This is just broken.
		#try:
		date = mail1.decodeDateHeader(date)
		#except 
	#end if
	spam_level = len(msg.get('X-Spam-Level', ''))
	spf = str(msg.get('Received-SPF', ''))
	# SPF Softfail is a useful indicator of spam, but..
	if spf.startswith('softfail'):
		# Constant Contact blows.
		if msg.get('DKIM-Signature', '') == '':
			spam_level = max(spam_level, 5)
			#print("SPF softfail without DKIM", subject, fname)
		#end if
	elif spf.startswith('none'):
		# This should certainly be configurable because some people are idiots. Most people are not.
		spam_level = max(spam_level, 3)
		#print("SPF none", subject, fname)
	#end if
	
	# DKIM is now an interesting value. I don't use it because in 
	# increases exposure on my server. It doesn't help me enough yet.
	if msg.get('DKIM-Signature', '') == '':
		spam_level += 2
		#print("no DKIM", subject, fname)
	#end if
	
	mfromSan = mfrom
	# TODO: Remove the name part and the angle brackets.
	# It could be in both lists, so we check.
	if (spam_level < 6) and (mfromSan in blacklist):
		spam_level = 6
		#print("blacklist", subject, fname, date)
	#end if
	if mfromSan in whitelist:
		spam_level = 0
	#end if
	
	# If we're on the sidelines and it has received header, check IP blacklist.
	if spam_level > 0 and spam_level < 6:
		mip = mail1.getReceivedIp(msg, config['Mail1'].get('server_name', ''))
		if mip:
			if mip in blacklist_ip:
				spam_level = 6
				# Cannot put this into the index but we would if we could.
			#end if
		#end if
	#end if

	#spam_icon = '<span style="background-color:' + mail1.spam_color[min(spam_level, len(mail1.spam_color)-1)] + '">&nbsp;</span>'
	msg_icon = iconList[0]
	if spam_level > 5:
		msg_icon = iconList[5]
	elif spam_level == 5:
		msg_icon = iconList[4]
	elif spam_level == 4:
		msg_icon = iconList[3]
	elif spam_level == 3:
		msg_icon = iconList[2]
	elif spam_level == 2:
		msg_icon = iconList[1]
	#end if
	smfrom = mfrom
	ssubject = sanitizeGtk(lsubject)
	lsubject = sanitizeGtk(lsubject)
	msg_row = [msg_icon, ssubject, smfrom, mto, date, lsubject, mfrom, spam_level]
	return msg_row
#end def getMessageRow(msg, config, blacklist=[], blacklist_ip=[], whitelist=[], iconList=[])

class AddressBookWindow:
	"""
	Allows you to compose and send a message.
	"""
	def __init__(self, parent):
		"""
		If you give no parent, showAll will call gtk.main(), so don't screw this up.
		"""
		self.parent = parent
		self.addressesChanged = False
		hidden_dir = mail1.hidden_dir
		
		self.builder = Gtk.Builder()
		self.builder.add_from_file(os.path.join(mail1.data_dir, 'glade/addressbook1.glade'))
		self.window = self.builder.get_object("AddressBook")
		if self.parent.composer:
			print('composer')
			self.window.set_transient_for(self.parent.composer.window)
		else:
			print('not composer', self.parent.composer)
			# FIXME: AddressBookWindow should just be a window. This causes MaildirWindow to get focus when To/Cc is pressed.
			#self.window.set_transient_for(self.parent.window)
		#end if
		handlers = {"on_AddressBook_destroy": self.close,
			"ToButton_clicked": self.clickTo,
			"CcButton_clicked": self.clickCc,
			"CancelButton_clicked": self.close,
			"AddButton_clicked": self.addContact,
			"UpdateButton_clicked":self.update,
			"Contacts_key_press": self.keyContacts
			}
		self.builder.connect_signals(handlers)
		self.contactStore = self.builder.get_object("ContactsStore")
		self.contactTable = self.builder.get_object("Contacts")
		rendererName = Gtk.CellRendererText()
		rendererName.ellipsize = Pango.EllipsizeMode.END
		rendererName.set_property('editable', True)
		rendererName.connect('edited', self.editedName)
		rendererEmail = Gtk.CellRendererText()
		rendererEmail.ellipsize = Pango.EllipsizeMode.END
		rendererEmail.set_property('editable', True)
		rendererEmail.connect('edited', self.editedEmail)
		name_column = Gtk.TreeViewColumn("Name", rendererName, text=0)
		name_column.set_resizable(True)
		name_column.set_fixed_width(260)
		email_column = Gtk.TreeViewColumn("E-Mail", rendererEmail, text=1)
		email_column.set_resizable(True)
		self.contactTable.append_column(name_column)
		self.contactTable.append_column(email_column)
		self.contactStore.clear()
		select = self.contactTable.get_selection()
		select.set_mode(Gtk.SelectionMode.MULTIPLE)
		self.undo = []
		# TODO: sort by mails direct from/to. Should be in an index or something.
		"""
		addrsList = []
		if os.path.exists(os.path.join(hidden_dir, 'whitelist.txt')):
			addrsList = open(os.path.join(hidden_dir, 'whitelist.txt'), 'r', errors='surrogatereplace').read().split('\n')
		#end if
		"""
		# ~/.mail/addressbook.json is generated from sent-mail.
		# TODO: Update address book as mail is sent to new people.
		addrs = []
		if os.path.exists(os.path.join(hidden_dir, 'addressbook.json')):
			if PY2:
				addrsList2 = open(os.path.join(hidden_dir, 'addressbook.json'), 'r').read().split('\n')
			else:
				addrsList2 = open(os.path.join(hidden_dir, 'addressbook.json'), 'r', errors='surrogatereplace').read().split('\n')
			#end if
			for addr in addrsList2:
				if addr == '': continue
				addrs.append(json.loads(addr))
			#next addr
			if None in addrs:
				addrs.remove(None)
			#end if
			if addrs == []:
				# Remove junk file.
				os.remove(os.path.join(hidden_dir, 'addressbook.json'))
			#end if
		else:
			addrs = self.update()
		#end if

		for name, addr, count in addrs:
			#print(name, addr, c)
			self.contactStore.append((name, addr))
		#next name, addr
		self.addrs = addrs
		self.onlyAddrs = [addr for name, addr, count in addrs]
		self.knownAddrs = None
	#end def AddressBookWindow(parent)
	
	def update(self, x=None):
		"""
		Address book becomes out of date.
		Don't remove a person from the address book.
		"""
		hidden_dir = mail1.hidden_dir
		# TODO: When was the address book last updated?
		# Build a list of recipients from sent-mail.
		sentDir = os.path.join(self.parent.maildir, 'sent-mail', 'cur')
		# Copied and pasted from drawMsg
		index = mail1.getIndex('sent-mail', 0)
		messages = mail1.readDir(sentDir)
		show = messages['files']
		show.sort()
		#print('show', len(show))
		addrs = []
		histogram_to = Counter()
		names = {}
		#indexed = 0
		for message in show:
			msg = getMessage(message, index, sentDir)
			addr = msg['to']
			if addr == None: continue
			tenc = decodeHeaderGtk(addr, False)
			addrstr = sanitizeGtk(tenc)
			#print('addr', addr.__class__, addr)
			addrs.extend(strToAddrList(addrstr, names, histogram_to))
		#next message
		print('hist', len(histogram_to))
		addrs = []
		for addr, c in histogram_to.most_common():
			if addr == '': continue
			addrs.append([names[addr], addr, c])
		#next addr, c
		# Save it to save a ton of time.
		open(os.path.join(hidden_dir, 'addressbook.json'), 'w').write('\n'.join([json.dumps(x) for x in addrs]))
		#print("indexed:", indexed)
		self.addrs = addrs
		self.onlyAddrs = [addr for name, addr, c in addrs]
		return addrs
	#end def update(x=None)
	
	def addMessage(self, tostr):
		"""
		Update the address book incrementally.
		"""
		news = []
		names = {}
		histogram_to = Counter()
		v = strToAddrList(tostr, names, histogram_to)
		for addr in v:
			if addr in self.onlyAddrs:
				# Update the count on addr
				addr_i = self.onlyAddrs.index(addr)
				self.addrs[addr_i][2] += 1
				# Is it worthwhile to update the file and the gui? Just the file.
				self.addressesChanged = True
			else:
				self.addrs.append([names[addr], addr, histogram_to[addr]])
				self.onlyAddrs.append(addr)
				# Add it to the file and the gui
				news.append(self.addrs[-1])
				self.addressesChanged = True
			#end if
		#next addr
		if news:
			for name, addr, count in news:
				#print(name, addr, c)
				self.contactStore.append((name, addr))
			#next name, addr
		#end if
		return True
	#end def addMessage(tostr)
	
	def loadKnownAddrs(self, x=None, update=False):
		"""
		Copied from above addressbook.json loader with minor changes.
		~/.mail/known_addrs.json is generated from inbox
		update means rebuild the list from scratch.
		"""
		hidden_dir = mail1.hidden_dir
		# TODO: threaded so that this doesn't slow down address book loading.
		# TODO: Update  as mail is sent to new people.
		addrs = []
		if (not update) and os.path.exists(os.path.join(hidden_dir, 'known_addrs.json')):
			if PY2:
				addrsList2 = open(os.path.join(hidden_dir, 'known_addrs.json'), 'r').read().split('\n')
			else:
				addrsList2 = open(os.path.join(hidden_dir, 'known_addrs.json'), 'r', errors='surrogatereplace').read().split('\n')
			#end if
			for addr in addrsList2:
				if addr == '': continue
				addrs.append(json.loads(addr))
			#next addr
			if None in addrs:
				addrs.remove(None)
			#end if
			if addrs == []:
				# Remove junk file.
				os.remove(os.path.join(hidden_dir, 'known_addrs.json'))
			#end if
		else:
			# Build a list of recipients from sent-mail.
			inbox_dir = self.parent.config['Mail1'].get('inbox', 'inbox')
			inboxPath = os.path.join(self.parent.maildir, inbox_dir, 'cur')
			# Copied and pasted from drawMsg
			index = mail1.getIndex(inbox_dir, 0)
			#print("index size:", len(index))
			messages = mail1.readDir(inboxPath)
			# TODO: More? What have you done for me lately?
			show = messages['files']
			show.sort(reverse=True)
			show = show[:2000]
			histogram_from = Counter()
			names = {}
			#indexed = 0
			for message in show:
				msg = getMessage(message, index, inboxPath)
				# From instead of to
				addr = msg['from']
				if addr == None: continue
				tenc = decodeHeaderGtk(addr, False)
				addrstr = sanitizeGtk(tenc)
				#print('addr', addr.__class__, addr)
				addrs.extend(strToAddrList(addrstr, names, histogram_from))
			#next message
			# We want the most common first to save computation.
			addrs = []
			for addr, c in histogram_from.most_common():
				addrs.append(addr)
			#next addr, c
			# Save it to save a ton of time.
			open(os.path.join(hidden_dir, 'known_addrs.json'), 'w').write('\n'.join([json.dumps(x) for x in addrs]))
			#print("indexed:", indexed)
		#end if
		self.knownAddrs = addrs
		return True
	#end def loadKnownAddrs(x=None, update=False)

	def showAll(self):
		"""
		Calls self.window.show_all()
		"""
		self.window.show_all()
		if self.parent == None:
			# This might screw stuff up...
			Gtk.main()
		#end if
		return True
	#end def showAll()
	
	def deleteContacts(self):
		"""
		Deletes a contact from the address book
		"""
		list_store, rows = self.contactTable.get_selection().get_selected_rows()
		to_delete = []
		j = 0
		for row in rows:
			i = row.get_indices()[0]
			contact = self.addrs.pop(i - j)
			self.onlyAddrs.pop(i - j)
			self.undo.append(('delete', i - j, contact))
			to_delete.append(i - j)
			self.contactStore.remove(self.contactStore.iter_nth_child(None, i - j))
			j += 1
		#next row
		self.addressesChanged = True
		return True
	#end def deleteContacts()
	
	def addContact(self, widget=None):
		"""
		Add a contact to the address book
		"""
		builder = Gtk.Builder()
		builder.add_from_file(os.path.join(mail1.data_dir, 'glade/contact_dialog.glade'))
		dialog = builder.get_object("contact_dialog")
		dialog.set_transient_for(self.window)
		name = builder.get_object("name")
		email = builder.get_object("email")
		response = dialog.run()
		if response == Gtk.ResponseType.OK:
			# Save the contact to the address book
			name_str = name.get_text()
			email_str = email.get_text()
			self.addrs.append([name_str, email_str, 1])
			self.onlyAddrs.append(email_str)
			self.contactStore.append((name_str, email_str))
			self.undo.append(('add', len(self.addrs) - 1))
			self.addressesChanged = True
		#end if
		dialog.destroy()
		
		return True
	#end def addContact(widget=None)
	
	def editedName(self, cell_renderer, path, data=''):
		"""
		Save edit
		"""
		path_i = int(path)
		self.undo.append(('edit', path_i, self.addrs[path_i][:]))
		self.contactStore[path][0] = data
		self.addrs[path_i][0] = data
		self.addressesChanged = True
		return True
	#end def editedName(self, cell_renderer, path, data='')
	
	def editedEmail(self, cell_renderer, path, data=''):
		"""
		Save edit
		"""
		path_i = int(path)
		self.undo.append(('edit', path_i, self.addrs[path_i][:]))
		self.contactStore[path][1] = data
		self.addrs[path_i][1] = data
		self.onlyAddrs[path_i] = data
		self.addressesChanged = True
		return True
	#end def editedEmail(self, cell_renderer, path, data='')
	
	def saveAddressBook(self):
		"""
		Save once on exit.
		"""
		#print('saving address book?', self.addressesChanged)
		if self.addressesChanged:
			hidden_dir = mail1.hidden_dir
			# Sort by histogram.
			self.addrs.sort(key=lambda x:x[2], reverse=True)
			open(os.path.join(hidden_dir, 'addressbook.json'), 'w').write('\n'.join([json.dumps(x) for x in self.addrs]))
			self.addressesChanged = False
		#end if
		return True
	#end def saveAddressBook(self)
	
	def showSearchContacts(self):
		# Currently searching contacts is handled by treeview but it might be nice to handle it ourselves.
		# See MaildirWindow.searchDir
		# TODO: Search contacts
		return True
	#end def showSearchContacts()
	
	def keyContacts(self, widget, event):
		"""
		Add the Contact to Cc:
		Carbon copy is useful for relationships where someone isn't intended to contribute.
		"""
		#print(repr(event.string), event.keyval)
		#if event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == 102:
		#	# Search Contacts
		#	self.showSearchContacts()
		#	return True
		#end if
		if event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == ord('z'):
			# ctrl-z Undo
			if len(self.undo) > 0:
				last_action = self.undo.pop()
				if last_action[0] == 'delete':
					self.addrs.insert(last_action[1], last_action[2])
					self.onlyAddrs.insert(last_action[1], last_action[2][1])
					self.contactStore.insert(last_action[1], last_action[2])
				elif last_action[0] == 'add':
					self.addrs.pop(last_action[1])
					self.onlyAddrs.pop(last_action[1])
					self.contactStore.remove(self.contactStore.iter_nth_child(None, last_action[1]))
				elif last_action[0] == 'edit':
					self.addrs[last_action[1]] = last_action[2]
					self.onlyAddrs[last_action[1]] = last_action[2][1]
					self.contactStore[last_action[1]] = last_action[2]
				#end if
			#end if
			return True
		#end if
		#print('contacts', repr(event.string), event.keyval)
		if event.string == '\x7f':
			# Delete key
			self.deleteContacts()
			return True
		#end if
		return False
	#end def keyContacts(widget, event)
	
	def clickCc(self, x=None):
		"""
		Add the Contact to Cc:
		Carbon copy is useful for relationships where someone isn't intended to contribute.
		"""
		return self.clickTo(x, cc=True)
	#end def clickCc(x=None)

	def clickTo(self, x=None, cc=False):
		"""
		Add the contact to To: or Cc:
		"""
		if self.parent == None:
			self.close()
			return True
		#end if
		if self.parent.composer == None:
			# We need composer.
			self.parent.clickCompose()
			# A rare case might be that they refuse to compose.
			if self.parent.composer == None:
				return True
			#end if
		#end if
		if cc:
			self.parent.composer.cc = self.parent.composer.inputCc.get_text()
			if self.parent.composer.cc:
				self.parent.composer.cc += ', '
			#end if
		else:
			self.parent.composer.to = self.parent.composer.inputTo.get_text()
			if self.parent.composer.to:
				self.parent.composer.to += ', '
			#end if
		#end if
		cursor = self.contactTable.get_cursor()
		# FIXME: Strange bug.
		if cursor == None or cursor[0] == None or cursor[0][0] == None:
			self.close()
			return True
		#end if
		#print('cursor', cursor, cursor[0][0], )
		pos = cursor[0][0]
		name, addr = self.contactStore[pos]
		fullAddr = addr
		if name != '' and name != addr:
			fullAddr = mail1.nameSanitize(name) + ' <' + addr + '>'
		#end if
		self.parent.composer.showAll()
		self.parent.composer.window.grab_focus()
		if cc:
			self.parent.composer.cc += fullAddr
			self.parent.composer.inputCc.set_text(self.parent.composer.cc)
			self.parent.composer.inputCc.grab_focus()
		else:
			self.parent.composer.to += fullAddr
			self.parent.composer.inputTo.set_text(self.parent.composer.to)
			self.parent.composer.inputTo.grab_focus()
		#end if
		self.close()
		return True
	#end def clickTo(self, x=None, cc=False)

	def close(self, x=None):
		self.saveAddressBook()
		self.window.hide()
		return True
	#end def close(self, x)
#end class AddressBookWindow

class FiltersWindow:
	"""
	Allows you to compose and send a message.
	"""
	def __init__(self, parent):
		"""
		If you give no parent, showAll will call gtk.main(), so don't screw this up.
		"""
		self.parent = parent
		self.filtersChanged = False
		hidden_dir = mail1.hidden_dir
		
		self.builder = Gtk.Builder()
		self.builder.add_from_file(os.path.join(mail1.data_dir, 'glade/filters.glade'))
		self.window = self.builder.get_object("FilterEditor")
		self.window.set_transient_for(self.parent.window)
		handlers = {"on_FilterEditor_destroy": self.close,
			"SaveButton_clicked": self.clickSave,
			"CancelButton_clicked": self.close,
			"AddButton_clicked": self.addFilter,
			"Filters_key_press": self.keyFilters
			}
		self.builder.connect_signals(handlers)
		self.filterStore = self.builder.get_object("FilterStore")
		self.filterTable = self.builder.get_object("Filters")
		rendererField = Gtk.CellRendererText()
		rendererField.ellipsize = Pango.EllipsizeMode.END
		rendererField.set_property('editable', True)
		rendererField.connect('edited', self.editedField)
		rendererValue = Gtk.CellRendererText()
		rendererValue.ellipsize = Pango.EllipsizeMode.END
		rendererValue.set_property('editable', True)
		rendererValue.connect('edited', self.editedValue)
		rendererDir = Gtk.CellRendererText()
		rendererDir.ellipsize = Pango.EllipsizeMode.END
		rendererDir.set_property('editable', True)
		rendererDir.connect('edited', self.editedDirectory)
		field_column = Gtk.TreeViewColumn("Field", rendererField, text=0)
		field_column.set_resizable(True)
		field_column.set_fixed_width(260)
		value_column = Gtk.TreeViewColumn("Value", rendererValue, text=1)
		value_column.set_resizable(True)
		value_column.set_fixed_width(260)
		dir_column = Gtk.TreeViewColumn("Directory", rendererDir, text=2)
		dir_column.set_resizable(True)
		self.filterTable.append_column(field_column)
		self.filterTable.append_column(value_column)
		self.filterTable.append_column(dir_column)
		self.filterStore.clear()
		select = self.filterTable.get_selection()
		select.set_mode(Gtk.SelectionMode.MULTIPLE)
		self.undo = []
		# TODO: sort by mails direct from/to. Should be in an index or something.
		"""
		addrsList = []
		if os.path.exists(os.path.join(hidden_dir, 'whitelist.txt')):
			addrsList = open(os.path.join(hidden_dir, 'whitelist.txt'), 'r', errors='surrogatereplace').read().split('\n')
		#end if
		"""
		filter_filename = 'list_filter1.json'
		filters = []
		filter_names = set()
		fn = os.path.join(hidden_dir, 'filter', filter_filename)
		filter_list = []
		if os.path.exists(fn):
			filter_list = open(fn, 'r').read().split('\n')
		#end if
		for filter_line in filter_list:
			if filter_line.strip() == '': continue
			x = json.loads(filter_line)
			if len(x) != 4:
				print("Warning: {0}".format(filter_filename), repr(filter_line))
				continue
			#end if
			name, field, value, loc = x
			if name in filter_names:
				print("Warning: {0}".format(filter_filename), repr(name), 'in list twice')
				continue
			#end if
			filter_names.add(name)
			filters.append(x)
		#next filter_line
		for name, field, value, directory in filters:
			#print(name, addr, c)
			x = self.filterStore.append((field, value, directory))
		#next field, value, directory
		self.filters = filters
	#end def FiltersWindow(parent)
	
	def showAll(self):
		"""
		Calls self.window.show_all()
		"""
		self.window.show_all()
		if self.parent == None:
			# This might screw stuff up...
			Gtk.main()
		#end if
		return True
	#end def showAll()
	
	def deleteFilters(self):
		"""
		Deletes filters from the filter editor
		"""
		list_store, rows = self.filterTable.get_selection().get_selected_rows()
		to_delete = []
		j = 0
		for row in rows:
			i = row.get_indices()[0]
			filterx = self.filters.pop(i - j)
			self.undo.append(('delete', i - j, filterx))
			to_delete.append(i - j)
			self.filterStore.remove(self.filterStore.iter_nth_child(None, i - j))
			j += 1
		#next row
		self.filtersChanged = True
		return True
	#end def deleteFilters()
	
	def addFilter(self, widget=None):
		"""
		Add a filter.
		"""
		self.parent.clickFilter(None, None, None, None)
		return True
	#end def addFilter()
	
	def editedField(self, cell_renderer, path, data=''):
		"""
		Save edit
		"""
		path_i = int(path)
		self.undo.append(('edit', path_i, self.filters[path_i][:]))
		self.filterStore[path][0] = data
		self.filters[path_i][1] = data
		self.filtersChanged = True
		return True
	#end def editedField(self, cell_renderer, path, data='')
	
	def editedValue(self, cell_renderer, path, data=''):
		"""
		Save edit
		"""
		path_i = int(path)
		self.undo.append(('edit', path_i, self.filters[path_i][:]))
		self.filterStore[path][1] = data
		self.filters[path_i][2] = data
		self.filtersChanged = True
		return True
	#end def editedValue(self, cell_renderer, path, data='')
	
	def editedDirectory(self, cell_renderer, path, data=''):
		"""
		Save edit
		"""
		path_i = int(path)
		self.undo.append(('edit', path_i, self.filters[path_i][:]))
		self.filterStore[path][2] = data
		self.filters[path_i][3] = data
		self.filtersChanged = True
		return True
	#end def editedDirectory(self, cell_renderer, path, data='')
	
	def saveFilters(self):
		"""
		Save once on exit.
		"""
		if self.filtersChanged:
			hidden_dir = mail1.hidden_dir
			open(os.path.join(hidden_dir, 'filter', 'list_filter1.json'), 'w').write('\n'.join([json.dumps(x) for x in self.filters]))
			self.filtersChanged = False
		#end if
		return True
	#end def saveFilters(self)
	
	def showSearchFilters(self):
		# Currently searching contacts is handled by treeview but it might be nice to handle it ourselves.
		# See MaildirWindow.searchDir
		# TODO: Search filters
		pass
	#end def showSearchFilters()
	
	def keyFilters(self, widget, event):
		"""
		Add the Contact to Cc:
		Carbon copy is useful for relationships where someone isn't intended to contribute.
		"""
		#print(repr(event.string), event.keyval)
		#if event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == 102:
		#	# Search Contacts
		#	self.showSearchContacts()
		#	return True
		#end if
		if event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == ord('z'):
			# ctrl-z Undo
			if len(self.undo) > 0:
				last_action = self.undo.pop()
				if last_action[0] == 'delete':
					self.filters.insert(last_action[1], last_action[2])
					self.filterStore.insert(last_action[1], last_action[2])
				elif last_action[0] == 'add':
					self.filters.pop(last_action[1])
					self.filterStore.remove(self.filterStore.iter_nth_child(None, last_action[1]))
				elif last_action[0] == 'edit':
					self.filters[last_action[1]] = last_action[2]
					self.filterStore[last_action[1]] = last_action[2]
				#end if
			#end if
			return True
		#end if
		#print('contacts', repr(event.string), event.keyval)
		if event.string == '\x7f':
			# Delete key
			self.deleteFilters()
			return True
		#end if
		return True
	#end def keyFilters(widget, event)
	
	def clickSave(self, x=None):
		"""
		Save and such
		"""
		self.saveFilters()
		self.close()
		return True
	#end def clickSave(x=None)

	def close(self, x=None):
		self.window.hide()
		return True
	#end def close(self, x)
#end class FiltersWindow

class ImportPubkeyDialog:
	"""
	Allows you to import a public key.
	"""
	def __init__(self, parent):
		"""
		If you give no parent, showAll will call gtk.main(), so don't screw this up.
		"""
		self.parent = parent
		self.filtersChanged = False
		
		self.builder = Gtk.Builder()
		self.builder.add_from_file(os.path.join(mail1.data_dir, 'glade/publickey1.glade'))
		self.window = self.builder.get_object("PublicKeyDialog")
		self.window.set_transient_for(self.parent.window)
		handlers = {"on_PublicKeyDialog_destroy": self.close,
			"ImportButton_clicked": self.clickImport,
			"CancelButton_clicked": self.close,
			}
		self.builder.connect_signals(handlers)
		self.pubkey_view = self.builder.get_object("pubkey_view")
		if mail1.gpg3 == None:
			# FIXME: We can't import a key. Disable things.
			alert("Importing a key isn't possible without gpgme and bindings.")
		#end if
	#end def ImportPubkeyDialog(parent)
	
	def showAll(self):
		"""
		Calls self.window.show_all()
		"""
		self.window.show_all()
		if self.parent == None:
			# This might screw stuff up...
			Gtk.main()
		#end if
		return True
	#end def showAll()

	def clickImport(self, x=None):
		"""
		Import the public key.
		"""
		
		valid_statuses = {1: 'Imported Key', 4:'Good Signature', 36:'Imported'}
		
		mb = self.pubkey_view.get_buffer()
		start = mb.get_start_iter()
		end = mb.get_end_iter()
		pubkey = mb.get_text(start, end, True)
		r = mail1.importPubKey(pubkey)
		if r == None:
			alert("Unable to import key.", parent=self.window)
		#end if
		print('Imported', r)
		message = ""
		if r.considered:
			keys_plural = r.considered > 1 and 'keys' or 'key'
			message += "Considered {0} {1}.\n".format(r.considered, keys_plural)
		#end if
		if r.imported:
			keys_plural = r.imported > 1 and 'keys' or 'key'
			message += "Imported {0} {1}.\n".format(r.imported, keys_plural)
		#end if
		if r.secret_imported:
			keys_plural = r.imported > 1 and 'keys' or 'key'
			message += "Imported {0} secret {1}.\n".format(r.secret_imported, keys_plural)
		#end if
		for imp_status in r.imports:
			# TODO: Actual statuses.
			status = imp_status.status
			if status in valid_statuses:
				status = valid_statuses[status]
			#end if
			message += "Status: {0}, fingerprint: {1}\n".format(status, imp_status.fpr)
		#next imp_status
		if message == '':
			message = "No public key imported. Are you sure that was a public key?"
		#end if
		alert(message, parent=self.window)
		self.close()
		return True
	#end def clickImport(x=None)

	def close(self, x=None):
		self.window.hide()
		return True
	#end def close(x=None)
	
#end class ImportPubkeyDialog


class GenKeyDialog:
	"""
	Allows you to import a public key.
	"""
	def __init__(self, parent):
		"""
		If you give no parent, showAll will call gtk.main(), so don't screw this up.
		"""
		self.parent = parent
		
		self.builder = Gtk.Builder()
		self.builder.add_from_file(os.path.join(mail1.data_dir, 'glade/genkey.glade'))
		self.window = self.builder.get_object("GenKeyDialog")
		self.window.set_transient_for(self.parent.window)
		handlers = {"on_GenKeyDialog_destroy": self.close
			}
		self.builder.connect_signals(handlers)
		self.name = self.builder.get_object("name")
		self.email = self.builder.get_object("email")
		self.comment = self.builder.get_object("comment")
		self.gen_progress = self.builder.get_object("gen_progress")
		if mail1.gpg3 == None:
			# FIXME: We can't generate a key. Disable things.
			alert("Generating a key isn't possible without gpgme and bindings.")
		#end if
	#end def GenKeyDialog(parent)

	def get_name(self):
		return self.name.get_text()
	def get_email(self):
		return self.email.get_text()
	def get_comment(self):
		return self.comment.get_text()
	
	def showAll(self):
		"""
		Calls self.window.show_all()
		"""
		self.window.show_all()
		if self.parent == None:
			# This might screw stuff up...
			Gtk.main()
		#end if
		return True
	#end def showAll()

	def close(self, x=None):
		self.window.hide()
		return True
	#end def close(x=None)
	
#end class GenKeyDialog

class MessageWindow:
	"""
	Allows you to compose and send a message.
	"""
	def __init__(self, parent, message=None, messageDir=None, useParentWindow=True):
		"""
		If you give no parent, showAll will call gtk.main(), so don't screw this up.
		"""
		self.parent = parent
		self.message = message
		self.messageDir = messageDir
		self.useParentWindow = useParentWindow
		self.msg = None
		self.found = None
		self.search_str = None
		
		self.builder = Gtk.Builder()
		self.builder.add_from_file(os.path.join(mail1.data_dir, 'glade/message.glade'))
		self.window = self.builder.get_object("MessageWindow")
		self.message_view = self.builder.get_object("message_view")
		self.message_search = self.builder.get_object("message_search")
		self.message_search.connect('search-changed', self.searchMessage)
		self.message_search.connect('next-match', self.searchMessage)
		self.message_search.connect('previous-match', lambda widget:self.searchMessage(widget, reverse=True))
		self.message_search.connect('stop-search', lambda widget:self.message_search.hide())
		#if self.parent and useParentWindow:
		#	self.window.set_transient_for(self.parent.window)
		#end if
		handlers = {"on_MessageWindow_destroy": self.close,
			#"SaveButton_clicked": self.clickSave,
			#"CancelButton_clicked": self.close
			"on_message_view_key_press_event":self.keyMsgView
			}
		self.builder.connect_signals(handlers)
	#end def FiltersWindow(parent, useParentWindow=True)
	
	def keyMsgView(self, widget, event):
		"""
		Handle key presses when message view is selected
		"""
		#print("key press", event, event.type, event.state, event.string, event.keyval)
		if event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == ord('f'):
			# Search Message Body is Ctrl-f
			self.showSearch()
			return True
		#end if
		if event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == ord('o'):
			# Open is Ctrl-o
			self.parent.openFile()
			return True
		#end if
		if event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == ord('1'):
			# Open in new window is Ctrl-1
			# This should never happen actually but it might because of cli?
			fakeEvent = Gdk.EventButton()
			fakeEvent.type = Gdk.EventType.BUTTON_PRESS
			if self.messageDir == 'drafts':
				self.parent.clickDraft(None, self.message_view, fakeEvent, None)
			#end if
			#print('cd', self.currentDir)
			return True
		#end if
		return False
	#end def keyMsgView(widget, event)
	
	def showSearch(self):
		"""
		Ctrl-f search the message.
		"""
		#print("Show Search Msg")
		self.message_search.show_all()
		self.message_search.grab_focus()
		return True
	#end def showSearch()

	def searchMessage(self, widget, reverse=False):
		"""
		Search the message for a string.
		http://stackoverflow.com/questions/14994684/python-search-for-text-in-a-gtk-textview
		"""
		textbuffer = self.message_view.get_buffer()
		search_str = self.message_search.get_text()
		if self.search_str != search_str:
			start_iter = textbuffer.get_start_iter()
			self.search_str = search_str
		else:
			if reverse:
				start_iter = self.found[0] if self.found else textbuffer.get_start_iter()
			else:
				start_iter = self.found[1] if self.found else textbuffer.get_start_iter()
			#end if
		#end if
		#print('searching from', start_iter.get_offset(), self.found, self.search_str)
		flags = Gtk.TextSearchFlags.CASE_INSENSITIVE
		if reverse:
			self.found = start_iter.backward_search(search_str, flags, None)
		else:
			self.found = start_iter.forward_search(search_str, flags, None)
		#end if
		if self.found:
			match_start,match_end = self.found
			textbuffer.select_range(match_start, match_end)
			self.message_view.scroll_to_iter(self.found[0], 0, False, 0, 0)
		#end if
		return True
	#end def searchMessage(self, reverse=False)

	def showAll(self):
		"""
		Calls self.window.show_all()
		"""
		self.window.show_all()
		# Message search should only take up space when it's called upon.
		self.message_search.hide()
		if self.parent == None or self.useParentWindow == False:
			# This might screw stuff up...
			Gtk.main()
		#end if
		return True
	#end def showAll()

	def close(self, widget=None):
		self.window.hide()
		if self.parent:
			self.parent.windowClose(self)
		#end if
		if self.parent == None or self.useParentWindow == False:
			Gtk.main_quit(widget)
		#end if
		return True
	#end def close(widget)
#end class MessageWindow

class UndoTextBuffer(Gtk.TextBuffer):
	
	def __init__(self):
		Gtk.TextBuffer.__init__(self)
		self.connect('insert-text', self.on_insert_text)
		self.connect('delete-range', self.on_delete_range)
		self.actions = []
		self.redo_actions = []
		self.pos = 0
		self.undoing = False
	#end UndoTextBuffer()
	
	def on_insert_text(self, textbuf, iterator, value, length):
		#print('insert-text', iterator, value, length)
		if self.undoing: return False
		pos = iterator.get_offset()
		if self.actions and self.actions[-1][0] == 'insert' and pos == self.pos:
			self.actions[-1][2] += value
		else:
			self.actions.append(['insert', pos, value])
		#end if
		self.pos = pos + len(value)
		#return Gtk.TextBuffer.insert(self, iterator, value, length)
		return True
	#end def on_insert_text(textbuf, iterator, value, length)
	
	def on_delete_range(self, textbuf, start, end):
		#print('delete-range', start, end)
		if self.undoing: return False
		pos = start.get_offset()
		value = self.get_text(start, end, True)
		#print('odr', pos, len(value), self.pos, pos + len(value), self.pos)
		if self.actions and self.actions[-1][0] == 'delete' and (pos + len(value) == self.pos):
			self.actions[-1][2] = value + self.actions[-1][2]
		else:
			self.actions.append(['delete', pos, value])
		#end if
		self.pos = pos
		#return Gtk.TextBuffer.delete(self, start, end)
		return True
	#end def on_delete_range(textbuf, start, end)
	
	def undo(self):
		"""
		Revert the last change, insert or delete
		"""
		#print('undo', self.actions)
		if not self.actions: return False
		self.undoing = True
		last_action = self.actions.pop()
		self.redo_actions.append(last_action)
		print(last_action)
		if last_action[0] == 'insert':
			# Remove the last text inserted
			start = self.get_iter_at_offset(last_action[1])
			end = self.get_iter_at_offset(last_action[1] + len(last_action[2]))
			self.delete(start, end)
		elif last_action[0] == 'delete':
			start = self.get_iter_at_offset(last_action[1])
			self.insert(start, last_action[2])
		#end if
		self.undoing = False
		return True
	#end def undo()
	
#end class UndoTextBuffer(Gtk.TextBuffer)

def getInstructions(result, hacker=False):
	"""
	Give the user instructions on how to deal with signature failure.
	Expired key, wrong password, no default key.
	"""
	user_instructions = """Unable to sign message. This is because your key is expired. To extend your key expiration date, follow these instructions:
	gpg --edit-key {0}
	expire
	save
	""".replace('\t','')
	hacker_instructions = """Unable to sign message. This is because your key is expired."""
	unknown_instructions = "Unable to sign message. \"{0}\" This is likely because your keyid is expired or not set."
	instructions = unknown_instructions
	if gpg3 and isinstance(result, gpg3.gpg.errors.GPGMEError):
		if gpg3.gpg.errors.gpgme.gpgme_strerror(result.error) == 'Unusable secret key':
			if hacker:
				instructions = hacker_instructions
			else:
				instructions = user_instructions
			#end if
		else:
			instructions = instructions.format(gpg3.gpg.errors.gpgme.gpgme_strerror(result.error))
		#end if
	else:
		instructions = instructions.format('')
	#end if
	return instructions
#end def getInstructions(result, hacker=False)

def joinRecipients(to, cc=''):
	"""
	Use comma to join to and cc.
	"""
	if to and cc:
		return to + ', ' + cc
	#end if
	if to:
		return to
	#end if
	if cc:
		return cc
	#end if
	return ''
#end def joinRecipients(to, cc='')

class ComposeWindow:
	"""
	Allows you to compose and send a message.
	"""
	def __init__(self, parent, mfrom='', to='', subject='', body='', cc='', bcc='', headers=''):
		"""
		If you give no parent, showAll will call gtk.main(), so don't screw this up.
		"""
		self.parent = parent
		self.mfrom = mfrom
		self.to = to
		self.subject = subject
		self.body = body
		self.cc = cc
		self.bcc = bcc
		self.headers = headers
		self.attachments = []
		self.fingerprints = {}
		self.fingerprints_changed = False
		hidden_dir = mail1.hidden_dir
		if os.path.exists(os.path.join(hidden_dir, 'fingerprints.json')):
			if PY2:
				self.fingerprints = json.load(open(os.path.join(hidden_dir, 'fingerprints.json'), 'r'))
			else:
				self.fingerprints = json.load(open(os.path.join(hidden_dir, 'fingerprints.json'), 'r', errors='surrogatereplace'))
			#end if
		#end if
		
		self.builder = Gtk.Builder()
		self.builder.add_from_file(os.path.join(mail1.data_dir, 'glade/compose1.glade'))
		self.window = self.builder.get_object("compose_window")

		handlers = {"on_compose_window_destroy": self.notReady,
			"on_send_clicked": self.send,
			"on_cancel_clicked": self.close,
			"on_attach_clicked": self.attach,
			"on_AddressButton_clicked": self.openAddressBook,
			"on_body_key_press": self.keyBody,
			"on_attach_tree_key_press_event": self.keyAttach,
			"on_to_key_press": self.keyTo,
			"on_cc_key_press": self.keyTo,
			"on_bcc_key_press": self.keyTo
			}
		self.builder.connect_signals(handlers)

		self.bodyView = self.builder.get_object("body")
		self.bodyView.set_buffer(self.newBuffer())
		 
		self.inputFrom = self.builder.get_object("from")
		self.inputTo = self.builder.get_object("to")
		self.inputCc = self.builder.get_object("cc")
		self.inputBcc = self.builder.get_object("bcc")
		self.inputSubject = self.builder.get_object("subject")
		self.inputHeader = self.builder.get_object("headers")
		self.attach_tree = self.builder.get_object("attach_tree")
		self.attach_store = self.builder.get_object("attach_store")
		self.encrypt_button = self.builder.get_object("encrypt_button")
		self.sign_button = self.builder.get_object("sign_button")
		self.openpgp_button = self.builder.get_object("openpgp_button")
		self.message_search = self.builder.get_object("message_search")
		self.message_search.connect('search-changed', self.searchMessage)
		self.message_search.connect('next-match', self.searchMessage)
		self.message_search.connect('previous-match', lambda widget:self.searchMessage(widget, reverse=True))
		self.message_search.connect('stop-search', lambda widget:self.message_search.hide())
		self.inputTo.grab_focus()
		self.inputTo.connect('key-press-event', self.keyTo)
		
		lang = "en_US"
		if 'LANG' in os.environ:
			lang = os.environ['LANG']
			if '.' in lang:
				lang = lang[:lang.index('.')]
			#end if
		#end if
		if GtkSpell:
			self.spellchecker = GtkSpell.Checker()
			self.spellchecker.set_language(lang)
			self.spellchecker.attach(self.bodyView)
			#self.spellchecker.attach(self.inputSubject)
		#end if
		self.found = None
		self.search_str = None
		
		if mfrom == '' and 'from' in self.parent.config['Mail1']:
			mfrom = self.parent.config['Mail1']['from']
		#end if
		self.sign = False
		self.encrypt = False
		self.traditional = False
		self.hacker = False
		self.encrypt = self.parent.config['Mail1'].getboolean('encrypt', False)
		self.sign = self.parent.config['Mail1'].getboolean('sign', False)
		self.traditional = self.parent.config['Mail1'].getboolean('traditional', False)
		self.hacker = self.parent.config['Mail1'].getboolean('hacker', False)
		self.sign_button.set_active(self.sign)
		self.encrypt_button.set_active(self.encrypt)
		self.openpgp_button.set_active(not self.traditional)
		self.inputFrom.set_text(mfrom)

		rendererFilename = Gtk.CellRendererText()
		rendererFilename.ellipsize = Pango.EllipsizeMode.END
		rendererType = Gtk.CellRendererText()
		rendererType.ellipsize = Pango.EllipsizeMode.END
		rendererImage = Gtk.CellRendererPixbuf()
		icon_column = Gtk.TreeViewColumn("", rendererImage, pixbuf=0)
		fn_column = Gtk.TreeViewColumn("Filename", rendererFilename, text=1)
		fn_column.set_resizable(True)
		#fn_column.connect("notify::width", self.updateConfig)
		#fn_column.set_fixed_width(100)
		type_column = Gtk.TreeViewColumn("Type", rendererType, text=2)
		type_column.set_resizable(True)
		#type_column.connect("notify::width", self.updateConfig)
		type_column.set_fixed_width(100)
		self.attach_tree.append_column(icon_column)
		self.attach_tree.append_column(fn_column)
		self.attach_tree.append_column(type_column)
		icon_size = '_16'
		self.att_icon = [GdkPixbuf.Pixbuf.new_from_file(os.path.join(mail1.data_dir, 'images/attach1{0}.png'.format(icon_size))), GdkPixbuf.Pixbuf.new_from_file(os.path.join(mail1.data_dir, 'images/envelope-closed{0}.png'.format(icon_size)))]

		self.ready = True
		self.sent = False
		self.md = None
		
	#end def ComposeWindow(mfrom='', to='', subject='', body='', cc='', bcc='', headers='')

	def newBuffer(self):
		"""
		Never use Gtk.TextBuffer() for bodyView!
		"""
		return UndoTextBuffer()
	#end def newBuffer()
	
	def setHeaders(self, s):
		mb = Gtk.TextBuffer()
		mb.set_text(s)
		self.inputHeader.set_buffer(mb)
		return True
	#end def setHeaders(s)
	
	def getHeaders(self):
		"""
		TextView requires a little more effort than an input box.
		"""
		mb = self.inputHeader.get_buffer()
		start = mb.get_start_iter()
		end = mb.get_end_iter()
		body = mb.get_text(start, end, True)
		return body
	#end def getHeaders(s)
	
	def openAddressBook(self, x=None):
		self.parent.openAddressBook(x)
		return True
	#end def openAddressBook(x=None)
	
	def keyTo(self, widget, event):
		"""
		When a key is pressed in the To, Cc, or Bcc input boxes
		Use widget to decide which box you are in if that matters.
		"""
		#print("key press", event, event.type, event.state, event.string, event.keyval)
		if event.string == '\t':
			# Tab complete
			if not self.parent.config['Mail1'].getboolean('autocomplete'):
				# No tab complete
				return False
			#end if
			if not self.parent.addressBook:
				self.parent.openAddressBook(show=False)
			#end if
			curr_data = widget.get_text()
			tosplit, sto = mail1.canonTo(curr_data)
			print(repr(tosplit), repr(sto))
			if len(tosplit) == 0: return False
			curr_to = tosplit[-1].lower()
			matches = []
			for name, addr, count in self.parent.addressBook.addrs:
				if name.lower().startswith(curr_to) or addr.lower().startswith(curr_to):
					matches.append((name, addr))
					# Exact match means either stop or accept.
					if name.lower() == curr_to or addr.lower() == curr_to:
						break
					#end if
				#end if
			#end if
			print('matched', matches)
			if len(matches) == 1:
				# One match, add it
				name, addr = matches[0]
				fullAddr = addr
				if name != '' and name != addr:
					fullAddr = mail1.nameSanitize(name) + ' <' + addr + '>'
				#end if
				new_text = ', '.join(tosplit[:-1] + [fullAddr])
				widget.set_text(new_text)
				widget.set_position(len(new_text))
				return True
			elif len(matches) > 1:
				# Multiple matches add as much as possible
				similar = curr_to
				pos = len(similar)
				sims = True
				while sims:
					vals = set()
					for name, addr in matches:
						if name.startswith(similar) and len(name) > pos:
							vals.add(name[pos])
						elif addr.startswith(similar) and len(addr) > pos:
							vals.add(addr[pos])
						else:
							sims = False
						#end if
					#next match
					if not sims: break
					if len(vals) == 1:
						similar += list(vals)[0]
					else:
						break
					#end if
					pos += 1
				#loop
				new_text = ', '.join(tosplit[:-1] + [similar])
				widget.set_text(new_text)
				widget.set_position(len(new_text))
				return True
			#end if
			return False
		#end if
		return False
	#end def keyTo(widget, event)
	
	def keyBody(self, widget, event):
		"""
		Handle shortcuts when the body is selected.
		"""
		#print("key press", event, event.type, event.state, event.string, event.keyval)
		if event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == ord('f'):
			# Search Message Body
			self.showSearch()
			return True
		elif event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == ord('z'):
			# ctrl-z undo
			self.bodyView.get_buffer().undo()
			return True
		#end if
		return False
	#end def keyBody(widget, event)
	
	def keyAttach(self, widget, event):
		"""
		When a key is pressed on attachment list (delete an attachment)
		"""
		#print("key press", event, event.type, event.state, event.string, event.keyval)
		if event.string == '\x7f':
			# Delete key
			cursor = widget.get_cursor()
			#print('cursor', cursor, cursor.__class__, cursor[0].__class__, dir(cursor), dir(cursor[0]))
			if cursor == None or cursor.path == None or cursor.path[0] == None:
				return False
			#end if
			pos = cursor.path[0]
			#print('pos', pos)
			self.attachments.pop(pos)
			cur_iter = self.attach_store.get_iter(cursor.path)
			self.attach_store.remove(cur_iter)
			return True
		#end if
		return False
	#end def keyAttach(widget, event)
	
	def showSearch(self):
		#print("Show Search Msg")
		self.message_search.show_all()
		self.message_search.grab_focus()
		return True
	#end def showSearch()

	def searchMessage(self, widget, reverse=False):
		"""
		Search the message for a string.
		http://stackoverflow.com/questions/14994684/python-search-for-text-in-a-gtk-textview
		"""
		textbuffer = self.bodyView.get_buffer()
		search_str = self.message_search.get_text()
		if self.search_str != search_str:
			start_iter = textbuffer.get_start_iter()
			self.search_str = search_str
		else:
			if reverse:
				start_iter = textbuffer.get_iter_at_offset(self.found[0]) if self.found else textbuffer.get_start_iter()
			else:
				start_iter = textbuffer.get_iter_at_offset(self.found[1]) if self.found else textbuffer.get_start_iter()
			#end if
		#end if
		#print('soff', start_iter.get_offset(), textbuffer.get_iter_at_offset)
		#print('searching from', start_iter.get_offset(), self.found, self.search_str)
		flags = Gtk.TextSearchFlags.CASE_INSENSITIVE
		if reverse:
			self.found = start_iter.backward_search(search_str, flags, None)
		else:
			self.found = start_iter.forward_search(search_str, flags, None)
		#end if
		if self.found:
			match_start, match_end = self.found
			textbuffer.select_range(match_start, match_end)
			self.bodyView.scroll_to_iter(self.found[0], 0, False, 0, 0)
			self.found = (match_start.get_offset(), match_end.get_offset())
		#end if
		return True
	#end def searchMessage(self, reverse=False)

	def traditionalSignEncrypt(self, body, recipients, attachments=[]):
		"""
		Input body, string of recipients and list of attachments
		returns body and attachments or None
		"""
		self_key = []
		if self.parent.keyid:
			# Encrypt with your key.
			self_key = [self.parent.keyid]
		#end if
		result = mail1.signEncryptTraditional(body, self_key + self.toFingerprint(strToAddrList(recipients)), self.parent.keyid)
		if result and len(result) == 2:
			result, body = result
			if result == False or body == None:
				if (not body) or (len(body) == 0):
					instructions = getInstructions(result, self.hacker)
					alert("Error: {0} keyid: {1}".format(instructions.format(self.parent.keyid or 'Unknown'), self.parent.keyid or 'Unknown'), parent=self.window)
					return None
				#end if
				email, keys = body
				if len(keys) == 0:
					alert("We don't have a key for recipient {0}, try importing a public key.".format(email), parent=self.window)
					self.parent.importPubkey()
					return None
				#end if
				self.getFingerprint(body)
				# unable to encrypt because of multiple keys or none
				return None
			#end if
			# Never should get a non-ascii.
			body = body.decode('ascii')
			att = []
			for i, attachment in enumerate(attachments):
				result = mail1.signEncryptTraditional(attachment.value, self_key + self.toFingerprint(strToAddrList(recipients)), self.parent.keyid)
				if result and len(result) == 2:
					result, encrypted_attachment = result
					if result == False or encrypted_attachment == None:
						if (not encrypted_attachment) or (len(encrypted_attachment) == 0):
							alert("Error: signature and encryption of attachment failed. {0}".format(result), parent=self.window)
							return None
						#end if
						email, keys = encrypted_attachment
						if len(keys) == 0:
							alert("We don't have a key for recipient {0}, try importing a public key.".format(email), parent=self.window)
							self.parent.importPubkey()
							return None
						#end if
						self.getFingerprint(encrypted_attachment)
						# unable to encrypt because of multiple keys or none
						return None
					#end if
					encrypted_attachment = encrypted_attachment.decode('ascii')
					att.append(mail1.FileAttachment('file_{0}.asc'.format(i), encrypted_attachment, 'application/pgp-encrypted'))
				#end if
			#next attachment
			attachments = att
		else:
			# Error message
			print("Error: couldn't sign 5.", result)
			alert("Error: couldn't sign 5. {0}".format(result), parent=self.window)
			return None
		#end if
		return body, attachments
	#end def traditionalSignEncrypt(body, recipients, attachments=[])
	
	def traditionalSign(self, body, attachments=[]):
		"""
		Input body and list of attachments
		returns body and attachments or None
		"""
		result = mail1.signTraditional(body, self.parent.keyid)
		if result and len(result) == 2:
			result, body = result
			if result == False or body == None:
				# Unable to sign
				instructions = getInstructions(result, self.hacker)
				alert("Error: {0} keyid: {1}".format(instructions.format(self.parent.keyid or 'Unknown'), self.parent.keyid or 'Unknown'), parent=self.window)
				return None
			#end if
			#print('sb', repr(body))
			body = body.decode('utf-8')
			# Sign Attachments
			att = []
			for i, attachment in enumerate(attachments):
				result = mail1.signOpaque(attachment.value, self.parent.keyid)
				if result and len(result) == 2:
					result, signed_attachment = result
					if result == False:
						if (not signed_attachment) or (len(signed_attachment) == 0):
							alert("Error: signature failed 1. {0}".format(result), parent=self.window)
							return None
						#end if
						# unable to encrypt because of multiple keys or none
						return None
					#end if
					att.append(mail1.FileAttachment('file_{0}.asc'.format(i), signed_attachment, 'application/octet-stream'))
				#end if
			#next attachment
			attachments = att
		else:
			# Error message
			print("Error: couldn't sign 3.", result)
			alert("Error: couldn't sign 3.", parent=self.window)
			return None
		#end if
		return body, attachments
	#end def traditionalSign(body, attachments=[])
	
	def traditionalEncrypt(self, body, recipients, attachments=[]):
		"""
		Input body, string of recipients, and list of attachments
		returns body and attachments or None
		"""
		self_key = []
		if self.parent.keyid:
			# Encrypt with your key.
			self_key = [self.parent.keyid]
		#end if
		result = mail1.encryptTraditional(body, self_key + self.toFingerprint(strToAddrList(recipients)))
		if result and len(result) == 2:
			result, body = result
			if result == False or body == None:
				if (not body) or (len(body) == 0):
					alert("Error: encryption failed 1. {0}".format(result), parent=self.window)
					return None
				#end if
				email, keys = body
				if len(keys) == 0:
					alert("We don't have a key for recipient {0}, try importing a public key.".format(email), parent=self.window)
					self.parent.importPubkey()
					return None
				#end if
				self.getFingerprint(body)
				# unable to encrypt because of multiple keys or none
				return None
			#end if
			body = body.decode('ascii')
			# Encrypt Attachments
			att = []
			for i, attachment in enumerate(attachments):
				result = mail1.encryptTraditional(attachment.value, self_key + self.toFingerprint(strToAddrList(recipients)))
				if result and len(result) == 2:
					result, encrypted_attachment = result
					if result == False or encrypted_attachment == None:
						if (not encrypted_attachment) or (len(encrypted_attachment) == 0):
							alert("Error: encrypt attachment failed 1. {0}".format(result), parent=self.window)
							return None
						#end if
						email, keys = encrypted_attachment
						if len(keys) == 0:
							alert("We don't have a key for recipient {0}, try importing a public key.".format(email), parent=self.window)
							self.parent.importPubkey()
							return None
						#end if
						self.getFingerprint(encrypted_attachment)
						# unable to encrypt because of multiple keys or none
						return None
					#end if
					att.append(mail1.FileAttachment('file_{0}.asc'.format(i), encrypted_attachment, 'application/octet-stream'))
				#end if
			#next attachment
			attachments = att
		else:
			# Error message
			print("Error: couldn't encrypt 4.", result)
			alert("Error: couldn't encrypt 4.", parent=self.window)
			return None
		#end if
		return body, attachments
	#end def traditionalEncrypt(body, recipients, attachments=[])

	def openPGPMimeSignEncrypt(self, msg, mime1, recipients, attachments=[]):
		"""
		Input mime, string of recipients and list of attachments
		returns body and attachments or None
		"""
		self_key = []
		if self.parent.keyid:
			# Encrypt with your key.
			self_key = [self.parent.keyid]
		#end if
		#mime1 = mail1.mime_encode(body)
		#print(mime1, repr(mime1))
		result = mail1.signDetatched(mime1, self.parent.keyid)
		if result and len(result) == 2:
			result, body_sig = result
			if result == False or body_sig == None:
				# unable to sign because of default
				instructions = getInstructions(result, self.hacker)
				alert("Error: {0} keyid: {1}".format(instructions.format(self.parent.keyid or 'Unknown'), self.parent.keyid or 'Unknown'), parent=self.window)
				return None
			#end if
			hash_algo = result.signatures[0].hash_algo
			if hash_algo not in mail1.sigalgs:
				# https://tools.ietf.org/html/rfc4880#section-9.4
				print("Error: Unknown hash algorithm in use:", hash_algo)
				alert("Error: Unknown hash algorithm in use: {0}".format(hash_algo), parent=self.window)
				return None
			#end if
			sigalg = mail1.sigalgs[hash_algo]
		else:
			# Error message
			print("Error: couldn't sign 1.", result)
			alert("Error: couldn't sign 1.", parent=self.window)
			return None
		#end if
		result = mail1.encryptTraditional(mail1.mime_encode_sign(msg, body_sig, sigalg), self_key + self.toFingerprint(strToAddrList(recipients)))
		if result and len(result) == 2:
			result, body = result
			if result == False or body_sig == None:
				if (not body) or (len(body) == 0):
					instructions = getInstructions(result, self.hacker)
					alert("Error: {0} keyid: {1}".format(instructions.format(self.parent.keyid or 'Unknown'), self.parent.keyid or 'Unknown'), parent=self.window)
					return None
				#end if
				email, keys = body
				if len(keys) == 0:
					alert("We don't have a key for recipient {0}, try importing a public key.".format(email), parent=self.window)
					self.parent.importPubkey()
					return None
				#end if
				self.getFingerprint(body)
				# unable to encrypt because of multiple keys or none
				return None
			#end if
			body = body.decode('ascii')
		else:
			# Error message
			print("Error: couldn't encrypt sign 6.", result)
			alert("Error: couldn't encrypt sign 6.", parent=self.window)
			return None
		#end if
		body_encrypted = mail1.mime_encode_encrypted(body)
		return body_encrypted, None
	#end def openPGPMimeSignEncrypt(msg, mime1, recipients, attachments=[])
		
	def openPGPMimeSign(self, msg, mime1, attachments=[]):
		"""
		Input body and list of attachments
		returns body and attachments or None
		"""
		#mime1 = mail1.mime_encode(body)
		result = mail1.signDetatched(mime1, self.parent.keyid)
		if result and len(result) == 2:
			result, body_sig = result
			if result == False or body_sig == None:
				# unable to sign because of default
				instructions = getInstructions(result, self.hacker)
				alert("Error: {0} keyid: {1}".format(instructions.format(self.parent.keyid or 'Unknown'), self.parent.keyid or 'Unknown'), parent=self.window)
				return None
			#end if
			hash_algo = result.signatures[0].hash_algo
			if hash_algo not in mail1.sigalgs:
				# https://tools.ietf.org/html/rfc4880#section-9.4
				print("Unknown hash algorithm in use:", hash_algo)
				alert("Unknown hash algorithm in use: {0}".format(hash_algo), parent=self.window)
			#end if
			sigalg = mail1.sigalgs[hash_algo]
		else:
			# Error message
			print("Error: couldn't sign 2.", result)
			alert("Error: couldn't sign 2. {0}".format(result), parent=self.window)
			return None
		#end if
		body = mail1.mime_encode_sign(msg, body_sig, sigalg, True)
		return body, None
	#end def openPGPMimeSign(msg, mime1, attachments=[])
	
	def openPGPMimeEncrypt(self, msg, mime1, recipients, attachments=[]):
		"""
		Input body, string of recipients, and list of attachments
		returns body and attachments or None
		"""
		self_key = []
		if self.parent.keyid:
			# Encrypt with your key.
			self_key = [self.parent.keyid]
		#end if
		#mime1 = mail1.mime_encode(body)
		result = mail1.encryptTraditional(mime1, self_key + self.toFingerprint(strToAddrList(recipients)))
		if result and len(result) == 2:
			result, body = result
			if result == False or body == None:
				if (not body) or (len(body) == 0):
					alert("Error: encryption failed 2. {0}".format(result), parent=self.window)
					return None
				#end if
				email, keys = body
				if len(keys) == 0:
					alert("We don't have a key for recipient {0}, try importing a public key.".format(email), parent=self.window)
					self.parent.importPubkey()
					return None
				#end if
				self.getFingerprint(body)
				# unable to encrypt because of multiple keys or none
				return None
			#end if
			body = body.decode('ascii')
		else:
			# Error message
			print("Error: couldn't encrypt 6.", result)
			alert("Error: couldn't encrypt 6. {0}".format(result), parent=self.window)
			return
		#end if
		body = mail1.mime_encode_encrypted(body)
		return body, None
	#end def openPGPMimeEncrypt(msg, mime1, recipients, attachments=[])
		
	def send(self, x=None):
		"""
		Send an e-mail.
		"""
		#print(x)
		sfrom = self.inputFrom.get_text().strip()
		to = self.inputTo.get_text().strip()
		cc = self.inputCc.get_text().strip()
		bcc = self.inputBcc.get_text().strip()
		subject = self.inputSubject.get_text().strip()
		header = self.getHeaders()
		mb = self.bodyView.get_buffer()
		start = mb.get_start_iter()
		end = mb.get_end_iter()
		body = mb.get_text(start, end, True)
		body_orig = body
		#print('body', body)
		#print(repr(body))
		self.sign = self.sign_button.get_active()
		self.encrypt = self.encrypt_button.get_active()
		self.traditional = not self.openpgp_button.get_active()
		self.parent.config['Mail1']['encrypt'] = str(self.encrypt)
		self.parent.config['Mail1']['sign'] = str(self.sign)
		self.parent.config['Mail1']['traditional'] = str(self.traditional)
		# TODO: This seems suspicious
		if self.parent.config['Mail1'].get('from', '') == '':
			self.parent.config['Mail1']['from'] = str(sfrom)
		#end if
		s = self.parent.config['Mail1'].get('hacker', 'false').lower()
		self.hacker = s in ['true', '1', 'yes', 'on', 'ja']
		if self.hacker:
			if not (self.sign and self.encrypt):
				# Error message popup
				msg_type = 'in plaintext'
				if self.sign: msg_type = 'signed plaintext'
				if self.encrypt: msg_type = 'unsigned encrypted'
				print("Weak modes disabled by default in hacker mode. You must sign and encrypt in hacker mode or click yes to a dialog that defaults to no.")
				response = yes_or_no("Weak modes disabled by default in hacker mode. Do you want to send {0} anyway?".format(msg_type), "Hacker mode override?", parent=self.window)
				if response == Gtk.ResponseType.YES:
					# They made the mistake, not me.
					pass
				else:
					# They decided against this weak mode.
					return True
				#end if
			#end if
		#end if
		
		attachments = []
		for fn, ctype, obj in self.attachments:
			if obj == None:
				attachments.append(mail1.createAttachment(fn, ctype))
			else:
				attachments.append(obj)
			#end if
		#next fn, ctype, obj
		
		if attachments == [] and 'attach' in body:
			verify = yes_or_no("Did you want to attach something?\nNo will send.", parent=self.window)
			if verify == Gtk.ResponseType.YES:
				return True
			#end if
		#end if
		
		if not self.parent.keyid:
			# User has not configured or has decided to not encrypt to self.
			if self.encrypt and not self.hacker:
				# TODO: allow them to avoid seeing this warning without hacker mode.
				# TODO: allow hackers to see this message because sometimes hackers.
				verify = yes_or_no("Warning: sending encrypted without knowning your keyid will result in your message not being encrypted with your key. You won't be able to decrypt. Are you sure you want to send?", parent=self.window)
				if verify != Gtk.ResponseType.YES:
					return True
				#end if
			#end if
		#end if
		
		encrypt_bcc = False
		if self.encrypt and bcc:
			# https://lists.gnupg.org/pipermail/gnupg-devel/2007-January/023538.html
			if self.hacker:
				# Hacker mode follows the advice of Werner Koch because he is a cryptographer.
				alert("Hacker mode absolutely refuses to encrypt and bcc at the same time.")
				return True
			#end if
			# Warn users, allow them to encrypt to bcc.
			result = yes_or_no("Bcc is incompatible with encryption. Do you want to encrypt to bcc? This will result in all recipients being easily able to learn to full contact list which is basically a horrible information disclosure.", parent=self.window)
			if result == Gtk.ResponseType.YES:
				encrypt_bcc = True
			#end if
		#end if
		
		if self.traditional:
			recipients = joinRecipients(to, cc)
			if encrypt_bcc:
				recipients = joinRecipients(recipients, bcc)
			#end if
			if self.sign and self.encrypt:
				result = self.traditionalSignEncrypt(body, recipients, attachments)
				if not result:
					return True
				#end if
				body, attachments = result
			elif self.sign:
				result = self.traditionalSign(body, attachments)
				if not result:
					return True
				#end if
				body, attachments = result
			elif self.encrypt:
				result = self.traditionalEncrypt(body, recipients, attachments)
				if not result:
					return True
				#end if
				body, attachments = result
			#end if
		else:
			# OpenPGP/MIME
			msg = None
			mime1 = None
			recipients = joinRecipients(to, cc)
			if encrypt_bcc:
				recipients = joinRecipients(recipients, bcc)
			#end if
			if self.sign or self.encrypt:
				# crypto means none of the headers should be attached.
				# But the rest should be handled in the same way to get the same result.
				msg = mail1.prepareMessage(sfrom, to, cc, bcc, subject, body, attachments, crypto=True)
				mime1 = msg.as_string()
				mime1 = mime1.replace('\n', '\r\n')
				#print('debug', mime1)
			#end if
			if self.sign and self.encrypt:
				result = self.openPGPMimeSignEncrypt(msg, mime1, recipients, attachments)
				if not result:
					return True
				#end if
				body, attachments = result
			elif self.sign:
				result = self.openPGPMimeSign(msg, mime1, attachments)
				if not result:
					return True
				#end if
				body, attachments = result
			elif self.encrypt:
				result = self.openPGPMimeEncrypt(msg, mime1, recipients, attachments)
				if not result:
					return True
				#end if
				body, attachments = result
			#end if
			# Get rid of attachments since we've signed/encrypted them
			if self.encrypt or self.sign:
				attachments = None
			#end if
		#end if
		#print(sfrom, to, cc, bcc, subject, body)
		server = self.parent.config['Mail1'].get('server')
		server_fingerprint = self.parent.config['Mail1'].get('server_fingerprint')
		server_dest = None
		if isinstance(server, str):
			server_dest = server.split(':')
			if len(server_dest) != 2:
				server_dest = (server, 25)
			else:
				server_dest[1] = int(server_dest[1])
			#end if
		#end if
		result = mail1.send(sfrom, to, cc, bcc, subject, body, attach=attachments, 
			 headers=header, server=server_dest, maildir=self.parent.maildir, 
			 fingerprint=server_fingerprint)
		# result is a string that should start with Success on success. Not otherwise.
		if not result.startswith('Success'):
			# Tell the user.
			print('Error:', result)
			alert(result, "Mail Send Failed", parent=self.window)
			return True
		#end if
		self.sent = True
		self.md = hashlib.sha512(body_orig.encode('utf-8')).digest()
		if not self.parent.addressBook:
			self.parent.openAddressBook(show=False)
		#end if
		# TODO: Cc and Bcc
		self.parent.addressBook.addMessage(to)
		self.close()
		return True
	#end def send(self, x=None)
	
	def toFingerprint(self, addresses):
		"""
		Convert a list of e-mail addresses to a list of fingerprints using the fingerprint map.
		"""
		output = []
		for address in addresses:
			if address in self.fingerprints:
				output.append(self.fingerprints[address])
			else:
				output.append(address)
			#end if
		#next address
		return output
	#end def toFingerprint(addresses)
	
	def getFingerprint(self, keys):
		"""
		Given multiple fingerprints, choose the correct one from a dialog box
		"""
		builder = Gtk.Builder()
		builder.add_from_file(os.path.join(mail1.data_dir, 'glade/fingerprint_dialog1.glade'))
		dialog = builder.get_object("fingerprint_dialog")
		#alert(str(dir(dialog)), 'Fingerprint Dialog', self.window)
		#dialog.set_parent(self.window)
		dialog.set_transient_for(self.window)
		fingerprints = builder.get_object("fingerprints")
		fingerprint_store = builder.get_object("fingerprint_store")
		#handlers = {"ok_clicked": dialog.hide, "cancel_clicked": dialog.hide}
		rendererUID = Gtk.CellRendererText()
		rendererUID.ellipsize = Pango.EllipsizeMode.END
		rendererFingerprint = Gtk.CellRendererText()
		rendererFingerprint.ellipsize = Pango.EllipsizeMode.END
		rendererCreated = Gtk.CellRendererText()
		rendererCreated.ellipsize = Pango.EllipsizeMode.END
		rendererExpired = Gtk.CellRendererText()
		rendererExpired.ellipsize = Pango.EllipsizeMode.END
		uid_column = Gtk.TreeViewColumn("UID", rendererUID, text=0)
		uid_column.set_resizable(True)
		fingerprint_column = Gtk.TreeViewColumn("Fingerprint", rendererFingerprint, text=1)
		fingerprint_column.set_resizable(True)
		created_column = Gtk.TreeViewColumn("Created", rendererCreated, text=2)
		created_column.set_resizable(True)
		expired_column = Gtk.TreeViewColumn("Expiry", rendererExpired, text=3)
		expired_column.set_resizable(True)
		#subject_column.connect("notify::width", self.updateConfig)
		#subject_column.set_fixed_width(int(self.config['Mail1GUI']['subject']))
		fingerprints.append_column(uid_column)
		fingerprints.append_column(fingerprint_column)
		fingerprints.append_column(created_column)
		fingerprints.append_column(expired_column)
		# Value we get is a tuple of email address and list of gpgme keys.
		email, keys = keys
		
		for key in keys:
			# FIXME: All subkeys need approx same timestamp and expiry date or..
			name = ', '.join(['{0} <{1}>'.format(uid.name, uid.email) for uid in key.uids])
			# FIXME: fix and test in python 2.7
			if (not PY2) and not name.isprintable():
				print('name bug', repr(name))
				name2 = mail1.guessEncoding(name.encode('utf-8','surrogateescape'), 'cp1252')
				if (not PY2) and name2.isprintable():
					name = name2
				else:
					name = name.encode('utf-8','surrogateescape').decode('utf-8','replace')
				#end if
			#end if
			fingerprint_store.append([name, key.fpr, time.strftime('%a %d %b %Y %T %Z', time.localtime(float(key.subkeys[0].timestamp))), key.subkeys[0].expires <= 0 and "No expiry" or time.strftime('%a %d %b %Y %T %Z', time.localtime(float(key.subkeys[0].expires)))])
		#next key
		response = dialog.run()
		if response == Gtk.ResponseType.OK:
			# Save the fingerprint/address to a map
			list_store, rows = fingerprints.get_selection().get_selected_rows()
			if len(rows) == 1:
				i = rows[0].get_indices()[0]
				#print('row indices:', i)
				try:
					self.fingerprints[email] = keys[i].fpr
				except Exception as e:
					print(e)
				#end try
				self.fingerprints_changed = True
			else:
				print('rows', rows)
			#end if
		#end if
		#print('fingerprints', self.fingerprints)
		dialog.destroy()
		return True
	#end def getFingerprint(key)
	
	def close(self, x=None):
		#self.saveDraft()
		self.window.hide()
		return True
	#end def close(self, x)
	
	def attach(self, x=None):
		"""
		Attach a file.
		"""
		# Stock file chooser.
		dialog = Gtk.FileChooserDialog("Attach File", self.window,
			Gtk.FileChooserAction.OPEN,
			(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
			Gtk.STOCK_OPEN, Gtk.ResponseType.OK))
		#if curr_filename:
		#	dialog.set_current_name(curr_filename)
		#end if
		dialog.set_do_overwrite_confirmation(True)
		response = dialog.run()
		if response == Gtk.ResponseType.OK:
			#print("Save clicked")
			fn = dialog.get_filename()
			#print("File selected: " + fn)
			#first_byte = open(fn, 'rb').read(1)
			# This is backwards and stupid, but... accidents happen.
			mime_ext = {'image/jpeg':'jpg', 'image/png':'png', 
		'image/gif':'gif', 'application/pdf':'pdf', 
		'application/zip':'zip', 'text/plain':'txt', 
		'application/pgp-encrypted':'txt', 'text/x-python': 'py', 
		'text/x-csrc': 'c', 'text/x-patch': 'patch', 'text/x-csrc': 'c', 
		'text/html':'html', 'text/x-c++src':'cxx', 'text/x-diff':'diff', 
		'audio/x-WAV':'wav', 'audio/ogg':'ogg', 'audio/flac':'flac', 
		'application/x-bittorrent':'torrent', 'application/x-rar-compressed':'rar',
		'video/mp4':'mp4', 'audio/mpeg':'mp3', 'audio/midi':'mid', 
		'video/x-msvideo':'avi', 'application/pgp-signature':'asc'}
			mime = 'application/octet-stream'
			ext = fn.split('/')[-1].split('.')[-1]
			if ext in mime_ext.values():
				for k, v in mime_ext.items():
					if ext == v:
						mime = k
						break
					#end if
				#next k, v
			#end if
			self.attachments.append([fn, mime, None])
			self.attach_store.append([self.att_icon[0], fn, mime])
			self.attach_tree.show_all()
		elif response == Gtk.ResponseType.CANCEL:
			#print("Cancel clicked")
			pass
		#end if
		dialog.destroy()
		return True
	#end def attach(self, x)
	
	def showAll(self):
		"""
		Calls self.window.show_all()
		"""
		self.window.show_all()
		self.message_search.hide()
		if len(self.attach_store) == 0:
			self.attach_tree.hide()
		#end if
		if self.parent == None:
			# This might screw stuff up...
			Gtk.main()
		#end if
		return True
	#end def showAll()
	
	def present(self):
		self.window.present()
		return True
	#end def present()
	
	def saveDraft(self):
		"""
		Save the data so we can move on.
		Triggered by destroy.
		"""
		# Save fingerprints to json file.
		if self.fingerprints_changed:
			print('saving fingerprints')
			hidden_dir = mail1.hidden_dir
			json.dump(self.fingerprints, open(os.path.join(hidden_dir, 'fingerprints.json'), 'w', errors='surrogatereplace'))
			self.fingerprints_changed = False
		#end if

		sfrom = self.inputFrom.get_text()
		to = self.inputTo.get_text()
		cc = self.inputCc.get_text()
		bcc = self.inputBcc.get_text()
		subject = self.inputSubject.get_text()
		header = self.getHeaders()
		mb = self.bodyView.get_buffer()
		start = mb.get_start_iter()
		end = mb.get_end_iter()
		body = mb.get_text(start, end, True)
		body_orig = body
		# Don't save any blank messages.
		if body_orig == '' and to == '' and subject == '' and cc == '' and bcc == '' and header == '':
			return False
		#end if

		if self.sent:
			md = hashlib.sha512(body.encode('utf-8')).digest()
			if md == self.md:
				# They sent this body, maybe to someone else.
				return False
			else:
				# They haven't sent the latest version.
				self.sent = False
			#end if
		#end if

		#print(repr(body))
		#print(sfrom, to, cc, bcc, subject, body)
		attachments = []
		for fn, ctype, obj in self.attachments:
			if obj == None:
				attachments.append(mail1.createAttachment(fn, ctype))
			else:
				attachments.append(obj)
			#end if
		#next fn, ctype, obj
		msg = mail1.prepareMessage(sfrom, to, cc, bcc, subject, body, attachments, header)
		path = os.path.join(self.parent.maildir, 'drafts', 'cur')
		if not os.path.exists(path):
			try:
				os.makedirs(path)
			except (NotADirectoryError, FileExistsError):
				path = os.path.join(self.parent.maildir, 'drafts2', 'new')
				if not os.path.exists(path):
					os.makedirs(path)
				#end if
			#end try
		#end if

		filename =  '{0:d}.{1:d}.{2}'.format(int(time.time()), os.getpid(), hex(random.randint(0, 1000000))[2:])
		open(os.path.join(path, filename), 'w').write(msg.as_string())
		##self.sent = True
		self.md = hashlib.sha512(body_orig.encode('utf-8')).digest()
		print("Saved draft {0}".format(os.path.join(path, filename)))
		return True
	#end def saveDraft()
	
	def notReady(self, x=None):
		self.saveDraft()
		self.ready = False
		return True
	#end def notReady(x=None)
	
#end class ComposeWindow

class SearchDialog(Gtk.Dialog):

	def __init__(self, parent):
		# TODO: This should be done in glade.
		Gtk.Dialog.__init__(self, "Search", parent,
			Gtk.DialogFlags.MODAL, buttons=(
			Gtk.STOCK_FIND, Gtk.ResponseType.OK,
			Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL))

		box = self.get_content_area()

		label = Gtk.Label("Insert text you want to search for:")
		box.add(label)

		self.entry = Gtk.Entry()
		self.entry.connect('key-press-event', self.keyEntry)
		self.body_search = Gtk.CheckButton()
		self.body_search.set_label('Search Body')
		self.case_sensitive = Gtk.CheckButton()
		self.case_sensitive.set_label('Case Sensitive')
		box.add(self.entry)
		box.add(self.body_search)
		box.add(self.case_sensitive)

		self.show_all()
	#end def SearchDialog(parent)
	
	def keyEntry(self, widget, event):
		#print('ke', repr(event.string), event.keyval)
		if event.string == '\r':
			self.response(Gtk.ResponseType.OK)
			return True
		#end if
		return False
	#end def keyEntry(widget, event)
	
#end class SearchDialog

def searchRow(search_term, row, message=None, path=None, case_sensitive=False, body_search=False):
	"""
	Utility function to search a row and optionally body
	message, path are only used if body_search is True.
	"""
	caseFunction = lambda x:x
	if not case_sensitive:
		caseFunction = lambda x:x.lower()
	#end if
	found_body = False
	found_headers = search_term in caseFunction(row[1]) or search_term in caseFunction(row[2])
	if body_search and not found_headers:
		# Slowly search each body only if it isn't found in the headers
		if isinstance(message, list):
			message = message[0]
		#end if
		messagePath = os.path.join(path, message)
		found_body = mail1.searchBody(search_term, messagePath, case_sensitive)
	#end if
	return found_headers or found_body
#end def searchRow(search_term, row, message=None, path=None, case_sensitive=False, body_search=False)

def openMessage(parent, useParentWindow=True):
	"""
	Open the parent's current Message in a different window
	"""
	message = parent.message
	message_window = MessageWindow(parent, message, parent.messageDir, useParentWindow)
	message_buffer = Gtk.TextBuffer()
	add = ''
	use_part = None
	parent.drawMsg(message_buffer, parent.maildir, parent.messageDir, message, add, use_part)
	# FIXME: Race condition.
	message_window.msg = parent.msg
	message_window.message_view.set_buffer(message_buffer)
	message_window.message_view.scroll_to_iter(message_buffer.get_start_iter(), 0.0, True, 0.0, 0.0)
	message_window.window.set_title(decodeHeaderGtk(parent.msg.get('subject'), True, removeNewline=True))
	# showAll on single window calls Gtk.main() and we aren't ready yet.
	if useParentWindow:
		message_window.showAll()
	#end if
	parent.message_wins.append(message_window)
	return message_window
#end def openMessage(parent, useParentWindow=True)

class MaildirWindow:
	"""
	On the left a list of directories.
	Top right, A list of messages in the current directory.
	Bottom right, The current message. Simple, no?
	"""
	# Types that won't cause us to download to disk.
	legit_types = ['multipart/mixed', 'multipart/signed', 'text/plain', 
		'application/pgp-signature', 'text/html', 'multipart/alternative', 'multipart/related', 'application/pkcs7-signature']
	
	def __init__(self, maildir, suppressDrawDir=False):
		self.running = False
		self.updateEvent = None
		self.auto_added = 0
		self.messages = []
		self.maildir = maildir
		self.message = None
		self.messageDir = None
		self.message_wins = []
		self.message_wins_exit = False
		self.add = ''
		self.composer = None
		self.addressBook = None
		self.showHeader = False
		self.extra_headers = []
		self.msg = None
		self.attachments = {}
		self.undoEvents = []
		self.cutBuffer = []
		self.copyBuffer = []
		self.to_move = []
		self.config = mail1.getConfig()
		if 'Mail1GUI' not in self.config:
			self.config['Mail1GUI'] = {'width': 800, 'height': 600, 'subject':384, 'from':144, 'date':200}
		#end if
		if 'Mail1' not in self.config:
			self.config['Mail1'] = {'inbox': 'inbox'}
		#end if
		self.currentDir = self.config['Mail1'].get('inbox', 'inbox')
		self.keyid = self.config['Mail1'].get('keyid')

		hidden_dir = mail1.hidden_dir
		if not os.path.exists(hidden_dir):
			os.mkdir(hidden_dir)
			if not os.path.exists(hidden_dir):
				raise Exception("Unable to create necessary hidden directory.")
			#end if
		#end if
		
		try:
			self.whitelist = open(hidden_dir + 'whitelist.txt','r').read().split('\n')
		except FileNotFoundError:
			self.whitelist = []
		#end try
		try:
			self.blacklist = open(hidden_dir + 'blacklist.txt','r').read().split('\n')
		except FileNotFoundError:
			self.blacklist = []
		#end try
		try:
			self.blacklist_ip = open(hidden_dir + 'blacklist_ip.txt','r').read().split('\n')
		except FileNotFoundError:
			self.blacklist_ip = []
		#end try
		
		self.drawDirRunning = False
		
		self.builder = Gtk.Builder()
		self.builder.add_from_file(os.path.join(mail1.data_dir, 'glade/mail1glade2.glade'))
		self.window = self.builder.get_object("window")
		#print(dict(self.config['Mail1GUI']))
		#w, h = self.window.get_preferred_size()
		#print('width height', w, h)
		self.window.resize(int(self.config['Mail1GUI']['width']), int(self.config['Mail1GUI']['height']))
		self.dirs_box = self.builder.get_object("dirs_box")
		self.dirs_store = self.builder.get_object("dirs_store")
		self.dirs_tree = self.builder.get_object("dirs_tree")
		self.als = []
		targets = Gtk.TargetList.new([])
		targets.add_text_targets(1)

		rendererName = Gtk.CellRendererText()
		rendererName.ellipsize = Pango.EllipsizeMode.END
		name_column = Gtk.TreeViewColumn("Name", rendererName, text=0, background=4)
		name_column.set_resizable(True)
		# TODO: add to config
		#name_column.connect("notify::width", self.updateConfig)
		name_column.set_fixed_width(self.config['Mail1GUI'].getint('dirname', 100))
		self.dirs_tree.append_column(name_column)
		self.dirChangedConnection = None
		self.drawIndex(maildir, self.dirs_box, self.dirs_store, targets)
		self.dirs_tree.enable_model_drag_dest([Gtk.TargetEntry.new('text/plain', 0, 0)], DRAG_ACTION)
		self.dirs_tree.drag_dest_set(Gtk.DestDefaults.ALL, [], DRAG_ACTION)
		self.dirs_tree.connect('drag-data-received', self.dragData)
		self.dirs_tree.connect('drag-drop', self.dragDrop)
		self.dirs_tree.drag_dest_set_target_list(targets)
		
		icon_size = '_16'
		self.unreadPixBuf = GdkPixbuf.Pixbuf.new_from_file(os.path.join(mail1.data_dir, 'images/envelope-closed{0}.png'.format(icon_size)))
		self.readPixBuf   = GdkPixbuf.Pixbuf.new_from_file(os.path.join(mail1.data_dir, 'images/envelope-open{0}.png'.format(icon_size)))
		self.spamPixBuf   = GdkPixbuf.Pixbuf.new_from_file(os.path.join(mail1.data_dir, 'images/x{0}.png'.format(icon_size)))
		self.twoPixBuf    = GdkPixbuf.Pixbuf.new_from_file(os.path.join(mail1.data_dir, 'images/question-mark2{0}.png'.format(icon_size)))
		self.threePixBuf  = GdkPixbuf.Pixbuf.new_from_file(os.path.join(mail1.data_dir, 'images/question-mark3{0}.png'.format(icon_size)))
		self.fourPixBuf   = GdkPixbuf.Pixbuf.new_from_file(os.path.join(mail1.data_dir, 'images/question-mark4{0}.png'.format(icon_size)))
		self.fivePixBuf   = GdkPixbuf.Pixbuf.new_from_file(os.path.join(mail1.data_dir, 'images/question-mark5{0}.png'.format(icon_size)))
		self.attachPixBuf = GdkPixbuf.Pixbuf.new_from_file(os.path.join(mail1.data_dir, 'images/attach1{0}.png'.format(icon_size)))
		self.twoAttachPixBuf    = GdkPixbuf.Pixbuf.new_from_file(os.path.join(mail1.data_dir, 'images/attach2{0}.png'.format(icon_size)))
		self.threeAttachPixBuf  = GdkPixbuf.Pixbuf.new_from_file(os.path.join(mail1.data_dir, 'images/attach3{0}.png'.format(icon_size)))
		self.fourAttachPixBuf   = GdkPixbuf.Pixbuf.new_from_file(os.path.join(mail1.data_dir, 'images/attach4{0}.png'.format(icon_size)))
		self.fiveAttachPixBuf   = GdkPixbuf.Pixbuf.new_from_file(os.path.join(mail1.data_dir, 'images/attach5{0}.png'.format(icon_size)))
		self.attachSpamPixBuf = GdkPixbuf.Pixbuf.new_from_file(os.path.join(mail1.data_dir, 'images/attachspam{0}.png'.format(icon_size)))

		self.msgs_win = self.builder.get_object("msgs_win")
		self.msg_store = self.builder.get_object("msg_store") # Gtk.TreeStore(GdkPixbuf.Pixbuf, str, str, str) 
		self.msg_table = self.builder.get_object("msg_table")
		#self.msg_table.set_model(self.msg_store)
		#print(self.window)
		# Multiple e-mails selected for moving, copying, archiving, and spam removal.
		select = self.msg_table.get_selection()
		select.set_mode(Gtk.SelectionMode.MULTIPLE)

		rendererSubject = Gtk.CellRendererText()
		rendererSubject.ellipsize = Pango.EllipsizeMode.END
		rendererFrom = Gtk.CellRendererText()
		rendererFrom.ellipsize = Pango.EllipsizeMode.END
		rendererDate = Gtk.CellRendererText()
		rendererDate.ellipsize = Pango.EllipsizeMode.END
		rendererImage = Gtk.CellRendererPixbuf()
		spam_column = Gtk.TreeViewColumn("", rendererImage, pixbuf=0, cell_background=4)
		subject_column = Gtk.TreeViewColumn("Subject", rendererSubject, text=1, background=4)
		subject_column.set_resizable(True)
		subject_column.connect("notify::width", self.updateConfig)
		subject_column.set_fixed_width(int(self.config['Mail1GUI']['subject']))
		from_column = Gtk.TreeViewColumn("From", rendererFrom, text=2, background=4)
		from_column.set_resizable(True)
		from_column.connect("notify::width", self.updateConfig)
		from_column.set_fixed_width(int(self.config['Mail1GUI']['from']))
		date_column = Gtk.TreeViewColumn("Date", rendererDate, text=3, background=4)
		date_column.set_resizable(True)
		date_column.connect("notify::width", self.updateConfig)
		date_column.set_fixed_width(int(self.config['Mail1GUI']['date']))

		self.msg_table.append_column(spam_column)
		self.msg_table.append_column(subject_column)
		self.msg_table.append_column(from_column)
		self.msg_table.append_column(date_column)
		self.index = None

		if not suppressDrawDir:
			self.drawDir(maildir, currentDir = self.currentDir, msg_store = self.msg_store)
		#end if

		self.cursorChangedConnection = self.msg_table.connect('cursor-changed', self.clickMsg)
		self.msg_table.enable_model_drag_source(Gdk.ModifierType.BUTTON1_MASK, [], DRAG_ACTION)
		self.msg_table.drag_source_set_target_list(targets)

		self.message_view = self.builder.get_object("message_view")
		self.message_view.connect('key-press-event', self.keyMsgView)

		self.message_search = self.builder.get_object("message_search")
		self.message_search.connect('search-changed', self.searchMessage)
		self.message_search.connect('next-match', self.searchMessage)
		self.message_search.connect('previous-match', lambda widget:self.searchMessage(widget, reverse=True))
		self.message_search.connect('stop-search', lambda widget:self.message_search.hide())
		
		self.found = None
		self.search_str = None
		self.filters_window = None
		self.import_pubkey = None
		self.readyToGroup = None
		self.updateDirLast = time.time()

		handlers = {"on_window_destroy": self.gtkExit,
			"on_compose_button_clicked": self.clickCompose,
			"on_window_configure_event": self.updateConfig,
			"on_quit": self.gtkExit,
			"on_open": self.openFile,
			"on_configure": self.clickConfig,
			"on_filters": self.clickFilters,
			"on_address_book": self.openAddressBook,
			"on_find": self.showSearchDir,
			"on_search_message": self.showSearch,
			"on_import_pubkey": self.importPubkey,
			"on_generate_key": self.genKeyPair,
			"on_msg_table_drag_begin": self.dragStart,
			"on_msg_table_drag_data_get": self.dragDataGet,
			"on_msg_table_key_press_event": self.keyMsgTable,
			"on_dirs_tree_key_press_event": self.keyMsgTable,
			"on_msg_table_click": self.clickMsgTable,
			"on_group_by_sender": self.groupBySender,
			"on_add_dir_click": self.addDirectory,
			}
		self.builder.connect_signals(handlers)
		
		self.dirs_tree.grab_focus()
	#end def MaildirWindow(maildir)
	
	def clickConfig(self, x=None):
		config_window = ConfigWindow(self.window)
		config_window.load_config(self.config)
		response = config_window.run()
		if response == Gtk.ResponseType.OK:
			#print("OK clicked")
			self.config = config_window.get_config()
			self.saveConfig()
			self.keyid = self.config['Mail1'].get('keyid')
			if not self.config['Mail1GUI'].getboolean('update_loop', False):
				# Turn off the auto update.
				if self.updateEvent != None:
					self.running = False
					GObject.source_remove(self.updateEvent)
					self.updateEvent = None
					self.running = True
				#end if
			else:
				# Turn on the auto update if it isn't running.
				if self.updateEvent == None:
					self.startUpdateLoop()
				#end if
			#end if
		elif response == Gtk.ResponseType.CANCEL:
			#print("Cancel clicked")
			pass
		#end if
		config_window.destroy()
		return True
	#end def clickConfig(x=None)
	
	def clickFilters(self, x=None):
		self.filters_window = FiltersWindow(self)
		self.filters_window.showAll()
		"""
		filters_window = FiltersWindow(self)
		response = filters_window.run()
		if response == Gtk.ResponseType.OK:
			#print("OK clicked")
			pass
		elif response == Gtk.ResponseType.CANCEL:
			#print("Cancel clicked")
			pass
		#end if
		filters_window.destroy()
		"""
		return True
	#end def clickFilters(x=None)
	
	def openMessage(self, x=None):
		"""
		Open a message in a new window
		"""
		openMessage(self)
		return True
	#end def openMessage(x=None)
	
	def importPubkey(self, x=None):
		"""
		Show Import public key dialog
		"""
		self.import_pubkey = ImportPubkeyDialog(self)
		self.import_pubkey.showAll()
		return True
	#end def importPubkey(x=None)

	def genKeyPair(self, x=None):
		"""
		Generate a PGP key pair for use in encryption and signing.
		"""
		#print("Show Search Dir")
		dialog = GenKeyDialog(self)
		response = dialog.window.run()
		if response == Gtk.ResponseType.OK:
			print("Generate clicked")
			name = dialog.get_name()
			email = dialog.get_email()
			comment = dialog.get_comment()
			result = gpg3.GPG_genkey(name, email, comment)
			if isinstance(result, gpg3.gpg.errors.GPGMEError):
				print("Error:", result)
				return
			#end if
			print([(x, getattr(result, x)) for x in dir(result)])
			if hasattr(result, 'fpr') and result.fpr:
				print("yay", result.fpr)
				self.config['Mail1']['keyid'] = result.fpr
				self.keyid = result.fpr
			#end if
		elif response == Gtk.ResponseType.CANCEL:
			print("Cancel clicked")
		#end if
		dialog.window.destroy()
		return True
	#end def genKeyPair(x=None)
	
	def openFile(self, x=None):
		"""
		Open any file as a message. This could get ugly.
		"""
		# Stock file chooser.
		dialog = Gtk.FileChooserDialog("Open File", self.window,
			Gtk.FileChooserAction.OPEN,
			(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
			Gtk.STOCK_OPEN, Gtk.ResponseType.OK))
		#if curr_filename:
		#	dialog.set_current_name(curr_filename)
		#end if
		response = dialog.run()
		if response == Gtk.ResponseType.OK:
			print("Open clicked")
			fn = dialog.get_filename()
			print("File selected: " + fn)
			self.showMsg(fn)
		elif response == Gtk.ResponseType.CANCEL:
			print("Cancel clicked")
		#end if
		dialog.destroy()
		return True
	#end def openFile(x=None)
	
	def openAddressBook(self, x=None, show=True):
		"""
		Show an address book. Optionally show it.
		"""
		if self.addressBook:
			if show:
				self.addressBook.showAll()
			#end if
			return True
		#end if
		self.addressBook = AddressBookWindow(self)
		if show:
			self.addressBook.showAll()
		#end if
		return True
	#end def openAddressBook(x=None)
	
	def dragStart(self, widget, context, user_data=None):
		"""
		This function supports drag and drop of files to folders.
		"""
		#print("Drag start", widget, context, user_data)
		#Gdk.drag_status(context, DRAG_ACTION, 0)
		return True
	
	def dragDrop(self, widget, context, x,y, time, user_data=None):
		"""
		This function supports drag and drop of files to folders.
		"""
		#print("Drag drop", widget, context, x, y, time, user_data)
		#print("clt", context.list_targets())
		#print('dgd', widget.drag_get_data(context, context.list_targets()[-1], time))
		return True
		
	def dragData(self, widget, drag_context, x,y, data,info, time, user_data=None):
		#print("Drag data", widget, drag_context, x,y, data,info, time, user_data)
		drop_info = widget.get_dest_row_at_pos(x, y)
		#print('row', drop_info, drop_info[0], drop_info[0][0])
		parent_dir = self.als[drop_info[0][0]]
		directory = parent_dir
		#print('parent', repr(directory))
		if isinstance(parent_dir, list):
			directory = parent_dir[0]
			if len(drop_info[0]) > 1:
				child_dir = parent_dir[drop_info[0][1] + 1]
				directory = child_dir
			#end if
		#end if
		#print('dir', directory)
		#return True
		#to_move_json = data.get_text()
		#print('data:', to_move_json)
		#to_move = json.loads(to_move_json)
		to_move = self.to_move
		# Drag a link from a browser.
		if to_move == None: return False
		for i in to_move:
			if isinstance(i, list):
				message1 = i[0]
				message = self.messages[message1][1][i[1]]
			else:
				message = self.messages[i]
			if isinstance(message, list):
				message = message[0]
			#end if
			messagePath = os.path.join(self.maildir, self.currentDir, 'cur', message)
			#end if
			destPath = os.path.join(self.maildir, directory, 'cur', message)
			print('Moving', messagePath, destPath)
			try:
				os.rename(messagePath, destPath)
			except FileNotFoundError:
				print("Warning: dragData is getting called twice. Once moved, you need not move again.", messagePath, destPath)
			#end try
		#next to_move
		return True
		
	def dragDataGet(self, widget, drag_context, data, info, time, user_data=None):
		#print("Drag data get", widget, drag_context, data, info, time, user_data)
		#print('info:', info)
		list_store, rows = widget.get_selection().get_selected_rows()
		#print("Multi", dir(list_store), (rows))
		to_move = []
		for row in rows:
			v = row.get_indices()
			if len(v) > 1:
				to_move.append(v)
			else:
				to_move.append(v[0])
			#end if
		#next row
		# we can reverse repr, so this is a reasonable solution
		#to_move_json = json.dumps(to_move)
		#data.set_text(to_move_json, len(to_move_json))
		data.set_text('move', 4)
		self.to_move = to_move
		return True
		
	def keyMsgTable(self, widget, event):
		"""
		When a message or multiple messages are selected and you press a key.
		"""
		print("key press", event, event.type, event.state, repr(event.string), event.keyval)
		if event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == ord('f'):
			# Search Messages is Ctrl-f
			self.showSearchDir()
			return True
		#end if
		if event.string == 's':
			# Spam is s
			self.clickSpamMulti()
			return True
		#end if
		if event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == ord('r'):
			# Reply is Ctrl-r
			fakeEvent = Gdk.EventButton()
			fakeEvent.type = Gdk.EventType.BUTTON_PRESS
			self.clickReply(None, widget, fakeEvent, None)
			return True
		#end if
		if event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == ord('o'):
			# Open is Ctrl-o
			self.openFile()
			return True
		#end if
		if event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == ord('1'):
			# Open in new window is Ctrl-1
			fakeEvent = Gdk.EventButton()
			fakeEvent.type = Gdk.EventType.BUTTON_PRESS
			if self.currentDir == 'drafts':
				self.clickDraft(None, None, fakeEvent, None)
			else:
				#self.clickReply(None, widget, fakeEvent, None)
				self.openMessage()
			#end if
			#print('cd', self.currentDir)
			return True
		#end if
		if event.string == 'r':
			# Refresh the current view.
			self.msg_table.disconnect(self.cursorChangedConnection)
			self.msg_store.clear()
			self.auto_added = 0
			self.cursorChangedConnection = self.msg_table.connect('cursor-changed', self.clickMsg)
			self.drawDir(self.maildir, currentDir = self.currentDir, msg_store = self.msg_store)
			return True
		#end if
		if event.string == 'u':
			# It's legit, unblock
			self.unblockMulti()
			return True
		#end if
		if event.string == '\x7f':
			# Delete key
			self.deleteMulti()
			return True
		#end if
		#print(event.keyval, repr(event.string), event.state)
		if event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == ord('z'):
			# ctrl-z undo
			return self.undo()
		#end if
		if event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == ord('x'):
			# ctrl-x cut
			self.cutMulti()
			return True
		#end if
		if event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == ord('c'):
			# ctrl-c copy
			self.copyMulti()
			return True
		#end if
		if event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == ord('v'):
			# ctrl-v paste
			self.paste()
			return True
		#end if
		return False
	#end def keyMsgTable(widget, event)
	
	def clickMsgTable(self, widget, event):
		"""
		Handle click and double-click of the list of messages.
		"""
		if event.type == Gdk.EventType._2BUTTON_PRESS:
			# Double-click
			fakeEvent = Gdk.EventButton()
			fakeEvent.type = Gdk.EventType.BUTTON_PRESS
			if self.currentDir == 'drafts':
				self.clickDraft(None, None, fakeEvent, None)
			else:
				#self.clickReply(None, None, fakeEvent, None)
				self.openMessage()
			#end if
			#print('cd', self.currentDir)
			return False
		#end if
		#print(widget, event, dir(event))
		#print(Gdk.EventType._2BUTTON_PRESS)
		#print([(x, getattr(event, x)) for x in dir(event)])
		return False
	#end def clickMsgTable(widget, event)
	
	def keyMsgView(self, widget, event):
		"""
		Handle key presses when message view is selected
		"""
		#print("key press", event, event.type, event.state, event.string, event.keyval)
		if event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == ord('f'):
			# Search Message Body is Ctrl-f
			self.showSearch()
			return True
		#end if
		if event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == ord('o'):
			# Open is Ctrl-o
			self.openFile()
			return True
		#end if
		if event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == ord('1'):
			# Open in new window is Ctrl-1
			fakeEvent = Gdk.EventButton()
			fakeEvent.type = Gdk.EventType.BUTTON_PRESS
			if self.currentDir == 'drafts':
				self.clickDraft(None, None, fakeEvent, None)
			else:
				#self.clickReply(None, None, fakeEvent, None)
				self.openMessage()
			#end if
			#print('cd', self.currentDir)
			return True
		#end if
		return False
	#end def keyMsgView(widget, event)
	
	def clickCompose(self, x=None):
		"""
		Compose button or menu item
		"""
		if self.composer != None and self.composer.ready:
			#self.composer.showAll()
			self.composer.present()
			return True
		#end if
		# If the composer isn't ready, it's been destroyed and is not available 
		# to be used, so create a new one.
		self.composer = ComposeWindow(self)
		self.composer.showAll()
		return True
	#end def clickCompose(x)
	
	def drawIndex(self, path, dirs_box, dirs_store, targets=None):
		"""
		Draw the list of directories in the dirs_box on the left.
		"""
		# select inbox if it's none
		selectInbox = True
		if self.dirChangedConnection != None:
			self.dirs_tree.disconnect(self.dirChangedConnection)
			selectInbox = False
		#end if
		inboxDir = self.config['Mail1'].get('inbox', 'inbox')
		# Make sure inboxDir ends with a / so that it matches cdir.
		if len(inboxDir) > 0 and inboxDir[-1] != '/':
			inboxDir += '/'
		#end if
		#TODO: what if inbox is a child? dun dun dunn!
		#TODO: total number of messages in the directory
		#TODO: unread messages in the directory
		#TODO: Background color
		#TODO: Drag and drop
		dirs_store.clear()
		self.als = []
		# directories on the left
		folders = mail1.readDir(path)
		dirs = 0
		for cdir in folders['dirs']:
			"""
			alld = mail1.readDir(os.path.join(path, cdir, 'cur', ''))
			#all_len = len(alld['files'])
			read_len = 0
			for filename in alld['files']:
				if filename[-4:] == ':2,S': read_len += 1
				if filename[-5:] == ':2,RS': read_len += 1
			#next file
			#unread_len = all_len - read_len
			"""
			q = cdir
			parent = dirs_store.append(None, (cdir[:-1], 0, 0, '', '#fff'))
			if selectInbox and cdir == inboxDir:
				sel = self.dirs_tree.get_selection()
				patha = dirs_store.get_path(parent)
				#print('inbox', parent, patha)
				sel.select_path(patha)
			#end if
			#als[-1].set_relief(Gtk.ReliefStyle.NONE)
			#print(als[-1].get_label())
			"""
			# Left aligned
			als[-1].xalign = 0.0
			als[-1].connect('clicked', self.clickDir, cdir)
			als[-1].drag_dest_set(Gtk.DestDefaults.ALL, [], DRAG_ACTION)
			als[-1].connect('drag-data-received', self.dragData, cdir)
			als[-1].connect('drag-drop', self.dragDrop)
			als[-1].drag_dest_set_target_list(targets)
			dirs_box.pack_start(als[-1], False, True, 0)
			"""
			# subdirs .inbox.directory/
			subdir = '.' + cdir[:-1] + '.directory' + os.sep
			if subdir in folders['hidden dirs']:
				q = [q]
				subdirs = mail1.readDir(os.path.join(path, subdir))
				for s in subdirs['dirs']:
					q.append(os.path.join(subdir, s))
					dirs_store.append(parent, (s[:-1], 0, 0, '', '#fff'))
					"""
					#als[-1].set_relief(Gtk.ReliefStyle.NONE)
					als[-1].connect('clicked', self.clickDir, os.path.join(subdir, s))
					als[-1].drag_dest_set(Gtk.DestDefaults.ALL, [], DRAG_ACTION)
					als[-1].connect('drag-data-received', self.dragData, os.path.join(subdir, s))
					als[-1].connect('drag-drop', self.dragDrop)
					als[-1].drag_dest_set_target_list(targets)
					# Left aligned
					als[-1].xalign = 0.0
					dirs_box.pack_start(als[-1], False, True, 0)
					"""
					dirs += 1
				#next s
			#end if
			#als is a tree of string or array of string
			self.als.append(q)
			dirs += 1
		#next cdir
		self.dirChangedConnection = self.dirs_tree.connect('cursor-changed', self.clickDir)
		return self.als
	#end def drawIndex(path, dirs_box = None)

	def clickDir(self, widget, path=None):
		"""
		When the user clicks on a directory, change the directory listing.
		"""
		if path != None:
			print("did you click on:", path)
			return True
		else:
			cursor = widget.get_cursor()
			# FIXME: Strange bug.
			if cursor == None or cursor[0] == None or cursor[0][0] == None:
				return False
			#end if
			#print('cursor', cursor, cursor[0][0], self.messages[cursor[0][0]])
			pos = cursor[0][0]
		#end if
		directory = self.als[pos]
		if len(cursor[0]) > 1 and isinstance(directory, list):
			# It's down a level.
			directory = directory[cursor[0][1] + 1]
		elif isinstance(directory, list):
			# Grab the parent.
			directory = directory[0]
		#end if
		print('dir', repr(directory))
		# This disconnection prevents the clickMsg from being fired hundreds of times.
		self.msg_table.disconnect(self.cursorChangedConnection)
		self.msg_store.clear()
		self.auto_added = 0
		self.cursorChangedConnection = self.msg_table.connect('cursor-changed', self.clickMsg)
		self.drawDir(self.maildir, directory, msg_store = self.msg_store)
		adj = self.msgs_win.get_vadjustment()
		adj.set_value(0)
		adj = self.msgs_win.get_hadjustment()
		adj.set_value(0)
		return True
	#end def clickDir(widget, path)

	def addDirectory(self, widget):
		"""
		Create a directory
		TODO: how does maildir handle multiple levels? Until drawIndex supports them, this function cannot.
		"""
		# Copied from clickFilter
		builder = Gtk.Builder()
		builder.add_from_file(os.path.join(mail1.data_dir, 'glade/add_dir_dialog.glade'))
		dialog = builder.get_object("add_dir_dialog")
		dialog.set_transient_for(self.window)
		name = builder.get_object("name")
		dir_store = builder.get_object("dir_store")
		dir_table = builder.get_object("dir_table")
		rendererDir = Gtk.CellRendererText()
		rendererDir.ellipsize = Pango.EllipsizeMode.END
		name_column = Gtk.TreeViewColumn("Name", rendererDir, text=0)
		name_column.set_resizable(True)
		dir_table.append_column(name_column)
		dir_store.clear()
		folders = mail1.readDir(self.maildir)
		for cdir in folders['dirs']:
			parent = dir_store.append(None, (cdir,))
			subdir = '.' + cdir[:-1] + '.directory' + os.sep
			if subdir in folders['hidden dirs']:
				subdirs = mail1.readDir(os.path.join(self.maildir, subdir))
				for s in subdirs['dirs']:
					#dir_store.append(parent, (os.path.join(cdir, s),))
					dir_store.append(parent, (s,))
				#next s
			#end if
		#next cdir
		response = dialog.run()
		if response == Gtk.ResponseType.OK:
			# Save the filter to ~/.mail/filter/x.txt
			dir_name = name.get_text()
			list_store, rows = dir_table.get_selection().get_selected_rows()
			dir_parent = ''
			for row in rows:
				v = row.get_indices()
				if len(v) > 1:
					row_i = v
					dir_parent = os.path.join(dir_store[row_i[0]][0], dir_store[row_i][0])
				else:
					row_i = v[0]
					dir_parent = dir_store[row_i][0]
				#end if
			#next row
			#print('okay', filter_name, filter_field, filter_value, filter_dir)
			if dir_parent != '':
				dirs = dir_parent.split(os.sep)
				while '' in dirs: dirs.remove('')
				if len(dirs) > 1:
					alert("BikeIM Mail doesn't support more than 1 level of directories yet", parent=self.window)
					dialog.destroy()
					return True
				#end if
				dir_parent = ''
				if len(dirs) > 0:
					dir_parent += '.{0}.directory'.format(dirs[0])
				#end if
			#end if
			if os.path.exists(os.path.join(self.maildir, dir_parent, dir_name, 'cur')):
				alert("Directory already exists", parent=self.window)
				dialog.destroy()
				return True
			#end if
			try:
				os.makedirs(os.path.join(self.maildir, dir_parent, dir_name, 'cur'))
				os.mkdir(os.path.join(self.maildir, dir_parent, dir_name, 'tmp'))
				os.mkdir(os.path.join(self.maildir, dir_parent, dir_name, 'new'))
			except Exception as e:
				alert("Creating Directory failed\n" + str(e), parent=self.window)
				dialog.destroy()
				return True
			#end try
			
			# Redraw the buttons on the left.
			targets = Gtk.TargetList.new([])
			targets.add_text_targets(1)
			self.drawIndex(self.maildir, self.dirs_box, self.dirs_store, targets)
			self.msg_table.drag_source_set_target_list(targets)
			self.window.show_all()
		#end if
		#print('fingerprints', self.fingerprints)
		dialog.destroy()
		return True
	#end def addDirectory(self, widget)
	
	def clickMsg(self, widget, add='', use_part=None):
		"""
		When the user clicks on a message, change the message.
		"""
		prev_msg = self.message
		self.message = None
		cursor = widget.get_cursor()
		# FIXME: Strange bug.
		if cursor == None or cursor[0] == None or cursor[0][0] == None:
			return False
		#end if
		#print('cursor', cursor, cursor[0][0], self.messages[cursor[0][0]])
		pos = cursor[0][0]
		#print(widget)
		message = self.messages[pos]
		if len(cursor[0]) > 1 and isinstance(message, list):
			# It's down a level.
			message = self.messages[pos][1][cursor[0][1]]
		elif isinstance(message, list):
			# Grab the parent.
			message = message[0]
		#end if
		#print('clickMsg', message)
		#fname = self.maildir + '/' + self.currentDir + '/cur/' + message
		if message == prev_msg and self.add == add:
			print("old", message, prev_msg)
			self.message = message
			self.messageDir = self.currentDir
			return True
		#end if
		self.showMsg(message, add, use_part)
		# FIXME: Race condition
		self.messageDir = self.currentDir
		return True
	#end def clickMsg(widget, add='', use_part=None)
	
	def showMsg(self, filename, add='', use_part=None, dontShow=False):
		"""
		Draw a message into the message_view.
		"""
		dirName, message = os.path.split(filename)
		if dirName:
			# This code path requires showMsg not from clickMsg for example cli.
			if dirName.startswith(self.maildir):
				dirName = dirName[len(self.maildir):].lstrip(os.sep)
			else:
				print("showMsg({0!r}): {1} not in maildir".format(filename, dirName))
			#end if
			if dirName.endswith(os.sep + 'cur'):
				dirName = dirName[:-len(os.sep + 'cur')]
			else:
				print("showMsg({0!r}): {1} not using cur".format(filename, dirName))
			#end if
			#self.currentDir = dirName.strip('/')
			#print('getting dirname from filename', dirName)
		else:
			# TODO: Make sure this is correct in all possible cases.
			dirName = self.currentDir
			#print('getting dirname from currentDir', dirName)
		#end if
		if dontShow == False:
			message_buffer = Gtk.TextBuffer()
			self.drawMsg(message_buffer, self.maildir, dirName, message, add, use_part)
			#message_buffer.set_text(message_text)
			self.message_view.set_buffer(message_buffer)
			self.message_view.scroll_to_iter(message_buffer.get_start_iter(), 0.0, True, 0.0, 0.0)
			# FIXME: This grab focus makes it impossible to multiselect with mouse
			if self.config['Mail1GUI'].getboolean('focus', False):
				self.message_view.grab_focus()
			#end if
		#end if
		self.message = message
		# FIXME: Race condition
		self.messageDir = dirName
		self.add = add
		return True
	#end def showMsg(filename, add='', use_part=None)
	
	def clickHtml(self, tag, widget, event, iterc):
		"""
		When the user clicks HTML, give them HTML in the message view.
		"""
		if event.type != Gdk.EventType.BUTTON_PRESS: return True
		self.clickMsg(self.msg_table, 'html')
		return True
	#end def clickHtml(tag, widget, event, iterc)
	
	def clickHtmlText(self, tag, widget, event, iterc):
		"""
		When the user clicks HTML text, give them HTML stripped in the message view.
		"""
		if event.type != Gdk.EventType.BUTTON_PRESS: return True
		self.clickMsg(self.msg_table, 'text_html')
		return True
	#end def clickHtml(tag, widget, event, iterc)
	
	def clickHtmlClean(self, tag, widget, event, iterc):
		"""
		When the user clicks HTML text, give them HTML without active 
		components in the message view.
		"""
		if event.type != Gdk.EventType.BUTTON_PRESS: return True
		self.clickMsg(self.msg_table, 'clean_html')
		return True
	#end def clickHtml(tag, widget, event, iterc)

	def clickHtmlFootnote(self, tag, widget, event, iterc):
		"""
		When the user clicks HTML text, give them HTML without active 
		components in the message view.
		"""
		if event.type != Gdk.EventType.BUTTON_PRESS: return True
		self.clickMsg(self.msg_table, 'footnote_html')
		return True
	#end def clickHtml(tag, widget, event, iterc)

	def clickPlain(self, tag, widget, event, iterc):
		"""
		When the user clicks Plain, give them Text in the message view.
		"""
		if event.type != Gdk.EventType.BUTTON_PRESS: return True
		self.clickMsg(self.msg_table)
		return True
	#end def clickPlain(tag, widget, event, iterc)

	def clickReply(self, tag, widget, event, iterc):
		"""
		When the user clicks Reply, open the compose window with the correct 
		data in the compose box so that they can reply.
		"""
		if event.type != Gdk.EventType.BUTTON_PRESS: return False
		#print(x)
		print("Reply")
		self.clickCompose()
		if self.composer == None:
			return True
		#end if
		self.composer.saveDraft()
		msg = None
		if widget == self.message_view:
			# The message must be the main window message if the widget is message_view
			msg = self.msg
		else:
			for win in self.message_wins:
				if win.message_view == widget:
					#print("Found the right window. Get its message.")
					msg = win.msg
				#end if
			#next win
		#end if
		if msg == None:
			print("Error: cannot find the widget in all the windows.", widget, self.message_wins)
			return True
		#end if
		tox = msg['from']
		if 'reply-to' in msg:
			tox = msg['reply-to']
		#end if
		tox = decodeHeaderGtk(tox, removeNewline=True)
		self.composer.inputTo.set_text(tox)
		subject = decodeHeaderGtk(msg['subject'], removeNewline=True)
		if subject[:3].lower() != 're:':
			subject = 'Re: ' + subject
		#end if
		self.composer.inputSubject.set_text(subject)
		mb = widget.get_buffer()
		start_mark = mb.get_mark("body_start")
		if start_mark != None:
			curr_iter = mb.get_iter_at_mark(start_mark)
			end_iter = mb.get_iter_at_mark(mb.get_mark("body_end"))
			body_data = mb.get_slice(curr_iter, end_iter, False)
			body_data = mail1.replyQuote(body_data)
			body_data = ('\n\nOn {0}, {1} wrote:\n'.format(msg['date'], msg['from'])) + body_data
			black_color = Gdk_color_parse('#000000')
			body_buffer = self.composer.newBuffer()
			writable_tag = body_buffer.create_tag("writable", editable=True,
				foreground_rgba=black_color,
				size_points=10.0)
			#print(dir(body_buffer))
			start_iter = body_buffer.get_start_iter()
			body_buffer.insert_with_tags(start_iter, sanitizeGtk(body_data), writable_tag)
			if 'Message-Id' in msg:
				self.composer.setHeaders('In-Reply-To: ' + msg['Message-Id'] + "\n" + 'References: ' + msg['Message-Id'] + " " + msg.get('References', ''))
			#end if
			self.composer.bodyView.set_buffer(body_buffer)
		#end if
		"""try:
			
		except Exception as e:
			print("Fail.")
			print(e)
		#end try"""
		self.composer.attachments = []
		self.composer.attach_store.clear()
		return True
	#end def clickReply(tag, widget, event, iterc)

	def clickDraft(self, tag, widget, event, iterc):
		"""
		When the user wants to resume a draft they saved.
		"""
		if event.type != Gdk.EventType.BUTTON_PRESS: return False
		if self.currentDir != 'drafts':
			# Easy path to accidental deletion.
			return True
		#end if
		if self.composer:
			self.composer.saveDraft()
		#end if
		self.clickCompose()
		if self.composer == None:
			return True
		#end if
		self.composer.inputTo.set_text('')
		self.composer.inputCc.set_text('')
		self.composer.inputBcc.set_text('')
		if 'to' in self.msg:
			tox = decodeHeaderGtk(self.msg['to'], removeNewline=True)
			self.composer.inputTo.set_text(tox)
		#end if
		if 'cc' in self.msg:
			cc = decodeHeaderGtk(self.msg['cc'], removeNewline=True)
			self.composer.inputCc.set_text(cc)
		#end if
		if 'bcc' in self.msg:
			bcc = decodeHeaderGtk(self.msg['bcc'], removeNewline=True)
			self.composer.inputBcc.set_text(bcc)
		#end if
		if 'subject' in self.msg:
			subject = decodeHeaderGtk(self.msg['subject'], removeNewline=True)
			self.composer.inputSubject.set_text(subject)
		#end if
		mb = self.message_view.get_buffer()
		start_mark = mb.get_mark("body_start")
		body_buffer = self.composer.newBuffer()
		if start_mark != None:
			curr_iter = mb.get_iter_at_mark(start_mark)
			end_iter = mb.get_iter_at_mark(mb.get_mark("body_end"))
			body_data = mb.get_slice(curr_iter, end_iter, False)
			black_color = Gdk_color_parse('#000000')
			writable_tag = body_buffer.create_tag("writable", editable=True,
				foreground_rgba=black_color,
				size_points=10.0)
			#print(dir(body_buffer))
			start_iter = body_buffer.get_start_iter()
			body_buffer.insert_with_tags(start_iter, sanitizeGtk(body_data), writable_tag)
		#end if
		self.composer.bodyView.set_buffer(body_buffer)
		# Headers copy.
		headers = ''
		knowns = ['from', 'to', 'cc', 'bcc', 'date', 'subject', 'content-type', 
			'mime-version', 'content-transfer-encoding']
		for h in self.msg.keys():
			if h.lower() in knowns: continue
			#print('h', h)
			headers += '{0}: {1}\n'.format(h, self.msg[h])
		#next h
		headers = headers.rstrip('\n')
		self.composer.setHeaders(headers)
		# Attachments copy.
		self.composer.attachments = []
		self.composer.attach_store.clear()
		for part in self.msg.walk():
			fn = None
			fn = part.get_filename()
			#print('message walk', fn, part.get_content_type())
			if not fn: continue
			ctype = part.get_content_type()
			data = part.get_payload(decode=True)
			obj = mail1.FileAttachment(fn, data, ctype)
			self.composer.attachments.append((fn, ctype, obj))
			self.composer.attach_store.append([self.composer.att_icon[0], fn, ctype])
		#next part
		if len(self.composer.attachments) > 0:
			self.composer.attach_tree.show_all()
		#end if
		# Make sure that the composer saves this draft if exit.
		self.composer.md = None
		# Delete draft
		print('Draft deleted', os.path.join(self.maildir, self.currentDir, 'cur', self.message))
		os.remove(os.path.join(self.maildir, self.currentDir, 'cur', self.message))
		return True
	#end def clickDraft(tag, widget, event, iterc)

	def clickFile(self, tag, widget, event, iterc):
		"""
		When the user clicks a file, save the file to disk.
		"""
		#print("ack")
		if event.type != Gdk.EventType.BUTTON_PRESS: return False
		print("File", tag, widget, event, iterc)
		buf = widget.get_buffer()
		rev_newline = iterc.backward_search('\n', 0, None)
		found_newline = iterc.forward_search('\n', 0, None)
		if not (found_newline and rev_newline): return True
		filename = buf.get_slice(rev_newline[1], found_newline[0], False)
		#print('hi', repr(filename))
		curr_att = None
		curr_filename = ''
		if filename in self.attachments:
			curr_att = self.attachments[filename]
			curr_filename = curr_att.get_filename()
			print(curr_filename)
		else:
			print(filename, self.attachments.keys())
		#end if
		if curr_att == None:
			return True
		#end if
		
		# Stock file chooser.
		dialog = Gtk.FileChooserDialog("Save File", self.window,
			Gtk.FileChooserAction.SAVE,
			(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
			Gtk.STOCK_SAVE, Gtk.ResponseType.OK))
		if curr_filename:
			dialog.set_current_name(curr_filename)
		else:
			dialog.set_current_name(filename)
		#end if
		dialog.set_do_overwrite_confirmation(True)
		response = dialog.run()
		if response == Gtk.ResponseType.OK:
			print("Save clicked")
			fn = dialog.get_filename()
			print("File selected: " + fn)
			open(fn, 'wb').write(curr_att.get_payload(decode=True))
		elif response == Gtk.ResponseType.CANCEL:
			print("Cancel clicked")
		#end if
		dialog.destroy()
		return True
	#end def clickFile(tag, widget, event, iterc)
	
	def clickSpam(self, tag, widget, event, iterc):
		"""
		When the user clicks spam, move it to the spam folder.
		"""
		if event.type != Gdk.EventType.BUTTON_PRESS: return False
		#print(x)
		print("Spam")
		message = None
		messageDir = None
		if widget == self.message_view:
			# The message must be the main window message if the widget is message_view
			message = self.message
			messageDir = self.messageDir
		else:
			for win in self.message_wins:
				if win.message_view == widget:
					#print("Found the right window. Get its message.")
					message = win.message
					messageDir = win.messageDir
				#end if
			#next win
		#end if
		if message == None or messageDir == None:
			print("Error: cannot find the widget in all the windows.", widget, self.message_wins)
			return True
		#end if
		message_buffer = Gtk.TextBuffer()
		
		self.moveToSpam(message, messageDir, message_buffer)
		
		self.message_view.set_buffer(message_buffer)
		return True
	#end def clickSpam(tag, widget, event, iterc)

	def clickFilter(self, tag, widget, event, iterc):
		"""
		When the user clicks filter, dialog guides them through creating and saving a filter.
		"""
		if event and event.type != Gdk.EventType.BUTTON_PRESS: return False
		print("Filter")

		message = self.message
		#end if
		if isinstance(message, list):
			# We're talking about the parent, not the children.
			message = message[0]
		#end if
		msg = self.msg
		if msg == None:
			msg = {'from':'', }
		#end if
		mfrom = msg['from']
		mfrom = decodeHeaderGtk(mfrom, True, removeNewline=True)
		listid = msg.get('list-id', '')
		listid = decodeHeaderGtk(listid, True, removeNewline=True)
		
		builder = Gtk.Builder()
		builder.add_from_file(os.path.join(mail1.data_dir, 'glade/filter_dialog1.glade'))
		dialog = builder.get_object("filter_dialog")
		dialog.set_transient_for(self.window)
		name = builder.get_object("name")
		field = builder.get_object("field")
		value = builder.get_object("value")
		dir_store = builder.get_object("dir_store")
		dir_table = builder.get_object("dir_table")
		#print(dir_store,  dir_store.append)
		name.set_text(mfrom)
		# TODO: field should be freeform and mail1.filterList should accept any value
		#field.set_text('From')
		field.append('from', 'From')
		field.append('to', 'To')
		field.append('listid', 'List-Id')
		field.append('cc', 'Cc')
		field.append('subject', 'Subject')
		if listid:
			field.set_active(2)
			value.set_text(listid)
		else:
			field.set_active(0)
			value.set_text(mfrom)
		#end if
		rendererDir = Gtk.CellRendererText()
		rendererDir.ellipsize = Pango.EllipsizeMode.END
		name_column = Gtk.TreeViewColumn("Name", rendererDir, text=0)
		name_column.set_resizable(True)
		dir_table.append_column(name_column)
		dir_store.clear()
		folders = mail1.readDir(self.maildir)
		for cdir in folders['dirs']:
			parent = dir_store.append(None, (cdir,))
			subdir = '.' + cdir[:-1] + '.directory' + os.sep
			if subdir in folders['hidden dirs']:
				subdirs = mail1.readDir(os.path.join(self.maildir, subdir))
				for s in subdirs['dirs']:
					#dir_store.append(parent, (os.path.join(cdir, s),))
					dir_store.append(parent, (s,))
				#next s
			#end if
		#next cdir
		response = dialog.run()
		if response == Gtk.ResponseType.OK:
			# Save the filter to ~/.mail/filter/x.txt
			filter_name = name.get_text()
			filter_field = field.get_active_text()
			filter_value = value.get_text()
			list_store, rows = dir_table.get_selection().get_selected_rows()
			filter_dir = ''
			for row in rows:
				v = row.get_indices()
				if len(v) > 1:
					row_i = v
					filter_dir = os.path.join(dir_store[row_i[0]][0], dir_store[row_i][0])
				else:
					row_i = v[0]
					filter_dir = dir_store[row_i][0]
				#end if
			#next row
			filter_dir = filter_dir.rstrip('/')
			if filter_value == '':
				alert('Unwilling to add a filter with no value. Try again.', parent=self.window)
				# TODO: let them retry, wow really?
				dialog.destroy()
				return True
			#end if
			if filter_dir == '':
				alert('Unwilling to add a filter with no destination. Try again.', parent=self.window)
				# TODO: let them retry, wow really?
				dialog.destroy()
				return True
			#end if
			#print('okay', filter_name, filter_field, filter_value, filter_dir)
			hidden_dir = mail1.hidden_dir
			# if they are filtering spam, it goes into junk_filter1.json
			# TODO: junk_filter1.json isn't currently used.
			filter_filename = 'list_filter1.json'
			junk_dir = os.path.join('inbox', 'spam1')
			if 'spam' in self.config['Mail1']:
				junk_dir_long = os.path.expanduser(self.config['Mail1']['spam'])
				if junk_dir_long.startswith(self.maildir):
					junk_dir_long = junk_dir_long[len(self.maildir):].strip(os.sep)
				#end if
				# Convert .inbox.directory/spam1 to inbox/spam1
				junk_dir = mail1.maildir_dir_re.sub(r'\1' + os.sep, junk_dir_long)
				junk_dir = junk_dir.strip(os.sep) + os.sep
			#end if
			# TODO: junk_filter1.json isn't currently used.
			#if filter_dir == junk_dir:
			#	filter_filename = 'junk_filter1.json'
			#end if
			# Filter name must be unique. Make this less expensive.
			filter_names = [json.loads(x)[0] for x in open(os.path.join(hidden_dir, 'filter', filter_filename), 'r').read().strip().split('\n') if len(x) > 13]
			if filter_name in filter_names:
				relevant_filter_names = [fn for fn in filter_names if fn.startswith(filter_name)]
				filter_name2 = '{0}1'.format(filter_name)
				i = 1
				while filter_name2 in relevant_filter_names:
					i += 1
					filter_name2 = '{0}{1}'.format(filter_name, i)
				#loop
				alert("Filter name must be unique. Renaming this filter {0}.".format(filter_name2), parent=self.window)
				# TODO: let them rename it, wow really?
				filter_name = filter_name2
			#end if
			open(os.path.join(hidden_dir, 'filter', filter_filename), 'a').write('\n' + json.dumps([filter_name, filter_field, filter_value, filter_dir]))
			if self.filters_window:
				self.filters_window.filterStore.append([filter_field, filter_value, filter_dir])
				self.filters_window.filters.append([filter_name, filter_field, filter_value, filter_dir])
			#end if
		#end if
		dialog.destroy()
		
		return True
	#end def clickFilter(tag, widget, event, iterc)

	def clickLink(self, tag, widget, event, iterc):
		"""
		Link to plaintext or whatever.
		"""
		# TODO: Save plaintext to file.
		return True
	#end def clickLink(tag, widget, event, iterc)

	def moveToSpam(self, message, messageDir, mb=None, curr_iter=None):
		"""
		Move a file in the current directory to the spam folder.
		Optionally tell the user about it with mb.
		"""
		if mb and not curr_iter:
			start_iter = mb.get_start_iter()
			curr_iter = start_iter
		#end if
		destDir = os.path.join(self.maildir, '.inbox.directory', 'spam1', 'cur')
		if 'spam' in self.config['Mail1']:
			destDir = os.path.expanduser(self.config['Mail1']['spam'])
		#end if
		
		messagePath = os.path.join(self.maildir, messageDir, 'cur', message)
		if not messagePath.startswith(self.maildir):
			print('Nothing outside Mail please: ' + messagePath)
			if mb:
				mb.insert(curr_iter, 'Nothing outside Mail please: ' + messagePath + '\n')
			#end if
			return True
		#end if
		if ('..' + os.sep) in messagePath:
			print('No directory traversal please: ' + messagePath)
			if mb:
				mb.insert(curr_iter, 'No directory traversal please: ' + messagePath + '\n')
			return True
		#end if
		message = messagePath.split(os.sep)[-1]
		if not os.path.exists(messagePath):
			print('Error: src does not exist:', messagePath)
			if mb:
				mb.insert(curr_iter, 'Error: src does not exist: ' + messagePath + '\n')
			return True
		#end if
		destPath = os.path.join(destDir, message)
		if os.path.exists(destPath):
			print('Error: dest exists:', destPath)
			if mb:
				mb.insert(curr_iter, 'Error: dest exists: ' + destPath + '\n')
			return True
		#end if
		if message == self.message:
			msg = self.msg
		else:
			if self.index == None:
				self.index = mail1.getIndex(messageDir, 0)
			#end if
			msg = getMessage(message, self.index, os.path.join(self.maildir, messageDir, 'cur'))
		#end if
		print('renamed', messagePath, destPath)
		if mb:
			mb.insert(curr_iter, 'renamed ' + messagePath + ' ' + destPath + '\n')
		os.rename(messagePath, destPath)
		# 644
		try:
			os.chmod(destPath, stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH)
		except OSError as e:
			print("Warning: chmod failed:", e)
			if mb: mb.insert(curr_iter, "Warning: chmod failed: " + str(e))
		#end try
		sfrom = baseUnicode(msg['from'])
		if sfrom not in self.blacklist:
			hidden_dir = mail1.hidden_dir
			self.blacklist.append(sfrom)
			# Append.
			open(os.path.join(hidden_dir, 'blacklist.txt'), 'a').write(sfrom + '\n')
		#end if
		return True
	#end def moveToSpam(self, message, messageDir, mb = None, curr_iter = None)
	
	def clickSpamMulti(self):
		"""
		Take the selection in the message table and send it to spam.
		"""
		list_store, rows = self.msg_table.get_selection().get_selected_rows()
		#print("Multi", dir(list_store), (rows))
		to_remove = []
		remove_iters = {}
		for row in rows:
			#print("row", row, row.__class__)
			v = row.get_indices()
			to_remove.append(v)
			cur_iter = self.msg_store.get_iter(row)
			remove_iters[repr(v)] = cur_iter
		#next row
		# Don't allow clickMsg until we've removed the row.
		self.msg_table.disconnect(self.cursorChangedConnection)
		message_buffer = Gtk.TextBuffer()
		start_iter = message_buffer.get_start_iter()
		curr_iter = start_iter
		file_list = ''
		rearrange = False
		# Reverse sort makes it possible to do this correctly without removing a parent before a child or an early sibling.
		# NOTE: Be very careful changing anything in this function because it can easily crash.
		to_remove.sort(reverse=True)
		for index in to_remove:
			can_remove = False
			message = None
			if isinstance(index, list):
				if len(index) >= 2:
					message = self.messages[index[0]][1].pop(index[1])
					can_remove = True
				else:
					if isinstance(self.messages[index[0]], baseUnicode) or len(self.messages[index[0]][1]) == 0:
						message = self.messages.pop(index[0])
						can_remove = True
					else:
						print("bug", isinstance(self.messages[index[0]], baseUnicode), self.messages[index[0]][1])
						message = self.messages[index[0]]
					#end if
				#end if
			else:
				print("Error: Sorting relies on lists")
				raise Exception("indexes should be lists")
				#message = self.messages[index]
			#end if
			#print('mess', message)
			if isinstance(message, list):
				# We're talking about the parent, not the children.
				message = message[0]
			#end if
			if self.moveToSpam(message, self.currentDir, message_buffer, curr_iter):
				file_list += message + '\n'
			#end if
			if can_remove:
				self.msg_store.remove(remove_iters[repr(index)])
			else:
				# TODO: If it's a parent, move the children where they go.
				# Parent needs to not have children.
				rearrange = True
			#end if
		#next index
		message_buffer.insert(curr_iter, 'Files moved: \n' + file_list + '\n')
		self.message_view.set_buffer(message_buffer)
		
		self.cursorChangedConnection = self.msg_table.connect('cursor-changed', self.clickMsg)

		return True
	#end def clickSpamMulti()
	
	def unblockMulti(self):
		"""
		Take the selection in the message table and unblock.
		"""
		list_store, rows = self.msg_table.get_selection().get_selected_rows()
		#print("Multi", dir(list_store), (rows))
		to_unblock = []
		for row in rows:
			v = row.get_indices()
			if len(v) > 1:
				to_unblock.append(v)
			else:
				to_unblock.append(v[0])
			#end if
		#next row
		blacklistModified = False
		blacklistIpModified = False
		for index in to_unblock:
			if isinstance(index, list):
				message1 = self.messages[index[0]]
				if len(index) >= 2:
					message = message1[1][index[1]]
				else:
					message = message1
				#end if
			else:
				message = self.messages[index]
			#end if
			if isinstance(message, list):
				# We're talking about the parent, not the children.
				message = message[0]
			#end if
			if message != self.message:
				# Load message, get IP, unblock, save
				path, currentDir = self.maildir, self.currentDir
				fname = os.path.join(path, currentDir, 'cur', message)
				fp = open(fname, 'rb')
				msg = mail1.message_from_binary_file(fp, fname)
			else:
				msg = self.msg
			#end if
			mip = mail1.getReceivedIp(msg, self.config['Mail1'].get('server_name', ''))
			if mip and mip in self.blacklist_ip:
				self.blacklist_ip.remove(mip)
				blacklistIpModified = True
				print(mip, "was in your IP blacklist.")
			#end if
			mfrom = msg['from']
			fenc = decodeHeaderGtk(mfrom, False)
			mfrom = sanitizeGtk(fenc)
			if mfrom in self.blacklist:
				self.blacklist.remove(mfrom)
				blacklistModified = True
				print(mfrom, "was in your blacklist.")
			#end if
		#next index
		if blacklistIpModified:
			hidden_dir = mail1.hidden_dir
			open(os.path.join(hidden_dir, 'blacklist_ip.txt'), 'w').write('\n'.join(self.blacklist_ip))
			print("Saved", hidden_dir + 'blacklist_ip.txt')
		#end if
		if blacklistModified:
			hidden_dir = mail1.hidden_dir
			open(os.path.join(hidden_dir, 'blacklist.txt'), 'w').write('\n'.join(self.blacklist))
			print("Saved", hidden_dir + 'blacklist.txt')
		#end if
		return True
	#end def unblockMulti()
	
	def deleteMulti(self):
		"""
		Take the selection in the message table and delete with undo.
		"""
		list_store, rows = self.msg_table.get_selection().get_selected_rows()
		#print("Multi", dir(list_store), (rows))
		to_delete = []
		remove_iters = {}
		for row in rows:
			v = row.get_indices()
			if len(v) > 1:
				to_delete.append(v)
			else:
				to_delete.append(v[0])
			#end if
			cur_iter = self.msg_store.get_iter(row)
			remove_iters[repr(v)] = cur_iter
		#next row
		# Don't allow clickMsg until we've removed the row.
		self.msg_table.disconnect(self.cursorChangedConnection)

		for index in to_delete:
			if isinstance(index, list):
				message1 = self.messages[index[0]]
				if len(index) >= 2:
					message = message1[1][index[1]]
				else:
					message = message1
				#end if
			else:
				message = self.messages[index]
			#end if
			if isinstance(message, list):
				# We're talking about the parent, not the children.
				message = message[0]
			#end if
			path, currentDir = self.maildir, self.currentDir
			fname = os.path.join(path, currentDir, 'cur', message)
			fp = open(fname, 'rb')
			fp.seek(0, io.SEEK_END)
			file_size = fp.tell()
			if file_size > 10000000:
				# 10 MB is our limit currently
				temp_deleted = os.path.join(path, 'deleted', 'tmp')
				if not os.path.exists(temp_deleted):
					os.makedirs(temp_deleted)
				#end if
				# Rename
				os.rename(fname, os.path.join(temp_deleted, message))
				self.undoEvents.append(('deleted file', fname, os.path.join(temp_deleted, message)))
				print('renamed for deletion on exit', fname)
				# TODO: can we remove?
				self.msg_store.remove(remove_iters[repr(index)])
				continue
			#end if
			# Save to memory
			fp.seek(0, io.SEEK_SET)
			self.undoEvents.append(('deleted memory', fname, fp.read()))
			# Delete here
			os.remove(fname)
			print('deleted', fname, 'ctrl-z to undo')
			# TODO: can we remove?
			self.msg_store.remove(remove_iters[repr(index)])
		#next index
		
		# now you can click again.
		self.cursorChangedConnection = self.msg_table.connect('cursor-changed', self.clickMsg)
		return True
	#end def deleteMulti()
	
	def cutMulti(self, copy=False):
		"""
		Take the selection in the message table and cut.
		"""
		list_store, rows = self.msg_table.get_selection().get_selected_rows()
		#print("Multi", dir(list_store), (rows))
		to_delete = []
		#remove_iters = {}
		for row in rows:
			v = row.get_indices()
			if len(v) > 1:
				to_delete.append(v)
			else:
				to_delete.append(v[0])
			#end if
			cur_iter = self.msg_store.get_iter(row)
			#remove_iters[repr(v)] = cur_iter
		#next row
		# Don't allow clickMsg until we've removed the row.
		self.msg_table.disconnect(self.cursorChangedConnection)
		self.cutBuffer = []
		self.copyBuffer = []

		for index in to_delete:
			if isinstance(index, list):
				message1 = self.messages[index[0]]
				if len(index) >= 2:
					message = message1[1][index[1]]
				else:
					message = message1
				#end if
			else:
				message = self.messages[index]
			#end if
			if isinstance(message, list):
				# We're talking about the parent, not the children.
				message = message[0]
			#end if
			path, currentDir = self.maildir, self.currentDir
			fname = os.path.join(path, currentDir, 'cur', message)
			if copy:
				self.copyBuffer.append(fname)
			else:
				self.cutBuffer.append(fname)
			#end if
		#next index
		
		# now you can click again.
		self.cursorChangedConnection = self.msg_table.connect('cursor-changed', self.clickMsg)
		return True
	#end def cutMulti(copy=False)

	def copyMulti(self):
		return self.cutMulti(True)
	
	def paste(self):
		if self.cutBuffer and self.copyBuffer:
			print("You need to fix the cut and copy code. We have cut and copy.", self.cutBuffer, self.copyBuffer)
		#end if
		if self.cutBuffer == []:
			if self.copyBuffer == []:
				# nothing to paste
				return True
			#end if
			# copy time
			for filename in self.copyBuffer:
				
				path, currentDir = self.maildir, self.currentDir
				message = os.path.basename(filename)
				output_filename = os.path.join(path, currentDir, 'cur', message)
				if output_filename == filename:
					print("Cannot copy a file to itself.", filename)
					break
				#end if
				if os.path.exists(output_filename):
					# TODO: overwrite?
					print("File already exists in this directory.", output_filename)
					continue
				#end if
				output_file = open(output_filename, 'wb')
				input_file = open(filename, 'rb')
				data = b'#'
				while data != b'':
					data = input_file.read(4096)
					output_file.write(data)
				#loop
				output_file.close()
				input_file.close()
			#next filename
		else:
			# cut time
			for filename in self.cutBuffer:
				
				path, currentDir = self.maildir, self.currentDir
				message = os.path.basename(filename)
				output_filename = os.path.join(path, currentDir, 'cur', message)
				if output_filename == filename:
					print("Cannot copy a file to itself.", filename)
					break
				#end if
				if os.path.exists(output_filename):
					# TODO: overwrite?
					print("File already exists in this directory.", output_filename)
					continue
				#end if
				os.rename(filename, output_filename)
			#next filename
		#end if
		return True
	#end def paste()
	
	def undo(self):
		"""
		Pop something off the undo buffer.
		"""
		if len(self.undoEvents) == 0: return False
		event = self.undoEvents.pop()
		if event[0] == 'deleted file':
			print('undo delete, moved file back', event[1])
			os.rename(event[2], event[1])
		elif event[0] == 'deleted memory':
			print('undo delete, wrote file from memory', event[1])
			open(event[1], 'wb').write(event[2])
		else:
			print("Error: undo failed. Sorry about that.", event, file=sys.stdout)
		#end if
		#TODO: put the message back into the directory list.
		return True
	#end def undo(self)

	def clickHeader(self, tag, widget, event, iterc):
		"""
		Show e-mail headers.
		"""
		if event.type != Gdk.EventType.BUTTON_PRESS: return False
		#print('header', tag, widget, event)
		if widget == None:
			# Allow this function to be called without valid args.
			widget = self.message_view
		#end if
		# Draw the extra headers.
		mb = widget.get_buffer()
		if not mb: return True
		if self.showHeader:
			curr_iter = mb.get_iter_at_mark(mb.get_mark("extra_headers"))
			tag_table = mb.get_tag_table()
			bold_tag = tag_table.lookup("bold")
			normal_tag = tag_table.lookup("normal")
			#i = 0
			for (k, v) in self.extra_headers:
				mb.insert_with_tags(curr_iter, sanitizeGtk(k) + ': ', bold_tag)
				if isinstance(v, bytes):
					v = mail1.guessEncoding(v)
				#end if
				mb.insert_with_tags(curr_iter, sanitizeGtk(v) + '\n', normal_tag)
			#next k,v
			
			mb.insert_with_tags(curr_iter, '\n', normal_tag)
			mb.create_mark("end_extra_headers", curr_iter, True)
			self.showHeader = False
			return True
		#end if
		# Remove the extra headers.
		curr_iter = mb.get_iter_at_mark(mb.get_mark("extra_headers"))
		end_iter  = mb.get_iter_at_mark(mb.get_mark("end_extra_headers"))
		# Remove all text.
		mb.delete(curr_iter, end_iter)
		self.showHeader = True
		return True
	#end def clickHeader(tag, widget, event, iterc)
	
	def drawDir(self, path, currentDir = '', page = 0, msg_store=None):
		"""
		Draw the directory listing in the treeview in the upper right.
		path is the maildir path.
		currentDir is the directory we're currently in relative to path.
		page must be 0. It is currently only so that index points to the correct file and will be deprecated.
		"""
		if self.drawDirRunning:
			self.drawDirExit = True
			# Wait until it's cleared
			#while self.drawDirRunning:
			#	time.sleep(4.2)
			#	print(self.drawDirRunning)
			##loop
		#end if
		if currentDir.strip('/') == 'sent-mail':
			#print('to', currentDir, self.msg_table.get_column(2).get_title())
			self.msg_table.get_column(2).set_title('To')
		else:
			self.msg_table.get_column(2).set_title('From')
		#end if
		loader = self.drawDirGenerator(path, currentDir, page, msg_store)
		GObject.idle_add(lambda: next(loader, False))
		return True
	#end def drawDir(self, path, currentDir = '', page = 0, msg_store=None)
	
	def drawDirGenerator(self, path, currentDir = '', page = 0, msg_store=None):
		"""
		A generator that yields True while it draws to msg_store.
		path is the maildir path.
		currentDir is the directory we're currently in relative to path.
		page must be 0. It is currently only so that index points to the correct file and will be deprecated.
		It exits as soon as possible when self.drawDirExit == True
		See drawDir() for usage.
		"""

		if msg_store == None:
			print("Error: drawDir called incorrectly, always need msg_store")
			return True
		#end if
		self.drawDirRunning = True
		# Unselect everything to fix a bug. Didn't work yet.
		select = self.msg_table.get_selection()
		if select: select.unselect_all()
		page_size = 50
		self.currentDir = currentDir.strip('/')
		# Disable this fun option unless configuration option.
		rainbow = False
		if self.config['Mail1GUI'].getboolean('rainbow', False) and 'rainbow' in self.currentDir:
			rainbow = True
		#end if
		
		self.index = mail1.getIndex(currentDir, page)
		
		# messages on the top right
		start = time.time()
		messages = mail1.readDir(os.path.join(path, currentDir, 'cur'))
		show = messages['files']
		show.sort(reverse = True)
		#show = show[page*pagesize:(page+1)*pagesize]
		max_messages = 10000
		try:
			max_messages = int(self.config['Mail1GUI'].getint('max_messages', 10000))
		except ValueError:
			print("Warning: invalid value in max_messages: {0}".format(self.config['Mail1GUI']['max_messages']))
		#end try
		date_local = self.config['Mail1GUI'].getboolean('date_local', False)
		show = show[:max_messages]
		self.messages = show[:]
		curr = time.time()
		print('readdir, sort dt:', curr - start)
		threaded_rows = []
		indexed = 0
		msg_i = 0
		msg_i_mod = 0
		show_i = 0
		# A way to get out of this very long loop.
		self.drawDirExit = False
		self.msg_table.freeze_child_notify()
		for message in show:
			#if msg_i < 300:
			#print('msg_i', msg_i)
			#end if
			if self.drawDirExit:
				print('exit draw dir')
				break
			#end if
			fname = os.path.join(path, currentDir, 'cur', message)
			if message not in self.index:
				fp = open(fname, 'rb')
				msg = mail1.message_from_binary_file(fp, fname)
				# Write to the index
				# These are Header objects, so we convert to str.
				msg_dict = {}
				msg_dict['subject'] = sanitizePy2(msg['subject'])
				msg_dict['from']    = sanitizePy2(msg['from'])
				msg_dict['to']      = sanitizePy2(msg['to'])
				msg_dict['date']    = sanitizePy2(msg['date'])
				if 'X-Spam-Level' in msg:
					msg_dict['X-Spam-Level'] = msg.get('X-Spam-Level', '')
				#end if
				if 'Received-SPF' in msg:
					msg_dict['Received-SPF'] = msg.get('Received-SPF', '')
				#end if
				if 'DKIM-Signature' in msg:
					msg_dict['DKIM-Signature']    = '1'
				#end if
				attach = False
				for b in msg.walk():
					ctype = b.get_content_type()
					# No false positive for encrypted openpgp/mime.
					if ctype == 'multipart/encrypted':
						break
					#end if
					if ctype not in MaildirWindow.legit_types:
						attach = True
						break
					#end if
				#next b
				if attach:
					msg['attach'] = msg_dict['attach'] = '1'
				#end if
				has_valid_text_plain = False
				for b in msg.walk():
					ctype = b.get_content_type()
					if ctype.startswith('text/plain'):
						charset = None
						if ctype and 'charset=' in ctype:
							charset = ctype[ctype.index('charset=')+8:]
						#end if
						pdata = b.get_payload(decode=True)
						# Why not just errors='replace'? Encoding is more 
						# valuable than that.
						pdata = mail1.guessEncoding(pdata, charset)
						if mail1.deWhitespace(pdata).strip() != '':
							has_valid_text_plain = True
							break
						#end if
					#end if
				#next b
				if has_valid_text_plain:
					msg['has_valid_text_plain'] = msg_dict['has_valid_text_plain'] = '1'
				#end if
				
				# A fairly cheap index.
				self.index[message] = msg_dict
			else:
				indexed += 1
				# Pull from the index
				msg = self.index[message]
			#end if
			#msg = {'subject':'', 'from':'', 'date':''}
			subject = msg['subject']
			lsubject = decodeHeaderGtk(subject, True, removeNewline=True)
			mfrom = msg['from']
			mfrom = decodeHeaderGtk(mfrom, True, removeNewline=True)
			mto = msg['to']
			mto = decodeHeaderGtk(mto, True, removeNewline=True)
			date = decodeHeaderGtk(msg['date'], True, removeNewline=True)
			if date_local:
				# FIXME: This is just broken.
				#try:
				date = mail1.decodeDateHeader(date)
				#except 
			#end if
			spam_level = len(msg.get('X-Spam-Level', ''))
			spf = str(msg.get('Received-SPF', ''))
			# SPF Softfail is a useful indicator of spam, but..
			if spf.startswith('softfail'):
				# Constant Contact blows.
				if msg.get('DKIM-Signature', '') == '':
					spam_level = max(spam_level, 5)
					#print("SPF softfail without DKIM", subject, fname)
				#end if
			elif spf.startswith('none'):
				# This should certainly be configurable because some people are idiots. Most people are not.
				spam_level = max(spam_level, 3)
				#print("SPF none", subject, fname)
			#end if
			
			# DKIM is now an interesting value. I don't use it because in 
			# increases exposure on my server. It doesn't help me enough yet.
			if msg.get('DKIM-Signature', '') == '':
				spam_level += 2
				#print("no DKIM", subject, fname)
			#end if
			
			mfromSan = mfrom
			# TODO: Remove the name part and the angle brackets.
			# It could be in both lists, so we check.
			if (spam_level < 6) and (mfromSan in self.blacklist):
				spam_level = 6
				#print("blacklist", subject, fname, date)
			#end if
			if mfromSan in self.whitelist:
				spam_level = 0
			#end if
			
			# If we're on the sidelines and it has received header, check IP blacklist.
			if spam_level > 0 and spam_level < 6:
				mip = mail1.getReceivedIp(msg, self.config['Mail1'].get('server_name', ''))
				if mip:
					if mip in self.blacklist_ip:
						spam_level = 6
						# Put this into the index.
						msg_dict['X-Spam-Level'] = '******'
						self.index[message] = msg_dict
						#print("blacklist IP", subject, fname)
					#end if
				#end if
			#end if
			
			iconList = [self.unreadPixBuf, self.twoPixBuf, self.threePixBuf, self.fourPixBuf, self.fivePixBuf, self.spamPixBuf]
			if msg.get('attach', None) == '1':
				iconList = [self.attachPixBuf, self.twoAttachPixBuf, self.threeAttachPixBuf, self.fourAttachPixBuf, self.fiveAttachPixBuf, self.attachSpamPixBuf]
			#end if

			#spam_icon = '<span style="background-color:' + mail1.spam_color[min(spam_level, len(mail1.spam_color)-1)] + '">&nbsp;</span>'
			msg_icon = iconList[0]
			if spam_level > 5:
				msg_icon = iconList[5]
			elif spam_level == 5:
				msg_icon = iconList[4]
			elif spam_level == 4:
				msg_icon = iconList[3]
			elif spam_level == 3:
				msg_icon = iconList[2]
			elif spam_level == 2:
				msg_icon = iconList[1]
			#end if
			smfrom = mfrom
			ssubject = sanitizeGtk(lsubject)
			lsubject = sanitizeGtk(lsubject)
			if self.currentDir == 'sent-mail':
				msg_row = [msg_icon, ssubject, mto, date, lsubject]
			else:
				msg_row = [msg_icon, ssubject, smfrom, date, lsubject]
			#end if
			if rainbow:
				# Easter egg
				rainbow = ['#ffc9c9','#ffe4c9','#ffffc9','#e4ffc9','#c9ffc9',
					'#c9ffe4','#c9ffff','#c9e4ff','#c9c9ff','#e4c9ff','#ffc9ff','#ffc9e4']
				background = rainbow[msg_i % len(rainbow)]
			elif self.currentDir != 'sent-mail':
				priority = self.getPriority(mfrom, spam_level)
				# eww yellow ffffc9
				# 1 green, 2 green-yellow, 3 white, 4 yellow, 5 orange, 6 red
				priority_background = {1:'#ccffcc', 2:'#e4ffc9', 3:'#ffffff', 4:'#fafae1', 5:'#ffdcba', 6:'#ffcccc'}
				background = priority_background[priority]
			else:
				background = '#ffffff'
			#end if
			# FIXME: mutex for auto_added
			thread = toThread(msg_row, threaded_rows, self.messages, show_i + self.auto_added, self.auto_added)
			if thread == None:
				msg_store.append(None, msg_row[:4] + [background])
				threaded_rows.append(msg_row[:5] + [[]])
				show_i += 1
			else:
				# adding auto_add fixes bug when auto update is occurring.
				thread = msg_store.iter_nth_child(None, thread + self.auto_added)
				msg_store.append(thread, msg_row[:4] + [background])
			#end if
			if msg_i_mod >= page_size:
				self.msg_table.thaw_child_notify()
				yield True
				self.msg_table.freeze_child_notify()
				msg_i_mod = -1
			#end if
			msg_i += 1
			msg_i_mod += 1
		#next message
		
		"""
		threaded_rows, showThread = toThreads(msgs_rows, show)
		self.messages = showThread
		#self.msg_table.freeze_child_notify()
		for row in threaded_rows:
			x = msg_store.append(None, row[:4])
			for row_child in row[5]:
				msg_store.append(x, row_child[:4])
			#next row_child
		#next row
		#self.msg_table.thaw_child_notify()
		"""
		self.msg_table.thaw_child_notify()
		print("indexed:", indexed)
		end = time.time()
		print('render dt:', end - curr)
		mail1.saveIndex(currentDir, page, self.index)
		self.drawDirRunning = False
		#return msg_list
		return True
	#end def drawDir(path = '/home/jvoss/Mail', currentDir = '', page = 0, msg_store=None)

	def drawMsg(self, mb, path, currentDir = '', currentMessage = '', add = '', use_part = None):
		"""
		Draw the message textview in the lower right.
		"""
		#if self.drawDirRunning:
		#	return False
		#end if
		if mb == None:
			print("Error: drawMsg called incorrectly, always need mb")
			return False
		#end if
		self.found = None
		self.attachments = {}

		if ('..' + os.sep) in currentDir or ('..' + os.sep) in currentMessage:
			# Alert instead?
			print("No directory traversal please...")
			mb.insert(mb.get_start_iter(), "No directory traversal please...")
			return False
		#end if
		
		black_color = Gdk_color_parse('#000000')
		green_light_color = Gdk_color_parse('#ccffcc')
		yellow_light_color = Gdk_color_parse('#ffffcc')
		red_light_color = Gdk_color_parse('#ffcccc')
		blue_light_color = Gdk_color_parse('#ccccff')
		blue_color = Gdk_color_parse('#0000ff')
		big_tag = mb.create_tag("big_yellow_bg", editable=False,
			foreground_rgba=black_color,
			background='yellow',
			size_points=12.0)
		normal_tag = mb.create_tag("normal", editable=False,
			foreground_rgba=black_color,
			size_points=10.0)
		spam_link_tag = mb.create_tag("spam_link", editable=False,
			foreground_rgba=blue_color,
			underline=Pango.Underline.SINGLE,
			size_points=10.0)
		spam_link_tag.connect('event', self.clickSpam)
		reply_link_tag = mb.create_tag("reply_link", editable=False,
			foreground_rgba=blue_color,
			underline=Pango.Underline.SINGLE,
			size_points=10.0)
		reply_link_tag.connect('event', self.clickReply)
		draft_link_tag = mb.create_tag("draft_link", editable=False,
			foreground_rgba=blue_color,
			underline=Pango.Underline.SINGLE,
			size_points=10.0)
		draft_link_tag.connect('event', self.clickDraft)
		headers_tag = mb.create_tag("headers", editable=False,
			foreground_rgba=blue_color,
			underline=Pango.Underline.SINGLE,
			size_points=10.0)
		headers_tag.connect('event', self.clickHeader)
		filter_link_tag = mb.create_tag("filter_link", editable=False,
			foreground_rgba=blue_color,
			underline=Pango.Underline.SINGLE,
			size_points=10.0)
		filter_link_tag.connect('event', self.clickFilter)
		html_link_tag = mb.create_tag("html_link", editable=False,
			foreground_rgba=blue_color,
			underline=Pango.Underline.SINGLE,
			size_points=10.0)
		html_link_tag.connect('event', self.clickHtml)
		html_text_link_tag = mb.create_tag("html_text_link", editable=False,
			foreground_rgba=blue_color,
			underline=Pango.Underline.SINGLE,
			size_points=10.0)
		html_text_link_tag.connect('event', self.clickHtmlText)
		html_clean_link_tag = mb.create_tag("html_clean_link", editable=False,
			foreground_rgba=blue_color,
			underline=Pango.Underline.SINGLE,
			size_points=10.0)
		html_clean_link_tag.connect('event', self.clickHtmlClean)
		html_footnote_link_tag = mb.create_tag("html_footnote_link", editable=False,
			foreground_rgba=blue_color,
			underline=Pango.Underline.SINGLE,
			size_points=10.0)
		html_footnote_link_tag.connect('event', self.clickHtmlFootnote)
		plain_link_tag = mb.create_tag("plain_link", editable=False,
			foreground_rgba=blue_color,
			underline=Pango.Underline.SINGLE,
			size_points=10.0)
		plain_link_tag.connect('event', self.clickPlain)
		file_link_tag = mb.create_tag("file_link", editable=False,
			foreground_rgba=blue_color,
			underline=Pango.Underline.SINGLE,
			size_points=10.0)
		file_link_tag.connect('event', self.clickFile)
		link_tag = mb.create_tag("link", editable=False,
			foreground_rgba=blue_color,
			underline=Pango.Underline.SINGLE,
			size_points=10.0)
		link_tag.connect('event', self.clickLink)
		bold_tag = mb.create_tag("bold", editable=False,
			foreground_rgba=black_color,
			weight=Pango.Weight.BOLD,
			size_points=10.0)
		red_tag = mb.create_tag("red", editable=False,
			foreground_rgba=black_color,
			paragraph_background_rgba=red_light_color,
			weight=Pango.Weight.BOLD,
			size_points=10.0)
		yellow_tag = mb.create_tag("yellow", editable=False,
			foreground_rgba=black_color,
			paragraph_background_rgba=yellow_light_color,
			weight=Pango.Weight.BOLD,
			size_points=10.0)
		green_tag = mb.create_tag("green", editable=False,
			foreground_rgba=black_color,
			paragraph_background_rgba=green_light_color,
			size_points=10.0)
		blue_tag = mb.create_tag("blue", editable=False,
			foreground_rgba=black_color,
			paragraph_background_rgba=blue_light_color,
			size_points=10.0)
		# When you change this, change drawMsgParts too.
		tags = (big_tag, normal_tag, spam_link_tag, reply_link_tag, draft_link_tag, headers_tag, html_link_tag, html_text_link_tag, html_clean_link_tag, html_footnote_link_tag, plain_link_tag, file_link_tag, link_tag, bold_tag, red_tag, yellow_tag, green_tag, blue_tag)

		# message body on the bottom right
		start_iter = mb.get_start_iter()
		#mb.insert_with_tags(start_iter, )
		#iter1 = self.get_start_iter()
		#iter2 = self.get_end_iter()
		curr_iter = start_iter
		
		#print('cd', currentDir)
		if not os.path.exists(os.path.join(path, currentDir, 'cur', currentMessage)):
			#print('drawDirRunning?', self.drawDirRunning)
			currentMessage2 = currentMessage + ':2,S'
			if not os.path.exists(os.path.join(path, currentDir, 'cur', currentMessage2)):
				print("File not found: ", os.path.join(path, currentDir, 'cur', currentMessage))
				mb.insert_with_tags(start_iter, 'File not found. Please try again.', big_tag)
				return False
			else:
				currentMessage = currentMessage2
			#end if
		#end if
		
		"""
		We don't have write access, so no go. This is easy but silly.
		if currentMessage[-4:] != ':2,S' and currentMessage[-5:] != ':2,RS':
			currentMessage2 = currentMessage + ':2,S'
			# It's been seen, so rename it (this is way weird)
			os.rename(path + '/' + currentDir + '/cur/' + currentMessage, path + '/' + currentDir + '/cur/' + currentMessage2)
			currentMessage = currentMessage2
		#end if
		"""
		
		#f = open(path + os.sep + currentDir + os.sep + 'cur' + os.sep + currentMessage, 'r')
		fname = os.path.join(path, currentDir, 'cur', currentMessage)
		fp = open(fname, 'rb')
		# TODO: decide whether we want to load this based on file length.
		#f.seek(0, 2)
		#file_size = f.tell()
		#f.seek(0, 0)
		msg = mail1.message_from_binary_file(fp, fname)
		self.msg = msg
		subject = msg['subject']
		ssubject = decodeHeaderGtk(subject, removeNewline=True)
		mfrom = decodeHeaderGtk(msg['from'], removeNewline=True)
		date = decodeHeaderGtk(msg['date'], removeNewline=True)
		mto = decodeHeaderGtk(msg['to'], removeNewline=True)
		message_id = None
		if 'Message-ID' in msg:
			message_id = decodeHeaderGtk(msg['Message-ID'])
		#end if
		mcc = None
		if 'cc' in msg:
			mcc = decodeHeaderGtk(msg['cc'])
		#end if
		has_valid_text_plain = False
		for b in msg.walk():
			ctype = b.get_content_type()
			if ctype.startswith('text/plain'):
				charset = None
				if ctype and 'charset=' in ctype:
					charset = ctype[ctype.index('charset=')+8:]
				#end if
				pdata = b.get_payload(decode=True)
				# Why not just errors='replace'? Encoding is more 
				# valuable than that.
				pdata = mail1.guessEncoding(pdata, charset)
				if mail1.deWhitespace(pdata).strip() != '':
					has_valid_text_plain = True
					break
				#end if
			#end if
		#next b
		if has_valid_text_plain:
			msg['has_valid_text_plain'] = '1'
		#end if

		mb.insert_with_tags(curr_iter, 'Reply', reply_link_tag)
		mb.insert_with_tags(curr_iter, ' ', normal_tag)
		mb.insert_with_tags(curr_iter, 'Spam', spam_link_tag)
		mb.insert_with_tags(curr_iter, ' ', normal_tag)
		mb.insert_with_tags(curr_iter, 'Filter', filter_link_tag)
		mb.insert_with_tags(curr_iter, ' ', normal_tag)

		mb.insert_with_tags(curr_iter, 'Headers', headers_tag)
		if currentDir.strip('/') == 'drafts':
			mb.insert_with_tags(curr_iter, ' ', normal_tag)
			mb.insert_with_tags(curr_iter, 'Resume Draft', draft_link_tag)
		#end if
		mb.insert_with_tags(curr_iter, '\n', normal_tag)
		mb.insert_with_tags(curr_iter, 'Subject: ', bold_tag)
		mb.insert_with_tags(curr_iter, ssubject + '\n', normal_tag)
		mb.insert_with_tags(curr_iter, 'From: ', bold_tag)
		mb.insert_with_tags(curr_iter, mfrom + '\n', normal_tag)
		mb.insert_with_tags(curr_iter, 'To: ', bold_tag)
		mb.insert_with_tags(curr_iter, mto + '\n', normal_tag)
		if mcc:
			mb.insert_with_tags(curr_iter, 'Cc: ', bold_tag)
			mb.insert_with_tags(curr_iter, mcc + '\n', normal_tag)
		#end if
		mb.insert_with_tags(curr_iter, 'Date: ', bold_tag)
		mb.insert_with_tags(curr_iter, date + '\n', normal_tag)
		if message_id:
			mb.insert_with_tags(curr_iter, 'Message-ID: ', bold_tag)
			mb.insert_with_tags(curr_iter, message_id + '\n', normal_tag)
		#end if
		#return output
		mb.create_mark("extra_headers", curr_iter, True)
		#mb.add_mark(mark_headers, curr_iter)
		self.showHeader = True
		self.extra_headers = msg.items()
		# See clickHeader for what goes here.
		return self.drawMsgParts(mb, curr_iter, msg, tags, add, use_part)
	#end def drawMsg(mb, path, currentDir = '', currentMessage = '', add = '', use_part = None)
	
	def drawMsgParts(self, mb, curr_iter, msg, tags, add = '', use_part = None, level = 1):
		"""
		Draw a message. Recursive due to PGP, but limited to 3 levels for now.
		Should only ever be one level since we only recurse on encryption. Who encrypts 3 levels? We do. What did you expect?
		"""
		if level > 3: return False
		html = (add == 'html')
		tryCleanHtml = False
		tryFootnoteHtml = False
		useTextHtml = False
		if add == 'clean_html':
			html = True
			tryCleanHtml = True
		if add == 'footnote_html':
			html = True
			tryFootnoteHtml = True
		elif add == 'text_html':
			html = True
			useTextHtml = True
		#end if
		app  = (add == 'app')
		gpg  = (add == 'gpg')
		(big_tag, normal_tag, spam_link_tag, reply_link_tag, draft_link_tag, headers_tag, html_link_tag, html_text_link_tag, html_clean_link_tag, html_footnote_link_tag, plain_link_tag, file_link_tag, link_tag, bold_tag, red_tag, yellow_tag, green_tag, blue_tag) = tags
		mime_ext = {'image/jpeg':'jpg', 'image/png':'png', 
		'image/gif':'gif', 'application/pdf':'pdf', 
		'application/zip':'zip', 'text/plain':'txt', 
		'application/pgp-encrypted':'txt', 'text/x-python': 'py', 
		'text/x-csrc': 'c', 'text/x-patch': 'patch', 'text/x-csrc': 'c', 
		'text/html':'html', 'text/x-c++src':'cxx', 'text/x-diff':'diff', 
		'audio/x-WAV':'wav', 'audio/ogg':'ogg', 'audio/flac':'flac', 
		'application/x-bittorrent':'torrent', 'application/x-rar-compressed':'rar',
		'video/mp4':'mp4', 'audio/mpeg':'mp3', 'audio/midi':'mid', 
		'video/x-msvideo':'avi', 'application/pgp-signature':'asc'}
		# TODO: disallow certain extensions
		#disallow_ext = ['exe', 'sh', 'py', 'bin']
		i = 0
		signed = False
		encrypted = False
		signed_data = None
		signed_part = None
		mark_body = None
		
		ctypes_available = [part.get_content_type() for part in msg.walk()]
		if add == '' and ('text/plain' not in ctypes_available):
			html = True
			tryCleanHtml = True
		#end if
		
		# A silly debug statement. put your username in if you want this assert.
		if 'jvoss' in self.maildir:
			assert([p for p in msg.walk()] == [q for q in mail1.walk(msg)])
		#end if
		
		for i, part in enumerate(mail1.walk(msg)):
			#if part.get_size() > 10000000:
			#	output = ''
			#end if
			ctype = part.get_content_type()
			f_raw = part.get_filename()
			if f_raw == None: f_raw = ''
			if f_raw != '':
				filename = decodeHeaderGtk(f_raw)
			else:
				filename = None
			#end if
			mb.insert_with_tags(curr_iter, sanitizeGtk(ctype) + '\n', normal_tag)
			#fname = part.get_param('name')
			if (ctype in mail1.KNOWN_MULTIPART):
				if ctype == 'multipart/signed':
					signed = True
					signed_part = i + 1
				#end if
				if ctype == 'multipart/encrypted':
					encrypted = True
				#end if
			elif ctype.startswith('image/'):
				ext = 'bin'
				if ctype in mime_ext: ext = mime_ext[ctype]
				if filename == None or (len(filename) == 0): filename = ('image{0:03d}.{1}'.format(i, ext))
				self.attachments[filename] = part
				mb.insert_with_tags(curr_iter, filename + '\n', file_link_tag)
			elif ctype.startswith('application/') or ctype.startswith('text/calendar') or filename != None:
				if ctype == 'application/pgp-encrypted':
					encrypted = True
				#end if
				ext = 'bin'
				if encrypted:
					ext = 'gpg'
					pdata = part.get_payload(decode=True)
					payload_decrypted = mail1.decryptPayload(pdata, self.keyid)
					if payload_decrypted != None:
						# Treat it like we just found a brand new text/plain.
						plaintext, good_sig, sig, good_msg, summary = payload_decrypted
						# Recursive now.
						#print(repr(plaintext), repr(good_sig), repr(sig), repr(good_msg))
						mb.insert_with_tags(curr_iter, "Encrypted Message\n", blue_tag)
						if good_sig:
							red = False
							mb.insert_with_tags(curr_iter, sig + "\n" + good_msg + "\n", green_tag)
							tag = green_tag
						else:
							red = True
							if summary != None:
								red = summary & gpg3.gpg.constants.sigsum.RED
							#end if
							tag = red_tag
							if not red: tag = yellow_tag
							mb.insert_with_tags(curr_iter, sig + "\n" + good_msg + "\n", tag)
						#end if
						# Encoding issues here. Matches /usr/lib64/python3.3/email/parser.py parsebytes.
						decrypted_msg = mail1.message_from_bytes(plaintext.encode('ASCII', errors='surrogateescape'))
						#print('dm', decrypted_msg)
						self.drawMsgParts(mb, curr_iter, decrypted_msg, tags, add, use_part, level + 1)
						curr_iter = mb.get_end_iter()
						if not red:
							mb.insert_with_tags(curr_iter, "End of Signed Message\n", tag)
						#end if
						mb.insert_with_tags(curr_iter, "End of Encrypted Message\n", blue_tag)
						continue
					#end if
				elif signed and signed_data and ctype == 'application/pgp-signature':
					# PGP/MIME signed
					sig = part.get_payload(decode=True)
					#print('sd', repr(signed_data), repr(sig))
					result = gpg3.GPG_verify2(signed_data, sig)
					red = True
					if not result:
						print("Warning: GPG_verify result is invalid", repr(result))
						result = (False, None, None, None)
					else:
						if not hasattr(result, 'signatures') or len(result.signatures) == 0:
							success = 0
							fpr, created, name = None, None, 'Unknown'
						else:
							success = result.signatures[0].summary & gpg3.gpg.constants.sigsum.VALID
							red = result.signatures[0].summary & gpg3.gpg.constants.sigsum.RED
							fpr = result.signatures[0].fpr
							created = result.signatures[0].timestamp
							if result.signatures[0].key:
								name = str(result.signatures[0].key)
							else:
								keys = gpg3.getKeylist(fpr)
								if len(keys) == 1:
									name = ', '.join(['{0} <{1}>'.format(uid.name, uid.email) for uid in keys[0].uids])
								elif len(keys) > 1:
									name = 'Multiple keys: '
									for key in keys:
										name += ', '.join(['{0} <{1}>'.format(uid.name, uid.email) for uid in key.uids]) + '; '
									#next key
								else:
									name = 'Unknown ' + fpr
								#end if
							#end if
						#end if
						if not success:
							message = ''
							if hasattr(result, 'signatures') and result.signatures:
								message = gpg3.summary_to_string(result.signatures[0].summary)
							#end if
							print("Warning: Bad signature 1", repr(result), hasattr(result, 'signatures') and result.signatures, message)
							good_msg = 'Signed by {0}'.format(name)
							if red:
								sig = "Warning: Bad signature {0}".format(message)
							else:
								sig = 'mail1: {0} Signature made {1}\nPrimary key fingerprint: {2}'.format(message, time.strftime('%a %d %b %Y %T %Z', time.localtime(float(created))), fpr)
							#end if
						#end if
						result = (success, fpr, created, name)
						(good_sig, fpr, created, name) = result
					#end if
					if good_sig:
						good_msg = 'Good signature from {0}'.format(name)
						sig = ('mail1: Signature made {0}\nPrimary key fingerprint: ' + fpr).format(time.strftime('%a %d %b %Y %T %Z', time.localtime(float(created))))
						mb.insert_with_tags(curr_iter, "End of Signed Message\n", green_tag)
						v_iter = mb.get_iter_at_mark(mark_body)
						#start_offset = v_iter.get_offset()
						#e_iter = mb.get_iter_at_mark(mark_body_end)
						#end_offset = e_iter.get_offset()
						# Before the body mark tag works so long as offset isn't 0.
						v_iter.set_offset(v_iter.get_offset() - 1)
						mb.insert_with_tags(v_iter, '\n' + sig + "\n" + good_msg, green_tag)
						#mark_body = mb.create_mark("body_start", v_iter, True)
						#doffset = v_iter.get_offset() - start_offset
						#e_iter.set_offset(end_offset + doffset)
						#mark_body_end = mb.create_mark("body_end", e_iter, True)
						curr_iter = mb.get_end_iter()
						continue
					else:
						print("Warning: Bad signature 2", repr(result))
						tag = red_tag
						if not red: tag = yellow_tag
						mb.insert_with_tags(curr_iter, "End of Signed Message\n", tag)
						v_iter = mb.get_iter_at_mark(mark_body)
						# Before the body mark tag works so long as offset isn't 0.
						v_iter.set_offset(v_iter.get_offset() - 1)
						mb.insert_with_tags(v_iter, '\n' + sig + "\n" + good_msg, tag)
						#mark_body = mb.create_mark("body_start", v_iter, True)
						curr_iter = mb.get_end_iter()
					#end if
				#else:
				#	print('sig without data', part.get_payload(decode=True))
				#end if
				
				if ctype in mime_ext: ext = mime_ext[ctype]
				if filename == None: filename = ('file{0:03d}.{1}'.format(i, ext))
				filenameu = filename
				if isinstance(filename, bytes):
					# TODO: Why not just errors='replace'? To fix bugs.
					try:
						filenameu = filename.decode('utf-8')
					except UnicodeDecodeError:
						filenameu = repr(filename)
					#end try
				#end if
				self.attachments[filenameu] = part
				mb.insert_with_tags(curr_iter, filenameu + '\n', file_link_tag)
			elif ctype.startswith('audio/'):
				ext = 'bin'
				if ctype in mime_ext: ext = mime_ext[ctype]
				if filename == None: filename = ('audio{0:03d}.{1}'.format(i, ext))
				filenameu = filename
					# TODO: Why not just errors='replace'? To fix bugs.
				if isinstance(filename, bytes):
					try:
						filenameu = filename.decode('utf-8')
					except UnicodeDecodeError:
						filenameu = repr(filename)
					#end try
				#end if
				self.attachments[filename] = part
				mb.insert_with_tags(curr_iter, filenameu, file_link_tag)
			elif ctype == 'text/html':
				if html == True:
					# Be very careful to ensure this does not get execed.
					pdata = part.get_payload(decode=True)
					ct = part.get('Content-Type')
					charset = None
					if ct and 'charset=' in ct:
						charset = ct[ct.index('charset=')+8:]
					#end if
					# Why not just errors='replace'? Encoding is more 
					# valuable than that.
					pdata = mail1.guessEncoding(pdata, charset)
					if not pdata.success:
						# warning. not always true.
						pdata = "Error: decoding failed, it's spam =)\n\n" + pdata
					#end if
					mark_body = mb.create_mark("body_start", curr_iter, True)
					if tryCleanHtml:
						# TODO GTK HTML?
						markdown = mail1.subNewline(mail1.anchorToMarkdown(mail1.clean_html(pdata)))
						mb.insert_with_tags(curr_iter, markdown, normal_tag)
					elif tryFootnoteHtml:
						markdown = mail1.subNewline(mail1.anchorToFootnote(mail1.clean_html(pdata)))
						mb.insert_with_tags(curr_iter, markdown, normal_tag)
					elif useTextHtml:
						mb.insert_with_tags(curr_iter, mail1.subNewline(mail1.text_html(pdata)), normal_tag)
					else:
						mb.insert_with_tags(curr_iter, sanitizeGtk(pdata), normal_tag)
					#end if
					mb.create_mark("body_end", curr_iter, True)
					# TODO: GTK HR
				#else:
				#end if
				if filename == None:
					filename = ''
				elif filename != '':
					filename = ' ' + filename
				#end if
				mb.insert_with_tags(curr_iter, '\n', normal_tag)
				mb.insert_with_tags(curr_iter, 'html' + filename + '\n', html_link_tag)
				mb.insert_with_tags(curr_iter, 'html text' + filename + '\n', html_text_link_tag)
				mb.insert_with_tags(curr_iter, 'html markdown' + filename + '\n', html_clean_link_tag)
				mb.insert_with_tags(curr_iter, 'html footnote' + filename + '\n', html_footnote_link_tag)
			elif ctype == 'text/plain':
				if (app or gpg) and (str(i) == use_part):
					ext = 'txt'
					if filename == None: filename = ('file{0:03d}.{1}'.format(i, ext))
					if gpg:
						# RFC3156 wants unencoded with headers.
						mb.insert_with_tags(curr_iter, '\n'.join([': '.join(a) for a in part.items()]) + '\n\n', normal_tag)
						mb.insert_with_tags(curr_iter, part.get_payload(decode=False), normal_tag)
					else:
						mb.insert_with_tags(curr_iter, part.get_payload(decode=True), normal_tag)
					#end if
					# TODO: GTK Create a file and save it somewhere.
					#return (ctype, output, filename)
				#end if
				if html == False:
					pdata = part.get_payload(decode=True)
					ct = part.get('Content-Type')
					# Use utf-8 because that's how the world works.
					charset = None
					if ct and ('charset=' in ct or 'charset =' in ct):
						if 'charset=' in ct:
							charset = ct[ct.index('charset=')+8:].strip()
						elif 'charset =' in ct:
							charset = ct[ct.index('charset =')+9:].strip()
						if ';' in charset: charset = charset[:charset.index(';')]
						if '"' in charset: charset = charset.replace('"','')
					#end if
					# Why not just errors='replace'? Encoding is more 
					# valuable than that.
					pdata = mail1.guessEncoding(pdata, charset)
					if not pdata.success:
						# warning.
						pdata = "Error: decoding failed, it's spam =)\n\n" + pdata
					#end if
					if '-----BEGIN PGP MESSAGE-----' in pdata:
						payload_decrypted = mail1.decryptPayload(pdata, self.keyid)
						if payload_decrypted != None:
							# Treat it like we just found a brand new text/plain.
							plaintext, good_sig, sig, good_msg, summary = payload_decrypted
							# Recursive now.
							#print(repr(good_sig), repr(sig), repr(good_msg))
							prefix = pdata[:pdata.index('-----BEGIN PGP MESSAGE-----')]
							suffix = ''
							if '-----END PGP MESSAGE-----' in pdata:
								suffix = pdata[pdata.index('-----END PGP MESSAGE-----')+len('-----END PGP MESSAGE-----\n'):]
							#end if
							if prefix: mb.insert_with_tags(curr_iter, prefix, normal_tag)
							mb.insert_with_tags(curr_iter, "Encrypted Message\n", blue_tag)
							if good_sig:
								tag = green_tag
								mb.insert_with_tags(curr_iter, sig + "\n" + good_msg + "\n", green_tag)
							else:
								red = True
								if summary != None:
									red = summary & gpg3.gpg.constants.sigsum.RED
								#end if
								tag = red_tag
								if not red: tag = yellow_tag
								mb.insert_with_tags(curr_iter, sig + "\n" + good_msg + "\n", tag)
							#end if
							mark_body = mb.create_mark("body_start", curr_iter, True)
							#print('plaintext', repr(plaintext))
							plaintext = plaintext.encode('utf-8', 'surrogateescape').decode('utf-8', 'replace')
							mb.insert_with_tags(curr_iter, sanitizeGtk(plaintext) + '\n', normal_tag)
							mb.create_mark("body_end", curr_iter, True)
							mb.insert_with_tags(curr_iter, "End of Signed Message\n", tag)
							mb.insert_with_tags(curr_iter, "End of Encrypted Message\n", blue_tag)
							if suffix: mb.insert_with_tags(curr_iter, suffix, normal_tag)
							# Don't print what we decrypted.
							continue
						#end if
					#end if
					if '-----BEGIN PGP SIGNED MESSAGE-----' in pdata:
						result = mail1.verifyClear(pdata)
						if (not result) or (len(result) != 2) or result[1] == None:
							result, plaintext = None, b''
						else:
							result, plaintext = result
							plaintext = plaintext.decode('utf-8', 'surrogateescape')
						#end if
						red = True
						if not result:
							print("Warning: GPG_verify result is invalid", repr(result))
							result = (False, None, None, None)
						else:
							if not hasattr(result, 'signatures') or len(result.signatures) == 0:
								success = 0
								fpr, created, name = None, None, None
							else:
								success = result.signatures[0].summary & gpg3.gpg.constants.sigsum.VALID
								red = result.signatures[0].summary & gpg3.gpg.constants.sigsum.RED
								fpr = result.signatures[0].fpr
								created = result.signatures[0].timestamp
								if result.signatures[0].key:
									name = str(result.signatures[0].key)
								else:
									keys = gpg3.getKeylist(fpr)
									if len(keys) == 1:
										name = ', '.join(['{0} <{1}>'.format(uid.name, uid.email) for uid in keys[0].uids])
									elif len(keys) > 1:
										name = 'Multiple keys: '
										for key in keys:
											name += ', '.join(['{0} <{1}>'.format(uid.name, uid.email) for uid in key.uids]) + '; '
										#next key
									else:
										name = 'Unknown ' + fpr
									#end if
								#end if
							#end if
							if not success:
								message = ''
								if hasattr(result, 'signatures') and result.signatures:
									message = gpg3.summary_to_string(result.signatures[0].summary)
								#end if
								print("Warning: Bad signature 1", repr(result), hasattr(result, 'signatures') and result.signatures, message)
								good_msg = 'Signed by {0}'.format(name)
								if red:
									sig = "Warning: Bad signature {0}".format(message)
								else:
									sig = 'mail1: {0} Signature made {1}\nPrimary key fingerprint: {2}'.format(message, time.strftime('%a %d %b %Y %T %Z', time.localtime(float(created))), fpr)
								#end if
							#end if
							result = (success, fpr, created, name)
						#end if
						(good_sig, fpr, created, name) = result
						if good_sig or plaintext:
							prefix = pdata[:pdata.index('-----BEGIN PGP SIGNED MESSAGE-----')]
							suffix = ''
							if '-----END PGP SIGNATURE-----' in pdata:
								suffix = pdata[pdata.index('-----END PGP SIGNATURE-----')+len('-----END PGP SIGNATURE-----\n'):]
							#end if
							if prefix: mb.insert_with_tags(curr_iter, prefix, normal_tag)
							#print(repr(plaintext), repr(good_sig), repr(sig), repr(good_msg))
							if good_sig:
								tag = green_tag
								good_msg = 'Good signature from {0}'.format(name)
								sig = ('mail1: Signature made {0}\nPrimary key fingerprint: ' + fpr).format(time.strftime('%a %d %b %Y %T %Z', time.localtime(float(created))))
								mb.insert_with_tags(curr_iter, sig + "\n" + good_msg + "\n", green_tag)
							else:
								tag = red_tag
								if not red: tag = yellow_tag
								mb.insert_with_tags(curr_iter, sig + "\n" + good_msg + "\n", tag)
							#end if
							mark_body = mb.create_mark("body_start", curr_iter, True)
							mb.insert_with_tags(curr_iter, sanitizeGtk(plaintext) + '\n', normal_tag)
							mb.create_mark("body_end", curr_iter, True)
							mb.insert_with_tags(curr_iter, "End of Signed Message\n", tag)
							if suffix: mb.insert_with_tags(curr_iter, suffix, normal_tag)
							# Don't print what we verified.
							continue
						#end if
					#end if
					mark_body = mb.create_mark("body_start", curr_iter, True)
					mb.insert_with_tags(curr_iter, sanitizeGtk(pdata) + '\n', normal_tag)
					mb.create_mark("body_end", curr_iter, True)
					# TODO: GTK HR
					#mb.insert_with_tags(curr_iter, 'plaintext' + '\n', link_tag)
				else:
					mb.insert_with_tags(curr_iter, 'text' + '\n', plain_link_tag)
				#end if
			else:
				# This is the else for content types. It gets called for:
				#message/delivery-status
				#message/rfc822
				# And other unknown message types.
				pdata = part.get_payload(decode=True)
				ct = part.get('Content-Type')
				charset = None
				if ct and 'charset=' in ct:
					charset = ct[ct.index('charset=')+8:]
				#end if
				# TODO: This is often a binary or something else...
				body = mail1.guessEncoding(pdata, charset)
				mb.insert_with_tags(curr_iter, body, normal_tag)
			#end if
			if signed and signed_part == i:
				# TODO: Make this link clickable.
				#mb.insert_with_tags(curr_iter, 'gpg' + '\n', link_tag)
				signed_data = mail1.as_bytes(part)
			#end if
		#next i, part
		return True
	#end def drawMsgParts(mb, path, currentDir = '', currentMessage = '', add = '', use_part = None)
	
	def searchMessage(self, widget, reverse=False):
		"""
		Search the message for a string.
		http://stackoverflow.com/questions/14994684/python-search-for-text-in-a-gtk-textview
		"""
		textbuffer = self.message_view.get_buffer()
		search_str = self.message_search.get_text()
		if self.search_str != search_str:
			start_iter = textbuffer.get_start_iter()
			self.search_str = search_str
		else:
			if reverse:
				start_iter = self.found[0] if self.found else textbuffer.get_start_iter()
			else:
				start_iter = self.found[1] if self.found else textbuffer.get_start_iter()
			#end if
		#end if
		#print('searching from', start_iter.get_offset(), self.found, self.search_str)
		flags = Gtk.TextSearchFlags.CASE_INSENSITIVE
		if reverse:
			self.found = start_iter.backward_search(search_str, flags, None)
		else:
			self.found = start_iter.forward_search(search_str, flags, None)
		#end if
		if self.found:
			match_start,match_end = self.found
			textbuffer.select_range(match_start, match_end)
			self.message_view.scroll_to_iter(self.found[0], 0, False, 0, 0)
		#end if
		return True
	#end def searchMessage(self, reverse=False)

	def showSearch(self, x=None):
		#print("Show Search Msg")
		self.message_search.show_all()
		self.message_search.grab_focus()
		return True
	#end def showSearch()

	def showSearchDir(self, x=None):
		#print("Show Search Dir")
		dialog = SearchDialog(self.window)
		response = dialog.run()
		unselected = False
		scrolled = False
		path = os.path.join(self.maildir, self.currentDir, 'cur')
		if response == Gtk.ResponseType.OK:
			search_term = dialog.entry.get_text()
			body_search = dialog.body_search.get_active()
			case_sensitive = dialog.case_sensitive.get_active()
			caseFunction = lambda x:x
			if not case_sensitive:
				search_term = search_term.lower()
				caseFunction = lambda x:x.lower()
			#end if
			model = self.msg_table.get_model()
			#print('sd', repr(search_term))
			# go through the message subjects for one.
			for i, row in enumerate(self.msg_store):
				found = []
				message = self.messages[i]
				if searchRow(search_term, row, message, path, case_sensitive, body_search):
					found.append([i])
				#end if
				if isinstance(message, list) and len(list(row.iterchildren())) != len(message[1]):
					print('error in searchdir', len(list(row.iterchildren())), repr(message))
					continue
				#end if
				for j, row_child in enumerate(row.iterchildren()):
					if searchRow(search_term, row_child, message[1][j], path, case_sensitive, body_search):
						found.append([i, j])
					#end if
				#next j, row_child
				#if row[1] == 'test':
				#	print(found, searchRow(search_term, row, message, path, case_sensitive, body_search))
				#end if
				if not found: continue
				#print(found)
				# TEMPORARY FOR SANITY
				#break
				parent_iter = model.iter_nth_child(None, i)
				for row_path in found:
					#print(row[1], row[2])
					if len(row_path) == 1:
						itera = parent_iter
					else:
						itera = model.iter_nth_child(parent_iter, row_path[1])
						parent_path = model.get_path(parent_iter)
						self.msg_table.expand_row(parent_path, False)
						#print('itera', itera, row_path)
					#end if
					patha = model.get_path(itera)
					#print('patha', patha, row_path)
					if not scrolled:
						self.msg_table.scroll_to_cell(patha, None, True, 0, 0)
						scrolled = True
					#end if
					sel = self.msg_table.get_selection()
					if not unselected:
						sel.unselect_all()
						unselected = True
					#end if
					sel.select_range(patha, patha)
					#sel.select_path(patha)
					#break
				#next row_path
			#next i, row
		dialog.destroy()
		return True
	#end def showSearchDir()

	def showAll(self):
		"""
		Calls self.window.show_all() and Gtk.main()
		"""
		self.running = True
		if self.config['Mail1GUI'].getboolean('update_loop', False):
			self.startUpdateLoop()
		#end if
		self.window.show_all()
		# Message search should only take up space when it's called upon.
		self.message_search.hide()
		Gtk.main()
		return True
	#end def showAll()
	
	def updateConfig(self, widget=None, event=None):
		"""
		Update the UI config values to the values currently used.
		"""
		window_size = self.window.get_size()
		messages_columns = self.msg_table.get_columns()
		#print(messages_columns)
		self.config['Mail1GUI']['width'] = str(window_size[0])
		self.config['Mail1GUI']['height'] = str(window_size[1])
		self.config['Mail1GUI']['subject'] = str(messages_columns[1].get_width())
		self.config['Mail1GUI']['from'] = str(messages_columns[2].get_width())
		self.config['Mail1GUI']['date'] = str(messages_columns[3].get_width())
		return True
	#end def updateConfig(widget=None, event=None)
	
	def saveConfig(self):
		"""
		Save the configuration on request.
		"""
		#print(dir(self.msg_table))
		mail1.saveConfig(self.config)
		return True
	#end def saveConfig()

	def gtkExit(self, widget=None):
		"""
		Useful quit function.
		"""
		#print("Quit was called", widget)
		self.running = False
		# Save that config file.
		self.saveConfig()
		if self.composer: self.composer.saveDraft()
		if self.addressBook: self.addressBook.close()
		Gtk.main_quit(widget)
		return True
	#end def gtkExit(widget)
	
	def handleWindows(self, message_wins):
		"""
		Exit when the message_wins are all closed.
		"""
		self.message_wins = message_wins
		self.message_wins_exit = True
		return True
	#end def handleWindows(message_wins)
	
	def windowClose(self, message_win):
		"""
		When a message window exits decide if we should exit also.
		"""
		if message_win in self.message_wins:
			self.message_wins.remove(message_win)
		#end if
		if self.message_wins_exit and self.message_wins == []:
			self.gtkExit(self.window)
		#end if
		return True
	#end def windowClose(message_win)
	
	def getPriority(self, mfrom, spam_level=None):
		"""
		1-6 based on how many times this person has mailed you.
		1 is green, 2 is green-yellow, 3 is white, 4 is yellow, 5 is orange, 6 is red.
		Possible to differentiate based on spammer with no other method.
		"""
		if not mfrom: return 3
		# TODO: This is expensive if it's the first time you've run it and you 
		# have a big sent-mail folder. Return 3 and put it in a thread or 
		# something.
		if self.addressBook == None:
			self.openAddressBook(show=False)
		#end if
		if self.addressBook == None:
			print("Error: with address book.")
			return 3
		#end if
		addr = strToAddrList(mfrom)
		# This is an warning condition but I don't know what to do about it.
		if len(addr) != 1: return 4
		addr = addr[0]
		if addr in self.addressBook.onlyAddrs:
			# If it's in the top half, they're a 1 green.
			addrPos = self.addressBook.onlyAddrs.index(addr)
			if addrPos < len(self.addressBook.onlyAddrs) >> 1:
				return 1
			else:
				# Bottom half, green-yellow
				return 2
			#end if
		#end if
		#print('spam_level', spam_level, mfrom)
		special = False
		if spam_level != None:
			priority = None
			if spam_level > 5:
				# Red
				priority = 6
			elif spam_level == 5:
				# Red
				priority = 6
			elif spam_level == 4:
				# Orange
				priority = 5
			elif spam_level == 3:
				# Yellow
				priority = 4
			elif spam_level == 2:
				# Spam level 2 and below is either good or unknown.
				priority = None
				special = True
			#end if
			if priority != None:
				return priority
			#end if
		#end if
		if self.addressBook.knownAddrs == None:
			self.addressBook.loadKnownAddrs()
		#end if
		#print('known addrs:', len(self.addressBook.knownAddrs), self.addressBook.knownAddrs[:5])
		if addr in self.addressBook.knownAddrs:
			return 3
		#end if
		if special:
			# They got a 2 on the spam level, so they're orange.
			return 5
		#end if
		return 4
	#end def getPriority(mfrom, spam_level=None)
	
	def updateDir(self):
		"""
		Copy from mail1 main
		If the user doesn't want to run sh am.sh, they can call this
		"""
		# No we cannot sleep.
		curTime = time.time()
		if curTime - self.updateDirLast < 2:
			return True
		#end if
		self.updateDirLast = curTime
		#print("update")
		filterFilename = 'list_filter1.json'
		home = mail1.home
		hidden_dir = mail1.hidden_dir
		# TODO: variable ~/.maildir
		dotMaildir = os.path.join(home, '.maildir')
		sentDest = os.path.join(self.maildir, 'sent-mail', 'cur', '')
		if not os.path.exists(sentDest):
			os.makedirs(sentDest)
		#end if
		if 'Mail1' not in self.config:
			self.config['Mail1'] = {'inbox': 'inbox'}
		#end if
		inboxDir = self.config['Mail1']['inbox']
		inboxCur = os.path.join(self.maildir, inboxDir, 'cur')
		moved = []
		moved2 = []
		if os.path.exists(os.path.join(dotMaildir, 'cur')):
			moved2 = mail1.moveAll(os.path.join(dotMaildir, 'cur'), inboxCur)
			moved = mail1.moveAll(os.path.join(dotMaildir, 'new'), inboxCur)
		#end if
		moved3 = mail1.moveAll(os.path.join(self.maildir, 'sent-mail', 'new', ''), 
					sentDest)
		if os.path.exists(os.path.join(hidden_dir, 'filter', filterFilename)):
			x, changes = mail1.filterList(filterFilename, moved, self.maildir, inboxDir)
		else:
			x = moved
		#end if
		if self.currentDir == 'sent-mail':
			moved += moved3
			# x gets a ref of moved so it doesn't need to be changed.
		#end if
		# x can be a string or a list.
		if x:
			scroll = False
			iconList = [self.unreadPixBuf, self.twoPixBuf, self.threePixBuf, self.fourPixBuf, self.fivePixBuf, self.spamPixBuf]
			iconListAtt = [self.attachPixBuf, self.twoAttachPixBuf, self.threeAttachPixBuf, self.fourAttachPixBuf, self.fiveAttachPixBuf, self.attachSpamPixBuf]
			print(x)
			currentPath = os.path.join(self.maildir, self.currentDir, 'cur')
			if inboxDir == self.currentDir:
				changes_fn = [filename for filename, path in changes]
				moved_here = [filename for filename in moved if filename not in changes_fn]
			else:
				moved_here = [filename for filename, path in changes if path == currentPath]
			#end if
			# Merge those files into the current view
			for filename in moved_here:
				# Depends on whether it was moved.
				if not os.path.exists(os.path.join(currentPath, filename)):
					print("Error: no file in current dir", filename)
					continue
				#end if
				# TODO: Order, threading
				self.messages.insert(0, filename)
				background = '#ffffff'
				iconListUse = iconList
				msg = getMessage(filename, {}, currentPath)
				if msg.get('attach', None) == '1':
					iconListUse = iconListAtt
				msg_row_data = getMessageRow(msg, self.config, self.blacklist, self.blacklist_ip, self.whitelist, iconListUse)
				if self.currentDir == 'sent-mail':
					msg_row = msg_row_data[:2] + msg_row_data[3:]
				else:
					msg_row = msg_row_data[:3] + msg_row_data[4:]
				#end if
				if self.currentDir != 'sent-mail':
					mfrom = msg_row_data[6]
					spam_level = msg_row_data[7]
					priority = self.getPriority(mfrom, spam_level)
					# eww yellow ffffc9
					# 1 green, 2 green-yellow, 3 white, 4 yellow, 5 orange, 6 red
					priority_background = {1:'#ccffcc', 2:'#e4ffc9', 3:'#ffffff', 4:'#fafae1', 5:'#ffdcba', 6:'#ffcccc'}
					background = priority_background[priority]
				#end if
				self.msg_table.freeze_child_notify()
				self.msg_store.insert(None, 0, msg_row[:4] + [background])
				self.auto_added += 1
				self.msg_table.thaw_child_notify()
				scroll = True
			#next filename
			if scroll:
				itera = self.msg_store.iter_nth_child(None, 0)
				patha = self.msg_store.get_path(itera)
				self.msg_table.scroll_to_cell(patha, None, True, 0, 0)
			#end if
		#end if
		# Must return true or else idle_add stops calling us.
		return True
	#end def updateDir()
	
	def startUpdateLoop(self):
		"""
		Run a second thread that updates.
		"""
		self.updateEvent = GObject.idle_add(self.updateDir)
		return True
	#end def startUpdateLoop()
	
	def groupBySender(self, widget=None):
		"""
		Change the directory to list one row per sender, then file all items under that tree item.
		"""
		# Get out of the drawDir loop or else it will fail.
		self.drawDirExit = True
		if self.drawDirRunning:
			#print('got to the idle_add')
			# Try again when idle.
			if self.readyToGroup == None:
				self.readyToGroup = GObject.idle_add(self.groupBySender)
			#end if
			return True
		#end if
		if self.readyToGroup != None:
			# Stop calling groupBySender when we finally made it to the core.
			GObject.source_remove(self.readyToGroup)
			self.readyToGroup = None
		#end if
		#print('got to the core')
		messages = []
		rows = []
		for i in range(min(len(self.msg_store), len(self.messages))):
			#print(self.messages[i], list(self.msg_store[i][:5]))
			rows.append(self.msg_store[i][:5])
			if isinstance(self.messages[i], list):
				# parent
				messages.append(self.messages[i][0])
				# Children
				for j in range(len(self.messages[i][1])):
					#print('   ', self.msg_store[i,j][:5])
					rows.append(self.msg_store[i,j][:5])
					messages.append(self.messages[i][1][j])
				#next j
			else:
				messages.append(self.messages[i])
			#end if
			#if i > 20: break
			#self.msg_store[i]
		#next i
		output_list, show = groupBySender(messages, rows)
		self.msg_table.disconnect(self.cursorChangedConnection)
		self.msg_store.clear()
		self.auto_added = 0
		# TODO: Disable update loop
		self.messages = show
		for msg_row in output_list:
			q = self.msg_store.append(None, msg_row[0])
			for child_row in msg_row[1]:
				self.msg_store.append(q, child_row)
			#next child_row
		#next msg_row
		self.cursorChangedConnection = self.msg_table.connect('cursor-changed', self.clickMsg)
		return True
	#end def groupBySender(widget=None)
#end class MaildirWindow

def main():
	import argparse
	# Calling GObject.threads_init() is not needed for PyGObject 3.10.2+ *shrug*
	GObject.threads_init()
	parser = argparse.ArgumentParser(description="BikeIM Mail Client")
	parser.add_argument('-w', '--window', action='store_true')
	parser.add_argument('filename', nargs="*")
	args = parser.parse_args()
	#print(args)
	maildir = os.path.join(os.path.expanduser('~'), 'Mail')
	# if args.window is true, we don't want to show the main window.
	w = MaildirWindow(maildir, args.window)
	message_wins = []
	if args.filename:
		useParentWindow = False
		if len(args.filename) > 1:
			useParentWindow = True
		#end if
		for filename in args.filename:
			if args.window:
				# TODO: multithreaded but GPG passphrase
				w.showMsg(filename, dontShow=True)
				message_win = openMessage(w, useParentWindow)
				message_wins.append(message_win)
			else:
				w.showMsg(filename)
			#end if
		#next filename
	#end if
	if not args.window:
		# Show the main window.
		w.showAll()
	else:
		# Someone needs to call Gtk.main(). If there's one, it's them.
		# If there's more than one, it's us.
		if len(args.filename) > 1:
			# Don't show the main window. Need to call gtk main myself.
			w.handleWindows(message_wins)
			Gtk.main()
		elif len(message_wins) == 1:
			message_wins[0].showAll()
		#end if
	#end if
	# It works!
#end def main()

if __name__ == '__main__':
	main()
#end if
